%*********************************************************************;
%*****         Calling recursively the built-in relations         ****;
%*********************************************************************;

module riccir!-callalgs;

% Author: Jozsef Kadlecsik

global '(!*msfone!* christoffel!* metric!* riemann!*);

fluid '(!*allterm !*curralg!* !*currten !*currtenfn !*ftermn 
        !*multtenfn !*only!-onefn !*revrelfn !*righttenfn);

% Calling the relations;

symbolic procedure call!-algs(u,v);
   begin scalar !*currtenfn,!*multtenfn,!*only!-onefn,!*currten,
                !*righttenfn,!*revrelfn;
      return if subfg!* and v then mp3msq call!-algs0(u,v) else u
   end;

symbolic procedure call!-algs0(u,v);
   if null v then u
    else call!-algs0(call!-algs!*(u,car v),cdr v);

symbolic procedure call!-algs!*(u,v);
   begin scalar !*curralg!*;
      return if v = '(partial covariant) and christoffel!*
               then call!-gdiffrel(u,'part!-cov) 
              else if v = '(covariant partial) and christoffel!*
               then call!-gdiffrel(u,'cov!-part)
              else if v = '(covariant lie) 
               then call!-gdiffrel(u,'cov!-lie)
              else if v = '(lie covariant) 
               then call!-gdiffrel(u,'lie!-cov)
              else if v = '(riemann ricci) 
               then << !*curralg!* := 1; remsq u>>
              else if v = '(ricci riemann) 
               then << !*curralg!* := 2; remsq u>>
              else if v = '(ricci invcurv) 
               then << !*curralg!* := 3; remsq u>>
              else if v = '(invcurv ricci) 
               then << !*curralg!* := 4; remsq u>>
              else if v = '(weyl riemann) 
               then << !*curralg!* := 5; remsq u>>
              else if v = '(riemann weyl) 
               then << !*curralg!* := 6; remsq u>>
              else if v = '(riemann definition) 
               then << !*curralg!* := 7; remsq u>>
              else if v = '(definition riemann) 
                      and christoffel!* and riemann!*
               then << set!-revrelfns 'riemann1; call!-revrel u>>  
              else if v = '(christoffel definition) 
               then << !*curralg!* := 8; remsq u>>
              else if v = '(definition christoffel) 
                      and metric!* and christoffel!*
               then << set!-revrelfns 'christoffel; call!-revrel u>>
              else if v = '(riemann secondcov) 
               then << set!-revrelfns 'riemann2; call!-revrel u>>
              else if eqcar(v,'bianchy1) 
               then << !*curralg!* := v; remsq u>>
              else if eqcar(v,'bianchy2) 
               then << !*curralg!* := v; remsq u>>
              else if v = 'bianchyc 
               then << !*curralg!* := 9; remsq u>>
              else if eqcar(v,'jacobi) 
               then << !*curralg!* := cdr v; call!-gdiffrel(u,'jacobi)>>
              else if v = 'ordercov and riemann!*
               then call!-gdiffrel(u,'ordercov)
              else if eqcar(v,'ordercov) and riemann!*
               then << !*curralg!* := cdr v; call!-gdiffrel(u,'ordercov)>>
              else u
   end;      
            
% 'currtenfn finds the given tensor in the argument of diff.ops, too.;

deflist('((riemann1 gdiff!-ten)
          (christoffel gdiff!-ten)
          (riemann2 riem!-ten)
         ),'currtenfn);             

% 'multtenfn gives the multiplicative term (if exists);

deflist('((riemann1 no!-mult!-ten)
          (christoffel mult!-metr)
          (riemann2 mult!-rten)
         ),'multtenfn);             

% 'only!-onefn works if the relation consists of one term only;

deflist('((riemann1 nil)
          (christoffel nil)
          (riemann2 t)
         ),'only!-onefn);           

% 'currten checks the given tensor;

deflist('((riemann1 partd!-chris)
          (christoffel partd!-metr)
          (riemann2 riemp)
         ),'currten);              

% 'righttenfn checks the two leading terms from the relation;

deflist('((riemann1 right!-riemdef)
          (christoffel right!-chrsdef)
          (riemann2 nil)
         ),'righttenfn);            

% 'revrelfn returns  (non evaluated left hand side) . 
%                   -(evaluated right hand side) of the given relation;

deflist('((riemann1 revrel!-riemdef)
          (christoffel revrel!-chrisdef)
          (riemann2 revrel!-riemdef2)
         ),'revrelfn);            

symbolic procedure set!-revrelfns u;
   << !*currtenfn   := get(u,'currtenfn);
      !*multtenfn   := get(u,'multtenfn);
      !*only!-onefn := get(u,'only!-onefn); 
      !*currten     := get(u,'currten);
      !*righttenfn  := get(u,'righttenfn); 
      !*revrelfn    := get(u,'revrelfn);
      !*curralg!*   := u>>;

symbolic procedure call!-revrel u;
   (if null x then rederr "Zero divisor"
     else multmsq(call!-revrel0(numr u,nil),
                  revpr x)) where x=call!-revrel0(denr u,nil);

symbolic procedure call!-revrel0(u,v);
   if null u or lpow u = 1 then addmsf!*!*(v,u) ./ !*msfone!*
    else (if x then apply!-revrel(u,x,v)
           else call!-revrel0(red u,addmsf!*!*(v,lt u .+ nil)))
              where x=find!-lterms(lpow u,lt u);

symbolic procedure find!-lterms(u,v);
   % The elementary str(ucture) returned is:  (op1 . <op2>) . orig
   % A term may contain more str, so from a term: nil | list of str's;
   if null u or numberp cdar u then nil
    else ((if x and y
             then if y eq 'one
                    then ((x . nil) . v) . find!-lterms(cdr u,v)
                    else if flagp(!*multtenfn,'more)
                          then append(mk!-from!-more(y,v),
                                      find!-lterms(cdr u,v))
                   else ((x . y) . v) . find!-lterms(cdr u,v)
            else find!-lterms(cdr u,v))
                  where y=(x and apply2(!*multtenfn,x,car v)))
              where x=apply1(!*currtenfn,car u);

symbolic procedure mk!-from!-more(u,v);
   (for each j in cdr u collect (x . j) . v) where x=car u; 

% Main working procedure I.;

symbolic procedure apply!-revrel(u,v,w);
   begin scalar x;
      v := if !*only!-onefn 
             then for each j in v 
                     collect (car j . (delete(caar j,cadr j) . cddr j))
            else find!-two!-terms(v,red u);
      return if v and (x := apply!-revrel0(u,v))
               then call!-revrel addmsq(x,w ./ !*msfone!*)
              else call!-revrel0(red u,addmsf!*!*(w,lt u .+ nil))
   end;

% If there are two part from two different terms, then let's check them;
% A leading term from the examined expression may contain more true parts.

symbolic procedure find!-two!-terms(u,v);
   if null v or lpow v = 1 then nil
    else append(find!-two!-terms0(u,find!-lterms(lpow v,lt v)),
                find!-two!-terms(u,red v));

symbolic procedure find!-two!-terms0(u,v); 
   if null v then nil
    else (if x then append(x,find!-two!-terms0(u,cdr v))
           else find!-two!-terms0(u,cdr v))
            where x=two!-terms(u,car v);

symbolic procedure two!-terms(u,v);
   if null u then nil
    else (if x then x . two!-terms(cdr u,v)
           else two!-terms(cdr u,v))
            where x=two!-terms0(car u,v);

symbolic procedure two!-terms0(u,v);
   apply(!*righttenfn,list(caar u,cdar u,cadr u,cddr u,
                           caar v,cdar v,cadr v,cddr v));
 
% Main working procedure II.;
     
symbolic procedure apply!-revrel0(u,v);
   % U : original expression;
   % V : terms being found;
   if null v then nil
    else (if x then x else apply!-revrel0(u,cdr v))
            where x=curr!-revrel(u,cdar v,mk!-curr!-revrel caar v);

symbolic procedure curr!-revrel(u,v,w);
   % U : original expression
   % V : ( (ops.) . sf) coefficient
   % W : left hand side . -(right hand side) of the relation;
   begin scalar x,y;
      if not msfonep denr (x := riccirsimp!* cdr w) 
        then << lprim "Please send input-output listing to J. Kadlecsik";
                return nil>>   %!!!!!!
       else x := numr x;
      v := (if car v then v else 1 .* cdr v) .+ nil;
      y := revrel!-addmsf(u,multmsf(v,x));
      return if null y then nil 
              else addmsq(y,multmsq(v ./ !*msfone!*,riccirsimp!* car w))
   end; 

symbolic procedure revrel!-addmsf(u,v);
   begin integer n,!*ftermn;
      n := length v;
      u := revrel!-addmsf0(u,v);
      return if (!*allterm and !*ftermn=n)
                or ((n=1 or n=2) and !*ftermn=n) 
                or (not !*allterm and n leq 2*!*ftermn)
               then u ./ !*msfone!* else nil
   end;

symbolic procedure revrel!-addmsf0(u,v);
   % U : original expression
   % V : ( (ops. . sf) * -(evaluated right hand side of the relation);
   if null u then v
    else if null v then u
    else if lpow u = 1 then addmsfm(u,v)
    else if lpow v = 1 then addmsfm(v,u)
    else if lpow u = lpow v 
     then (if null x then << !*ftermn := !*ftermn + 1; y>> 
            else lpow u .* x .+ y)
              where x=addf(lc u,lc v),y=revrel!-addmsf0(red u,red v)
    else if riccir!-ordp(lpow u,lpow v) then lt u .+ revrel!-addmsf0(red u,v)
    else lt v .+ revrel!-addmsf0(u,red v);

symbolic procedure mk!-curr!-revrel u;
   if apply1(!*currten,car u) then apply2(!*revrelfn,car u,cdr u)
    else (caar u . car x . cddar u) . (caar u . cdr x . cddar u)
              where x=mk!-curr!-revrel (cadar u . cdr u);

symbolic procedure addmsf!*!*(u,v);
   if null u then v else lt u .+ addmsf!*!*(red u,v);

endmodule;

end;
