%********************************************************************;
%*****                        Declarations                       ****;
%********************************************************************;

module riccir!-decls;

% Author: Jozsef Kadlecsik

global '(indexlist!* signature!* stdim!* symbollist!* tensorlist!*);

% indexlist!*   =   List of user defined index identifiers;
% symbollist!*  =   List of user defined symbol identifiers;
% tensorlist!*  =   List of user defined tensor identifiers;

% A tensor or symbol identifier may be index identifier, too;

% signature!*   =   Signature of metric tensor;
% stdim!*       =   Dimension of space-time;

rlistat('(tensor symbol!* index spacedim signature));

symbolic procedure tensor u;
   %
   %   TENSOR id1,id2,...,idn; 
   %
   begin rmsubs();
      for each j in u do 
         if atom j then mktensor j
          else errpri2(j,'hold)
   end;

symbolic procedure mktensor u;
   if flagp(u,'diffop) 
     then typerr(list("Differential operator",u),"tensor.")
    else if u memq symbollist!* 
     then typerr(list("Symbol",u),"tensor.")
    else << put(u,'simpfn,'riccirsimpiden);
            put(u,'idvalfn,'riccirsimpiden);
            put(u,'riccirfn,'ricciriden);
            put(u,'rtype,'indexed!-quantity);
            flag(list u,'full);
            put(u,'prifn,'indvarprt);
            put(u,'fancy!-pprifn,'fancy!-indvarprt);  % added: HM
            tensorlist!* := union(list u,tensorlist!*)>>;

%**** Because of the existing 'symbol procedure of REDUCE, a parsing
%**** procedure is needed to avoid the confusion;

put('symbol,'stat,'symbolstat); 
                                 
symbolic procedure symbolstat;
   begin scan();
      return 'symbol!* . remcomma xread1 'lambda
   end;

symbolic procedure symbol!* u;
   %
   %   SYMBOL id1,id2,...,idn;   
   %
   begin rmsubs();
      for each j in u do 
         if atom j then mksymbol j
          else errpri2(j,'hold)
   end;

symbolic procedure mksymbol u;
   if flagp(u,'diffop) 
     then typerr(list("Differential operator",u),"symbol.")
    else if u memq tensorlist!* 
     then typerr(list("Tensor",u),"symbol.")
    else << put(u,'simpfn,'riccirsimpiden);
            put(u,'idvalfn,'riccirsimpiden);
            put(u,'riccirfn,'ricciriden);
            put(u,'rtype,'indexed!-quantity);
            flag(list u,'full);
            put(u,'prifn,'indvarprt);
            put(u,'fancy!-pprifn,'fancy!-indvarprt);  % added: HM
            symbollist!* := union(list u,symbollist!*)>>;

put('indexed!-quantity,'revlisfn,'riccir!-revlis);

symbolic procedure index u;
   %
   %   INDEX id1,id2,...,idn;
   %
   begin rmsubs();
      if atomlis u then indexlist!* := ordn union(indexlist!*,u)
       else errpri2(u,'hold)
   end;

put('index,'clearfn,'clear!-index);
put('clear!-index,'formfn,'formrlis);

symbolic procedure clear!-index u;
   %
   %   CLEAR INDEX id1,id2,...,idn;
   %
   begin rmsubs();
      if atomlis u then indexlist!* := ordn setdiff(indexlist!*,u)
       else errpri2(u,'hold)
   end;

symbolic procedure spacedim u;
   %
   %   SPACEDIM sd;       sd ::= integer | id
   %
   begin rmsubs();
      if null cdr u and car u then stdim!-signature(car u,signature!*)
       else errpri2(u,'hold)
   end;

put('spacedim,'clearfn,'clear!-spacedim);
put('clear!-spacedim,'formfn,'formgenclear);

symbolic procedure clear!-spacedim u; stdim!* := nil;

symbolic procedure signature u; 
   %
   %   SIGNATURE sg;      sg ::= integer | id
   %
   begin rmsubs();
      if null cdr u and car u then stdim!-signature(stdim!*,car u)
       else errpri2(u,'hold)
   end; 

symbolic procedure clear!-signature u; signature!* := nil;

put('signature,'clearfn,'clear!-signature);
put('clear!-signature,'formfn,'formgenclear);

flag('(clear!-spacedim clear!-signature),'emptyarg);

symbolic procedure stdim!-signature(u,v);
   begin
      if u 
        then << if denr (u := simp u) neq 1  
                  then typerr(prepsq u,"the dimension of space-time.")
                  else u := if null numr u then 0 else numr u;
                % the dimension is an id. or a positive integer;
                if numberp u and not(u>0)
                  then typerr(u,"the dimension of space-time.")>>;
      if v
        then if null u 
               then rerror('riccir!-decls,1,
                  "The dimension of space-time is not defined yet.")
              else if denr(v := simp v) neq 1 
               then typerr(prepsq v,
                           "the signature of the metric tensor.")
              else v := if null numr v then 0 else numr v;
      % u is the dimension, v is the signature
      % then u>=abs v and (u+v)/2 is an integer
      if u and v and numberp u and numberp v
         and (u<abs v or null quotdd(u+v,2))
        then rerror('riccir!-decls,2,list("The signature",v,
                    "and the dimension",u,"are not compatible."));
      if u then stdim!* := prepf u;
      if v then signature!* := prepf v;
  end;

endmodule;

end;
