%********************************************************************;
%*****                  Definitions and relations                ****;
%********************************************************************;

module riccir!-defs;

% Author: Jozsef Kadlecsik

global '(!*msfone!* christoffel!* diffindices!*
         metric!* riemann!* signature!* stdim!* weyl!*);

fluid '(!*curralg!*);

% I. Definitions

% Definition of the Christoffel symbol

symbolic procedure christoffel!-def(u,v,w);
   if downp u then christoffel!-def0(u,v,w)
    else christoffel!-def1(u,v,w);

symbolic procedure christoffel!-def1(u,v,w);
   %     
   %     u      1   un
   %  GAM    = --- g  * ( g     + g     - g    )
   %      vw    2          nv,w    nw,v    vw,n
   %
   list('times,
        list(metric!*,u,'!&000),
        christoffel!-def0('(minus !&000),v,w));    

symbolic procedure christoffel!-def0(u,v,w);
   list('quotient,
        list('plus,
             metric!-diff(u,v,w),
             metric!-diff(u,w,v),
             list('minus,metric!-diff(v,w,u))),
        2);

symbolic procedure metric!-diff(u,v,w);
   list('partdiff,list(metric!*,u,v),w);

% Definition of the Riemann tensor;

symbolic procedure riemann!-def u;
   begin scalar v,w,z,x;
      v := cadr u;
      w := caddr u;
      z := cadddr u;
      u := car u;
      x := list(riemann!-def0(indexabs u,indexabs v,
                              indexabs w,indexabs z),1);
      if downp u then x := list(metric!*,u,'(minus !&001)) .
                               indexsubst('!&001 ,cadr u,x);
      if atom v then x := list(metric!*,v,'!&002) .
                              indexsubst('(minus !&002),indexinv v,x);
      if atom w then x := list(metric!*,w,'!&003) .
                              indexsubst('(minus !&003),indexinv w,x);
      if atom z then x := list(metric!*,z,'!&004) .
                              indexsubst('(minus !&004),indexinv z,x);
      x := 'times . x;
      return if cadr riemann!* = 1 then x      % MTW definition;
              else list('minus,x)
   end;

symbolic procedure riemann!-def0(u,v,w,z);
   %     
   %   u         u          u          u        n        u        n
   %  R     = GAM      - GAM      + GAM    * GAM    + GAM    * GAM
   %    vwz       vz,w       vw,z       wn       vz       zn       vw
   %
   list('plus,
         christoffel!-diff(u,v,z,w),
         list('minus,christoffel!-diff(u,v,w,z)),
         christoffel!-prod(u,w,'!&005,'!&005,v,z),
         list('minus,christoffel!-prod(u,z,'!&005,'!&005,v,w)));

symbolic procedure christoffel!-diff(u,v,w,z);
   list('partdiff,
        list(christoffel!*,u,list('minus,v),list('minus,w)),
        list('minus,z));

symbolic procedure christoffel!-prod(u,v,w,a,b,c);
   list('times,
        list(christoffel!*,u,list('minus,v),list('minus,w)),
        list(christoffel!*,a,list('minus,b),list('minus,c)));

% Definition of the Ricci tensor;

symbolic procedure ricci!-def(u,v);
   %
   %         n
   %  R   = R
   %   uv     unv
   %
   (if cddr riemann!* = 1 then x      % MTW definition;
     else list('minus,x))
       where x= list(car riemann!*,'!&006,u,'(minus !&006),v);

% Definition of the invariant curvature (Einstein scalar);

symbolic procedure invcurv!-def;
   %
   %       n
   %  R = R
   %        n
   %
   list(car riemann!*,'!&007,'(minus !&007));

% Definition of the Weyl tensor;

symbolic procedure weyl!-out!-def(u,v,w,z);
   %
   %                          s2
   %  W     = s1 * R     + ------- * ( g   * R   - g   * R   
   %   uvwz         uvwz    n - 2       uw    vz    uz    vw
   %
   %                                 + g   * R   - g   * R  ) 
   %                                    vz    uw    vw    uz
   %
   %                      R
   %             + --------------- * ( g   * g   - g   * g  )
   %               (n - 2)*(n - 1)      uz    vw    uw    vz
   %
   %    where n is the dimension of the space-time,
   %    and s1,s2 are determined by the signs of the definitions of 
   %    the Riemann and Ricci tensors;
  list('plus,
       list('times,
             signs!-def(),
             list(car riemann!*,u,v,w,z)),
       weyl!-def!-part(u,v,w,z));

symbolic procedure weyl!-in!-def(u,v,w,z);
   list('times,
        signs!-def(),
        list('plus,
             list(weyl!*,u,v,w,z),
             list('minus,weyl!-def!-part(u,v,w,z)))); 

symbolic procedure signs!-def;
   if minusp((cadr riemann!*)*(cddr riemann!*)) then '(minus 1) 
    else 1;
 
symbolic procedure weyl!-def!-part(u,v,w,z);
   list('plus,
        list('times,
             list('quotient,
                  if minusp(cddr riemann!*) then 1 else '(minus 1),
                  list('plus,stdim!*,'(minus 2))),
             list('plus,
                  ricci!-metric!-prod(u,w,v,z),
                  list('minus,ricci!-metric!-prod(u,z,v,w)),
                  list('minus,ricci!-metric!-prod(v,w,u,z)),
                  ricci!-metric!-prod(v,z,u,w))),
        list('times,
             list('quotient,1,
                   list('times,list('plus,stdim!*,'(minus 1)),
                               list('plus,stdim!*,'(minus 2)))),
             car riemann!*,
             list('plus,
                  metrics!-prod(u,w,v,z),
                  list('minus,metrics!-prod(u,z,v,w)))));
                
symbolic procedure ricci!-metric!-prod(u,v,w,z);
   list('times,
        list(car riemann!*,u,v),
        list(metric!*,w,z));
 
symbolic procedure metrics!-prod(u,v,w,z);
   list('times,
        list(metric!*,u,v),
        list(metric!*,w,z));

% II. Identities

% First Bianchy identity

%
%   i         i       i
%  R     = - R     - R
%    jkl       klj     ljk
%

symbolic procedure bianchy1p(u,v);
   v and (bianchy1p0(u,car v) or bianchy1p(u,cdr v));

symbolic procedure bianchy1p0(u,v);
   (genindexp car v
    and condcheck1(list car u,list car v) 
    and condcheck2(cdr u,cadr v) 
    and bianchy0 cdr u) or
   (not genindexp car v
    and condcheck1(cdddr u,cdr v) 
    and condcheck2(u := first!-three u,car v)
    and bianchy0 u); 

% Second Bianchy identity

%
%   ij          ij        ij
%  R       = - R       - R
%     kl;m        lm;k      mk;l
%

symbolic procedure bianchy2p(u,v);
   v and (bianchy2p0(u,car v) or bianchy2p(u,cdr v));

symbolic procedure bianchy2p0(u,v);
   (genindexp car v
    and condcheck1(first!-two u,first!-two v) 
    and condcheck2(cddr u,caddr v) 
    and bianchy0 cddr u) or
   (not genindexp car v
    and condcheck1(first!-two cddr u,cdr v) 
    and condcheck2(u := append(first!-two u,cddddr u),car v)
    and bianchy0 u); 

symbolic procedure condcheck1(u,v);
   null u or (condindexp(car u,car v) and condcheck1(cdr u,cdr v));
   
symbolic procedure condcheck2(u,v);
   condcheck1(u,v) and ordcheck(u,v);

symbolic procedure ordcheck(u,v);
   if (tildep car v and (tildep cadr v or tildep caddr v))
      or (tildep cadr v and tildep caddr v)
     then (v := for each j in v collect remtilde j)
          and ordp(car u,cadr u) eq ordp(car v,cadr v)
          and ordp(car u,caddr u) eq ordp(car v,caddr v)
          and ordp(cadr u,caddr u) eq ordp(cadr v,caddr v)
    else t;

symbolic procedure tildep u;
   eqcar(u,'!~) or (eqcar(u,'minus) and eqcar(cadr u,'!~));

symbolic procedure remtilde u;
   if atom u then u
    else if car u eq '!~ then cadr u
    else list(car u,remtilde cadr u);

symbolic procedure condindexp(u,v);
   tildep v or u = v or 
   (eqcar(u,'minus) and eqcar(v,'minus) and condindexp(cadr u,cadr v));

symbolic procedure bianchy0 u; car even!-odd!-perms u;

symbolic procedure bianchy(u,v);
   list('minus,list('plus,where!-list(car v,caddr v,u),
                          where!-list(cadr v,caddr v,u)));

% Contracted Bianchy identities;

symbolic procedure riem!-biancc(u,a);
   %
   %   i
   %  R       = R     - R
   %    jln;i    jn;l    jl;n
   %
   (if x = a then riem!-biancc0(y,z,w)
     else if y = a then list('minus,riem!-biancc0(x,z,w))
     else if z = a then riem!-biancc0(w,x,y)
     else list('minus,riem!-biancc0(z,x,y)))
       where x=car u,y=cadr u,z=caddr u,w=cadddr u;

symbolic procedure riem!-biancc0(x,y,z);
   list('times,cddr riemann!*,
               list('plus,list('d,list(car riemann!*,x,z),y),
                          list('minus,list('d,list(car riemann!*,x,y),z))));

symbolic procedure ricci!-biancc(u,a);
   %
   %   i       1 
   %  R      = - R
   %     n;i   2  ;n
   %
   list('quotient,list('d,car riemann!*,if car u = a then cadr u else car u),
                  2);

% Jacobi idenity

symbolic procedure jacobi u;
   %
   %     
   %  LD(A,B,C) + LD(B,C,A) + LD(C,A,B) = 0
   %
   %
   list('minus,list('plus,u,
                          list(car u,caddr u,cadddr u,cadr u),
                          list(car u,cadddr u,cadr u,caddr u)));

% Relation for the products of Levi-Civita's;

symbolic procedure lcc!-cond(u,v);
   ((levi!-civitasp u and levi!-civitasp v
     and sameposl cdr u and sameposl cdr v) or
    (levi!-civitatp u and levi!-civitatp v))
   and stdim!-signature!-exist(); 

symbolic procedure stdim!-signature!-exist;
   << stdim!-signature(stdim!*,signature!*);
      if not(stdim!* and signature!*)
        then rerror('riccir!-defs,1,
           list(if stdim!* then "The signature of the metric tensor" 
                 else "The dimension of space-time","is not defined"))
       else T>>;

symbolic procedure levi!-civita!-rel(u,v);
   %
   %         a1 ... an                       q         [a1        an]
   %     TEPS          * TEPS          = (-1)  * n! * G      ... G
   %                         b1 ... bn                    b1        bn
   %
   %         a1 ... an                          [a1        an]
   %     SEPS          * SEPS          =  n! * G      ... G
   %                         b1 ... bn             b1        bn
   %
   %  where TEPS is the Levi-Civita tensor, 
   %        SEPS is the Levi-Civita symbol
   %        and q=(n-s)/2
   %
   begin scalar a,b,n,bool;
      n := if levi!-civitasp u then 1
            else list('expt,'(minus 1),
                            list('quotient,list('plus,stdim!*,
                                                list('minus,signature!*)),
                                 2));
      a := ordn(u := cdr u);
      b := ordn(v := cdr v);
      u := tpermp(u,a);
      v := tpermp(v,b);
      if (u and not v) or (not u and v) then bool := t;
      a := even!-odd!-perms a;
      u := 'plus . for each j in car a collect metricprods(j,b);
      v := 'plus . for each j in cdr a collect metricprods(j,b);
      if bool then u := list('minus,u)
       else v := list('minus,v);
      return list('times,n,list('plus,u,v));
   end;         

symbolic procedure metricprods(u,v);
   'times . for each j in pair(u,v) collect list(metric!*,car j,cdr j);

% III. Differential relations;

symbolic procedure call!-gdiffrel(u,type);
   multmsq(call!-gdiffrel!-msf(numr u,type,
                               leading!-diffops(numr u,type)),
           revpr call!-gdiffrel!-msf(denr u,type,
                                     leading!-diffops(denr u,type)));

symbolic procedure call!-gdiffrel!-msf(u,type,dfls);
   if null dfls then u ./ !*msfone!*
     else (if msfonep denr x and numr x = u then x
            else call!-gdiffrel(x,type))
              where x=if type memq '(cov!-lie jacobi)
                        then call!-cov!-lie!-msf0(u,type,car dfls)
                       else call!-gdiffrel!-msf0(u,type,car dfls);

symbolic procedure call!-gdiffrel!-msf0(u,type,dfls);
   if null dfls then u ./ !*msfone!*
    else addmsq(car x,call!-gdiffrel!-msf0(red u,type,cdr x))
            where x=call!-gdiffrel!-mst(lt u,type,dfls);

symbolic procedure call!-cov!-lie!-msf0(u,type,dfls);
   if null dfls or null u then u ./ !*msfone!*
    else (if msfonep denr car x and null red numr car x
             and lt numr car x = lt u
            then addmsq(car x,call!-cov!-lie!-msf0(red u,type,cdr x))
           else call!-gdiffrel(addmsq(car x,red u ./ !*msfone!*),type))
             where x=call!-gdiffrel!-mst(lt u,type,dfls);

symbolic procedure call!-gdiffrel!-mst(u,type,dfls);
   begin scalar x,y,z,v,w;
      v := tpow u;
      if type neq 'cov!-lie then go to a;
      z := appinvs sdummy movable!-argindices v;
      for each j in dfls do 
         if j member v 
           then if (y := findlievec(delete(j,v),indicesinv covindices j))
                   and (atom caaar y or caaar y member z)
                   and (not (w := buildvec list cadr j) 
                        or (pairp caaar w and not(caaar w memq z))
                        or ((w := where!-list(caar y,caar w,
                                          where!-dummy(cdar y,cdar w,cdr w)))
                            neq cdr y and riccir!-ordp(w,cdr y))) 
                  then << v := delete(j,setdiff(v,cdr y)); 
                          dfls := delete(j,dfls);
                          x := (if atom caaar y then j . y
                                 else indexsubst(caaar y,indexinv caaar y,j)
                                       . ((list indexinv caaar y . cdar y) .
                                 indexsubst(indexinv caaar y,caaar y,cdr y)))
                               . x>>
                 else dfls := delete(j,dfls);
      if null x then return (u .+ nil ./ !*msfone!*) . dfls;
      x := reverse x;
      go to b;
   a: if v = 1 or null (x := xn(v,dfls))
        then return (u .+ nil ./ !*msfone!*) . dfls
       else if type eq 'jacobi then go to c;
      v := setdiff(v,x); 
      dfls := setdiff(dfls,x);
      x := for each j in x collect j . nil;
   b: y := if v then (v .* tc u) .+ nil ./ !*msfone!*
            else !*dsf2msq tc u;
      for each j in x do
         y := multmsq(call!-gdiffrel!-msv(car j,cdr j,type),y);
      return y . dfls;
   c: while x and null y do << 
        w := !*curralg!*;
        while w and null y do <<
          (if (z := buildvec(cadar x . delete(car x,v)))
              and ldmultmk2!* cdr z = ldmultmk2!* cdr zz
              and corr!-arg caddr car x = ldmultmk2!* caddr car w
              and corr!-arg cadddr car x = ldmultmk2!* cadddr car w 
             then << y := (cdar zz . car w) . 
                          (cdar z . setdiff(v,car x . cdr z)); 
                     dfls := delete(car x,dfls)>>)
                       where zz=buildvec corr!-arg cadar w;
          w := cdr w>>;
        x := cdr x>>;
      return (if y then addmsq(multmsqsq(jacobi!-rel(car y,cdr y),tc u ./ 1),
                               u .+ nil ./ !*msfone!*)
               else u .+ nil ./ !*msfone!*) . dfls
   end;

symbolic procedure ldmultmk2!* u;
   lpow multmk2(corr!-arg u,nil);

symbolic procedure corr!-arg u;
   if car u eq 'times then cdr u else list u;

symbolic procedure jacobi!-rel(u,v);
   riccirsimp!* (if cdr v then 'times . append(mk!*expt cdr v,x)
                  else x) where x=jacobi where!-dummy(car v,car u,cdr u); 

symbolic procedure call!-gdiffrel!-msv(u,v,type);
   if lastreltype(u,v,type) then call!-gdiffrel!-msv0(u,v,type)
    else begin scalar x,ttype;
            if (ttype := car u) neq 'ld then
              diffindices!* := append(upping cddr u,diffindices!*);
            x := call!-gdiffrel!-msv(cadr u,v,type);
            for each j in cddr u do 
               if ttype neq 'ld
                 then << diffindices!* := cdr diffindices!*;
                         x := gendiffmsq(x,j,ttype)>>
                else gendiffmsq(x,lieargeval j,ttype);
            return x
         end;

symbolic procedure call!-gdiffrel!-msv0(u,v,type);
   (if type eq 'cov!-part then cov!-part(arg,ind,'partdiff)
     else if type eq 'part!-cov then cov!-part(arg,ind,'d)
     else if type eq 'ordercov then order!-covd u
     else lie!-cov(arg,ind,v)) where arg=cadr u,ind=cddr u;

symbolic procedure leading!-diffops(u,type);
   if null u or lpow u = 1 then nil
    else rally!-diffops(leading!-diffops0(lpow u,type),
                        leading!-diffops(red u,type));
                      
symbolic procedure leading!-diffops0(u,type);
   if null u or numberp cdar u then nil
    else rally!-diffops(find!-dtype(car u,car u,type,0),
                        leading!-diffops0(cdr u,type));

symbolic procedure find!-dtype(u,v,type,n);
   if diffopp!* u
     then if reltype(u,type) 
            then find!-dtype(cadr u,v,type,n + length cddr u)
           else find!-dtype(cadr u,v,type,n)
    else if n = 0 then nil
    else list v . n;

symbolic procedure rally!-diffops(u,v);
   if null u then v
    else if null v then u
    else if cdr u > cdr v then u
    else if cdr u = cdr v then append(car u,car v) . cdr u
    else v;

symbolic procedure reltype(u,type);
   if type eq 'part!-cov then car u eq 'partdiff
    else if type memq '(cov!-part cov!-lie) then car u eq 'd
    else if type eq 'lie!-cov then car u eq 'ld
    else if type eq 'ordercov 
     then cdddr u and ((!*curralg!* and car u memq '(d partdiff))
                       or (car u eq 'd and not cov2ordp cddr u))
    else if type eq 'jacobi 
     then car u eq 'ld and cdddr u and not atom cadr u
    else nil;

symbolic procedure cov2ordp u; ordn x = x where x=upping u;

symbolic procedure lastreltype(u,v,type);
   (type eq 'cov!-lie and car u eq 'd and indexinv caaar v member cddr u)
    or (reltype(u,type) and not find!-dtype(cadr u,nil,type,0));

symbolic procedure covindices u;
   if diffopp!* u
     then if car u eq 'd then append(covindices cadr u,cddr u)
           else covindices cadr u
    else nil;
    
symbolic procedure findlievec(u,v);
   if null u or numberp cdar u or null v then nil
    else if car v member indices car u
     then buildvec u or findlievec(cdr u,cdr v)
    else findlievec(cdr u,v);    

% Relation between covariant and partial differentials;

symbolic procedure cov!-part(u,v,type);
   %
   %                A    = A
   %                 ; I    , I
   %
   %     I1 ... Ik           I1 ... Ik
   %    A                = A
   %        J1 ... Jl; I        J1 ... Jk , I
   %
   %                         I1 ..c.. Ik       Ii
   %                     + A              * GAM
   %                            J1 ... Jl       I c
   %
   %                         I1 ... Ik           c
   %                     - A                * GAM
   %                            J1 ..c.. Jl      Ji I
   %
   % U : argumentum
   % V : 'variables'
   % type : the name of the diff.op. after using the relation
   %        partdiff => d
   %               d => partdiff
   begin scalar x,y,z; integer n;
      v := reverse v;
      n := if type eq 'd then  -1 else 1;
      if cdr v then u := (if type eq 'd then 'partdiff else 'd) .
                            (u . reverse cdr v);
      v := car v;
      z := sfree allindices u;
      y := downs z;
      for each j in setdiff(z,y) do x := 
         list('times,indexsubst('!&000,j,u),
                     list(christoffel!*,j,'(minus !&000),v),
                     n) . x;
      for each j in y do x :=
         list('times,indexsubst('(minus !&000),j,u),
                     list(christoffel!*,'!&000,j,v),
                     -n) . x;
      x := 'plus . mkgendiff(u,list v,type) . 
               if atom v then for each j in x collect
                 list('times,cadr j,indexsubst('(minus !&001),v,caddr j),
                             list(metric!*,'!&001,v),cadddr j)
                else x;
      return riccirsimp!* x
   end;

% Relation between Lie and covariant derivatives;

symbolic procedure lie!-cov(u,ind,v);
   %
   %                                 c
   %               LD(A,X) = A    * X
   %                          ; c
   %
   %       I1 ... Ik          I1 ... Ik          c
   %   LD(A            ,X) = A                * X
   %          J1 ... Jl          J1 ... Jl; c
   %
   %                          I1 ..c.. Ik     Ii
   %                       - A             * X
   %                             J1 ... Jl      ; c
   %
   %                          I1 ... Ik         c
   %                       + A               * X
   %                             J1 ..c.. Jl     ; Ji
   begin scalar x,y,z,w;
      if null v 
        then << v := reverse ind; 
                ind := nil;
                if cdr v then u := 'ld . (u . reverse cdr v);
                z := sfree allindices u;
                y := buildvec!* car v;
                w := caaar y;
                v := mk!-dummy0(car v,cdar y,union(z,caar y))>>
       else << w := caaar v;
               if indexinv w = car ind then ind := list cdr ind
                else << u := mkgendiff(u,x := remgd(indexinv w,ind),'d);
                        ind := list cdr setdiff(ind,x)>>;
               z := sfree allindices u;
               v := 'times . cdr v>>;
      y := downs z;
      for each j in setdiff(z,y) do x := 
         list('times,indexsubst('!&000,j,u),
                     mkgendiff(indexsubst(j,w,v),list '(minus !&000),'d),
                     -1) . x;
      for each j in y do x :=
         list('times,indexsubst('(minus !&000),j,u),
                     mkgendiff(indexsubst('!&000,w,v),list j,'d)) . x;
      x := if ind 
             then if x then list('plus,mkgendiff(u,list v,'ld),
                                       list('minus,'plus . x))
                   else mkgendiff(u,list v,'ld)
            else 'plus . list('times,mkgendiff(u,list '(minus !&000),'d),
                                 indexsubst('!&000,w,v)) . x;
      return riccirsimp!* if ind and car ind then mkgendiff(x,car ind,'d)
                           else x
   end;

symbolic procedure buildvec!* u;
   buildvec if car u eq 'times then cdr u else list u;

symbolic procedure remgd(u,v);
   if u = car v then nil else car v . remgd(u,cdr v);

% Using the definition of the Riemann tensor;

symbolic procedure order!-covd u;
   if !*curralg!* then order!-covd1 u
    else order!-covd2 u;

symbolic procedure order!-covd1 u;
    (if x member !*curralg!*
       then if car u eq 'd
              then riccirsimp!* (if cddddr u 
                                   then mkgendiff(y,cddddr u,car u)
                                  else y) where y=order!-covd0 x
             else if christoffel!*
              then call!-gdiffrel(!*msk2msq list 
                                    mkgendiff(cadr u,
                                              cadddr u . caddr u . cddddr u,
                                              car u),
                                 'part!-cov)
             else !*msk2msq list mkgendiff(cadr u,cddr u,car u)
      else if null cddddr u then !*msk2msq list mkgendiff(cadr u,cddr u,car u)
      else order!-covd1(car u . mkgendiff(cadr u,list caddr u,car u) . cdddr u))
          where x=car u . cadr u . first!-two cddr u;

symbolic procedure order!-covd2 u;
   if null cdddr u then mkmsq mkgendiff(cadr u,cddr u,car u)
    else if ordp(indexabs caddr u,indexabs cadddr u)
     then order!-covd2(car u . mkgendiff(cadr u,list caddr u,car u) . cdddr u)
    else riccirsimp!* (if cddddr u then mkgendiff(x,cddddr u,car u)
                         else x) where x=order!-covd0 u;

symbolic procedure order!-covd0 u;
   'plus . mkgendiff(cadr u,reverse x,car u) . riemanndef2(cadr u,x)
        where x=list(caddr u,cadddr u);

symbolic procedure riemanndef2(u,v);
   %
   %                       c
   %      A     - A     = R      * A
   %       I;JK    I;KJ     I JK    c
   %
   %       I       I         I        c
   %      A     - A     = - R      * A
   %        ;JK     ;KJ       c JK    
   %
   %   I1 ... Ik          I1 ... Ik            Ii        I1 ..c.. Ik
   %  A                - A                = - R       * A           
   %      J1 ... Jl;JK       J1 ... Jl;KJ        c JK       J1 ... Jl
   %
   %                                           c         I1 ... Ik   
   %                                        + R       * A      
   %                                            Ji JK       J1 ..c.. Jl 
   begin scalar x,y,z,w; integer n;
      n := cadr riemann!*;
      z := sfree allindices u;
      y := downs z;
      w := car riemann!*;
      for each j in setdiff(z,y) do x :=
         list('times,indexsubst('!&000,j,u),
                     list(w,j,'(minus !&000),car v,cadr v),
                     -n) . x;
      for each j in y do x :=
         list('times,indexsubst('(minus !&000),j,u),
                     list(w,j,'!&000,car v,cadr v),
                     n) . x;
      return x
   end;

endmodule;

end;
