%*********************************************************************;
%*****             General differential functions                 ****;
%*********************************************************************;

module riccir!-gdiff;

% Author: Jozsef Kadlecsik

global '(!*msfone!* !*msqnull!* commutelist!* diffindices!* 
         indexlist!* metric!*);

fluid '(liedind!*);

%*********************************************************************;
%*****                                                           *****;
%*****   General differential rules:                             *****;
%*****                                                           *****;
%*****   GD(A + B,C) = GD(A,C) + GD(B,C)           (Linearity)   *****;         
%*****                                                           *****;
%*****   GD(A * B,C) = GD(A,C) * B + A * GD(B,C)   (Leibnitz-rule) ***;
%*****                                                           *****;
%*********************************************************************;
%*****                                                           *****;
%*****   Special rules for Lie-derivatives:                      *****;
%*****                                                           *****;
%*****       c                    c                              *****;
%*****   LD(X ,Y)         = - LD(Y ,X)                           *****;
%*****                                                           *****;
%*****       ...                ...           ...                *****;
%*****   LD(A     ,X + Y) = LD(A    ,X) + LD(A    ,Y)            *****;
%*****        ...                ...           ...               *****;
%*****                                                           *****;
%*****        ...                ...                             *****;
%*****   LD(A     ,a * X) = LD(A    ,X) * a                      *****;
%*****         ...                ...                            *****;
%*****                                                           *****;
%*****                ..c..    ic          ...      c            *****;
%*****             - A      * X   * a   + A      * X  * a        *****;
%*****                 ...           ;c     ..c..        ;jc     *****;
%*****                                                           *****;
%*********************************************************************;

symbolic procedure mkdiffop u;
   for each j in u 
     do << put(j,'simpfn,'simpgendiff);
           put(j,'riccirfn,'gendiff);
           put(j,'rtype,'indexed!-quantity);
           flag(list j,'diffop);
           flag(list j,'full);
           put(j,'prifn,'gendiffprt);
           % Next line added by Herbert Melenk
           put(j,'fancy!-pprifn,'fancy!-print!-riccirdf)>>;

mkdiffop '(partdiff d ld);

flag('(partdiff),'young!-identity);
flag('(ld),'lie!-type);
put('ld,'prifn,'liedprt);
% Next two lines added by Herbert Melenk
put('ld,'fancy!-pprifn,'fancy!-riccirld);
put('ld,'fancy!-functionsymbol,'(ascii 163));

commutelist!* := '((d metric) (ld partdiff metric));

symbolic procedure commutep(u,v);
   (x and v memq cdr x) where x=assoc(u,commutelist!*);

symbolic procedure simpgendiff u;
   !*msq2sq gendiff u;

symbolic procedure gendiff u;
 if null subfg!* then mkmsq gendiffrevindl u
  else gendiff0 u;

symbolic procedure gendiffrevindl u;
   if atom u then u
    else if tensorp u or symbolp u then car u . revalindexlist cdr u
    else if diffopp u and not flagp(car u,'lie!-type)
     then car u . gendiffrevindl cadr u . revalindexlist cddr u
    else car u . for each j in cdr u collect gendiffrevindl j;

symbolic procedure gendiff0 u;
   begin scalar v,type;
      type := car u;
      if flagp(type,'lie!-type) then << 
        v := if type eq 'ld then lieindexlist cddr u
              else revalindexlist cddr u;
        u := riccirsimp cadr u;
        % the args of 'ld are as follows: 
        % expression,vec1,vec2,...
        for each j in v do u := gendiffmsq(u,j,type);
        return u>>;
      v := revalindexlist cddr u;
      if ups v and null metric!*
        then rederr list("Metric tensor is not defined:",v);
      diffindices!* := append(upping v,diffindices!*);
      % We have to use DIFFINDICES!* to store the args of the
      % differential operator. So we can avoid that these indices
      % play a role as dummy indices in the expression from the
      % first arg. of the differential operator.
      u := riccirsimp cadr u;
      for each j in v do << 
       diffindices!* := cdr diffindices!*;
       u := gendiffmsq(u,j,type)>>;
      return u
   end;

symbolic procedure lieindexlist u;
   if null u then nil
    else lieargeval car u . lieindexlist cdr u;

symbolic procedure lieargeval u;
   begin scalar y,z;
      y := riccirsimp u;
      if null numr y 
        then rederr list("Wrong Lie-argumentum:",u,"has zero value")
       else if null car(z := free!-and!-dummy2 msfindices numr y) 
               or cdar z or downp caar z
        then rederr list("Wrong Lie-argumentum:",u,"is not vector");
      return if caar z eq car indexlist!* then y
              else << u := choose!-dummy(cdr z,cdr indexlist!*,nil);
                      where!-list(car indexlist!* . appinvs u,
                                  caar z . appinvs cdr z,
                                  numr y) ./ denr y>>
   end;
     
symbolic procedure gendiffmsq(u,v,type);
   % U is a modified standard quotient, V a kernel.
   % Value is the modified standard quotient derivative of U wrt V.
   % Algorithm: df(x/y,z)= (x'-(x/y)*y')/y;
   multmsq(addmsq(gendiffmsf(numr u,v,type),
                  negmsq multmsq(u,gendiffmsf(denr u,v,type))),
           !*msfone!* ./ denr u);

symbolic procedure gendiffmsf(u,v,type);
   if null u or lpow u = 1 then !*msqnull!*
    else addmsq(multmsqsq(gendiffk(lpow u,v,type),lc u ./ 1),
                gendiffmsf(red u,v,type));

symbolic procedure gendiffk(u,v,type);
   if type eq 'ld 
     then begin scalar liedind!*;
             liedind!* := sdummy movable!-argindices u;
             return lddiffk(u,v)
          end
    else if cdr u then difftimes(u,v,type)
    else gendiffk!*(car u,v,type);

symbolic procedure gendiffk!*(u,v,type);
   if atom u then one!-level!-down(u,list v,type)
    else if kroneckerp u
            or (type eq 'd and (metricp u or levi!-civitatp u))
            or (type eq 'partdiff and levi!-civitasp u)
           then !*msqnull!*
    else if numberp cdr u 
           then if cdr u = 1 then gendiffk!*(car u,v,type)
                 else gendiffp0(u,v,type)
    else begin scalar x,y,z;
            if diffopp u and commutep(car u,type)
              then << x := gendiffk(list if atom cadr u then cadr u . 1 
                                          else cadr u,v,type);
                      % DIFFINDICES!* is OK!!
                      for each j in cddr u do 
                         x := gendiffmsq(x,j,car u);
                      return x>>;
            if null(x := get(car u,'dfn)) 
              then return one!-level!-down(u,list v,type); 
            z := cdr u;
            y := !*msqnull!*;
            for each j in
               for each k in z collect gendiff list(type,k,v) do
             <<if j then
                y := addmsq(multmsq(j,riccirsimp!*
                                       subla(pair(caar x,z),cdar x)),
                        y);
               x := cdr x>>;
            return y
         end;

symbolic procedure gendiffp0(u,v,type);
   begin scalar pv,n,z;
      pv := car u;
      n := pdeg u;
      return if (z := gendiffk!*(pv,v,type))
               then if n = 1 then z
                     else multmsq(z,list(pv to (n - 1)) .* n 
                                    .+ nil ./ !*msfone!*)
              else !*msqnull!*
  end;

symbolic procedure difftimes(u,v,type);
   % U is a times argument, result is a msq.
   if null cdr u then gendiffk!*(car u,v,type)
    else addmsq(multmsq(!*msk2msq list car u,difftimes(cdr u,v,type)),
                multmsq(gendiffk!*(car u,v,type),!*msk2msq cdr u));

symbolic procedure one!-level!-down(u,v,type);
   if type eq 'partdiff and atom car v and null cdr v
     then multmsq(!*msk2msq list tkern list(metric!*,car v,'!&009),
                  one!-level!-down(u,list '(minus !&009),type))
    else if eqcar(u,type)
     then << v := append(cddr u,v);
             u := cadr u;
             % Torsion-free;
             if type eq 'd and atom u and downp car v and downp cadr v
               then v := append(ordn first!-two v,cddr v);
             recursive!-let(u,v,type)>>
    else recursive!-let(u,v,type);

symbolic procedure recursive!-let(u,v,type);
   begin scalar x,y;
      for each j in v do
         if null x then << y := j . y;
                           x := kvalue!-opmtch(type . 
                                               (u . reverse y))>>;
      y := setdiff(v,reverse y);
      return if x 
               then if y then gendiffmsq(car x,y,type)
                     else car x
              else if length v = 1 then mkmsq mkgendiff(u,v,type)
              else recursive!-let(mkgendiff(u,list car v,type),
                                  cdr v,type)
   end;  

symbolic procedure mkgendiff(u,v,type);
   type . if pairp u and car u eq type then cadr u . append(cddr u,v)
            else u . v;

symbolic procedure lddiffk(u,v);
   if msfonep denr v
     then if null red numr v
            then if lc numr v = 1 
                   then ((if y
                            then lddiffk!-msq(u,!*msk2msq cdr x,
                                              !*msk2msq y,sdummy argindices y)
                           else lddiffk0(u,x,nil))
                                 where y=setdiff(lpow numr v,cdr x)) 
                             where x=buildvec lpow numr v
                  else lddiffk!-msq(u,!*msk2msq lpow numr v,
                                    !*dsf2msq lc numr v,nil)
           else addmsq(lddiffk(u,lt numr v .+ nil ./ !*msfone!*),
                       lddiffk(u,red numr v ./ !*msfone!*))
    else lddiffk!-msq(u,numr v ./ !*msfone!*,!*msfone!* ./ denr v,
                      sdummy msfindices denr v);

symbolic procedure buildvec u;
   buildvec0(free!-and!-dummy2 indices car u . list car u)
   or (cdr u and buildvec0 buildmonom(car u,cdr u));

symbolic procedure buildvec0 u;
   if caar u and null cdaar u then u else nil;

symbolic procedure ldveceq(u,v);
   length cdar u = length cdar v
   and (atom caaar u or indexabs caaar u memq liedind!*)
   and cdr u = where!-list(caar u,caar v,where!-dummy(cdar u,cdar v,cdr v));

symbolic procedure lddiffk0(u,v,u0);
   begin scalar x,y,z,w;
      if null (x := setdiff(u,scalpart u))
         or null (x := buildvec x) then return lddiffk1(u,v,u0);
      z := x; w := cdr x;
   a: if ldveceq(x,v) 
        then return if (y := setdiff(u,cdr x))
                      then lddiffk0(y,v,append(cdr x,u0))
                     else !*msqnull!*;
   b: if null(w := cdr w) then go to c
       else if (x := buildvec w) then go to a
       else go to b;
   c: return if (atom caaar z or indexabs caaar z memq liedind!*)
                and riccir!-ordp(cdr v,where!-list(caar v,caar z,cdr z))
               then changed!-lie(setdiff(u,cdr z),v,z,u0)
              else if (y := setdiff(u,cdr z))
                     then addmsq(lddiffk0(y,v,append(cdr z,u0)),
                                 lddiffk1(cdr z,v,append(u0,y)))
                    else lddiffk1(cdr z,v,u0)
   end;

symbolic procedure changed!-lie(u,v,x,u0);
   if atom caaar x 
     then if u then addmsq(lddiffk0(u,v,append(u0,cdr x)),
                           negmsq changed!-lie0(v,x,append(u0,u)))
           else negmsq changed!-lie0(v,x,u0)
    else changed!-lie(where!-list(caar x,upping caar x,u),v,
                      (upping caar x . cdar x) . 
                         where!-list(upping caar x,caar x,cdr x),
                      where!-list(caar x,upping caar x,u0));

symbolic procedure changed!-lie0(u,v,u0);
   lddiffk1(where!-list(caar v,caar u,
                        mk!-dummy0(cdr u,cdar u,append(caar u,caar v))),
            (caar u . cdar v) . where!-list(caar u,caar v,cdr v),u0);

symbolic procedure lddiffk1(u,v,u0);
   if null cdr u 
     then (if u0 then multmsq(ldmultmk2(u0,nil),x)
            else x) where x=lddiffmsk(car u,cdr v)
    else addmsq(lddiffk0(cdr u,v,car u . u0),
                multmsq(lddiffmsk(car u,cdr v),ldmultmk2(cdr u,u0)));

symbolic procedure ldmultmk2(u,v);
   multmk2(mkscal!-last(u,nil),mkscal!-last(v,nil)) ./ !*msfone!*;

symbolic procedure mkscal!-last(u,v);
   if null u then v
    else if numberp cdar u then mkscal!-last(cdr u,car u . v)
    else car u . mkscal!-last(cdr u,v);

symbolic procedure lddiffmsk(u,v);
   begin scalar x; integer n;
      v := if cdr v then 'times . v else car v;
      return if numberp cdr u
               then << n := cdr u;
                       u := car u;
                       x := gendiffk!*(u,v,'ld);
                       if n = 1 then x
                        else multmsq(x,list(u to (n - 1)) .* n
                                       .+ nil ./ !*msfone!*)>>
              else one!-level!-down(u,list v,'ld)
   end;
   
symbolic procedure lddiffk!-msq(u,v,w,a);
   % U : arg of LD;
   % V : vector valued msq;
   % W : mult. factor to V;
   % A : indices of W; 
   (addmsq(multmsq(lddiffk(u,v),w),
           lddiffk!-addterm(u,v,cdr z,car z,car sfree x,y))
              where z=lddiffk!-dterm(w,x,y,a))
         where y=argindices u,x=msfindices numr v;

symbolic procedure lddiffk!-dterm(u,v,w,a);
   begin 
      v := union(union(upping v,upping w),a);
      w := diffindices!*;
      diffindices!* := union(diffindices!*,v);
      a := car choose!-dummy(list 1,setdiff(indexlist!*,v),nil);
      u := gendiffmsq(u,indexinv a,'d);
      diffindices!* := w;
      return a . u
   end;  

symbolic procedure lddiffk!-addterm(u,v,w,a,b,z);
   % U : arg of LD;
   % V : vector valued msq;
   % W : cov. deriv. of the mult. factor with respect A;
   % A : dummy index;
   % B : free index of V;
   % Z : indices of U;
   begin scalar x,y;
      if null numr w or null z then return !*msqnull!*;
      u := prepsq !*msq2sq !*msk2msq u;
      v := prepsq !*msq2sq v;
      w := prepsq !*msq2sq w;
      y := downs z;
      for each j in setdiff(z,y) do x :=
         list('times,indexsubst(a,j,u),indexsubst(j,b,v),w,-1) . x;
      for each j in y do x :=
         list('times,indexsubst(indexinv a,j,u),indexsubst(a,b,v),
                     indexsubst(j,indexinv a,w)) . x;
      return riccirsimp!*('plus . x)
   end;
                                          
endmodule;

end;
