%********************************************************************;
%********************************************************************;
%                       Ricci Calculus Package                       ;
%********************************************************************;
% This version runs in REDUCE 3.5                                    ;
%********************************************************************;
% Version: 1.01                                                      ;
% Date   : 03.03.94                                                  ;
%********************************************************************;
% Author : Jozsef Kadlecsik                                          ;
%                                                                    ;
% Central Research Institute for Physics                             ;
% H-1525 Budapest 114, P.O.Box 49, Hungary                           ;
%                                                                    ;
% Phone  : +36 (1) 169 9499                                          ;
% Fax    : +36 (1) 155 3376                                          ;
% E-mail : kadlec@sunserv.kfki.hu | kadlec@blackhole.kfki.hu         ;
%********************************************************************;
%********************************************************************;

module riccir!-intro;

% Author: Jozsef Kadlecsik

%********************************************************************;
%*****               General signs and modifications             ****;
%********************************************************************;
 
global '(christoffelsign!* indexnam!* levi!-civitamods!*
         levi!-civitasign!* liesign!* partdiffsign!*
         version!*);


% Version-dependent variables

symbolic procedure version!-dependent;
   if version!* = "REDUCE 3.4"
     then << % Reduce 3.4
             global '(frlis!* subfg!*)>>
    else if version!* = "REDUCE 3.5"
     then << % Reduce 3.5
             fluid '(frlis!*);
             global '(subfg!*)>>
    else if version!* = "REDUCE 3.6"
     then << % Reduce 3.6
             fluid '(frlis!* subfg!*)>>
    else <<  % Other versions: try 3.6...
             fluid '(frlis!* subfg!*)>>;

%*************** Sign used in input-output of RICCIR **************;

partdiffsign!*     := '!@;   % Partial differential operator;

%********* Signs used in output of RICCIR in case of ON NAT ********;
%**** If there are no available characters for them, then they may be 
%**** equal to nil;

christoffelsign!*  := nil;   % Christoffel symbol (capital gamma);
levi!-civitasign!* := nil;   % Levi-Civita symbol (epsilon);
levi!-civitamods!* := nil;   % modification for Levi-Civita tensor (hat);
liesign!*          := nil;   % Lie differential operator (pound sign);

indexnam!* := 'i;   % The names of the dummy indices introduced
                    % by the system consist of INDEXNAM*
                    % and a serial number;

%**** PLUS and DIFFERENCE are changed because of the non-homogenous terms;

deflist('
  ((difference getrtypeor)
   (plus getrtypeor)
   ),'rtypefn);

%**** The algebraic functions and operators are changed in order that 
%**** they could be differentiated with tensor expressions in their args.;

deflist('((acos getrtypecar) (asin getrtypecar) (atan getrtypecar)
          (acosh getrtypecar) (asinh getrtypecar) (atanh getrtypecar)
          (acot getrtypecar) (cos getrtypecar) (sin getrtypecar) 
          (tan getrtypecar) (sec getrtypecar) (sech getrtypecar)
          (csc getrtypecar) (csch getrtypecar) (cot getrtypecar) 
          (acot getrtypecar) (coth getrtypecar) (acoth getrtypecar)
          (cosh getrtypecar) (sinh getrtypecar) (tanh getrtypecar)
          (abs getrtypecar) (sqrt getrtypecar) (log getrtypecar)
   ),'rtypefn);

symbolic procedure mkop u;
   begin scalar x;
        if null u then typerr("Local variable","operator")  
         else if (x := gettype u) eq 'operator
          then lprim list(u,"already defined as operator")
         else if x and not(x memq '(fluid global procedure))
          then typerr(u,'operator)
         else if u memq frlis!* then typerr(u,"free variable")
         else << put(u,'simpfn,'simpiden);
                 put(u,'rtypefn,'getrtypeor)>>;  %*** modification ***% 
   end;

%**** We use NEWTOKCALL and REMNEWTOK to introduce and remove new
%**** tokens. Their arguments are:
%**** U := nil (no sign) | token | list of tokens: (!: !=);
%**** V := the name of the token;

symbolic procedure newtokcall(u,v);
   if u then newtok list(newtokarg u,v);

symbolic procedure remnewtok(u,v);
   if u then begin scalar x;
                u := list(newtokarg u,v);
                remprop(caar u,'switch!*);
                remprop(cadr u,'prtch);
                if x := get(cadr u,'unary) then remprop(x,'prtch)
             end;

symbolic procedure newtokarg u;
   if atom u then list u
    else if listp u then u
    else rerror('riccir!-intro,1,
     "Wrong new token definition (variable) in the source code");

symbolic procedure mkriccirsigns();
   begin
      newtok list(newtokarg partdiffsign!*,'partdiff);
      newtokcall(liesign!*,'ld);  % maybe there is no spec. sign;
   end;

mkriccirsigns();

endmodule;

end;
