%********************************************************************;
%*****               Special LET function for tensors            ****;
%********************************************************************;

module riccir!-letfn;

% Author: Jozsef Kadlecsik

global '(!*msqnull!* christoffel!* imatch!* indxlist!* ipowlis!* 
         kronecker!* levi!-civitas!* levi!-civitat!* metric!* 
         riemann!* symbollist!* tensorlist!* weyl!* 
         erfg!* mcond!*);

fluid '(!*let!*
        !*sub2 frasc!* newrule!* oldrules!* wtl!*);

version!-dependent();

put('indexed!-quantity,'typeletfn,'tensorletfn);

flag('(tensorletfn),'direct);

symbolic procedure tensorletfn(u,v,ltype,b,rtype);
   begin scalar x,y,z,flg;
      if ltype eq rtype then go to a
       else if null rtype
              then if v = 0 
                      or (eqcar(v,'!*sq) and null numr cadr u)
                      or (null v and null b)
                     then go to a
                    else typerr(list(ltype,u),'scalar)
       else if null ltype
              then if not atom u or not indexedp list u 
                     then typerr(u,rtype)
             else put(u,'rtype,rtype)
       else typerr(ltype,rtype);
   a: x := subla(frasc!*,u);
      if x neq u
        then if atom x then return errpri1 u
               else <<flg := t; u := x>>;
      x := subla(frasc!*,v);
      if x neq v then v := x;
      % Check for unmatched free variables.
      x := smemql(frlis!*,mcond!*);
      y := smemql(frlis!*,u);
      if (z := setdiff(x,y))
         or (z := setdiff(setdiff(smemql(frlis!*,v),x),
                  setdiff(y,x)))
        then <<lprie ("Unmatched free variable(s)" . z);
               erfg!* := 'hold;
               return nil>>;
      if atom u and null b 
        then <<remprop(u,'avalue);
               remprop(u,'rtype);    % just in case
               remprop(u,'rtypefn);
               remprop(u,'kvalue);
               remprop(u,'opmtch);
               remprop(u,'simpfn);
               remprop(u,'idvalfn);
               remprop(u,'symmetry);
               remprop(u,'riccirfn);
               remprop(u,'prifn);
               tensorlist!* := delete(u,tensorlist!*);
               symbollist!* := delete(u,symbollist!*);
               refresh!-alliqlist();
               remflag(list u,'full);
               wtl!* := delasc(u,wtl!*);
               if u eq metric!* then clear!-metric u
                else if u eq kronecker!* then clear!-kronecker u
                else if u eq christoffel!* then clear!-christoffel u
                else if eqcar(levi!-civitat!*,u)
                 then clear!-levi!-civita list 'tensor
                else if eqcar(levi!-civitas!*,u)
                 then clear!-levi!-civita list 'symbol
                else if eqcar(riemann!*,u) then clear!-riemann u
                else if u eq weyl!* then clear!-weyl u;
                rmsubs(); % since all kernel lists are gone.
               return nil>>;
      x := riccirsimp0 u;
      return if not domainp numr x 
               then tensorletexprn(u,v,x,b,flg)
              else errpri1 u
   end;

symbolic procedure riccirsimp0 u;
   begin scalar !*factor,!*let!*,x,y;
      y := setkorder frlis!*;
      x := subfg!* . !*sub2;
      subfg!* := nil;
      !*let!* := t;  % Because of the special MSQ => SQ conversation;
      u := !*msq2sq riccirsimp!* u;
      subfg!* := car x;
      !*sub2 := cdr x;
      setkorder y;
      return u
   end;

symbolic procedure tensorletexprn(u,v,x,b,flg);
   % replacement of tensor expressions;
   begin scalar y,z,flgg;
      if denr x neq 1
        then return let2(let!-prepf numr x,
                         list('times,let!-prepf denr x,v),nil,b)
       else if red(x := numr x)
        then return let2(let!-prepf !*t2f lt x,
                         list('difference,v,let!-prepf red x),nil,b)
       else if null (y := kernlp x)
        then <<y := term!-split x;
               return let2(let!-prepf car y,
                           list('difference,v,let!-prepf cdr y),nil,b)>>
       else if y neq 1
        then return let2(let!-prepf quotf!*(x,y),
                         list('quotient,v,let!-prepf y),nil,b);
      if car (z := find!-riccir!-cond(x,v,b))
        then << x := caar z; v := cdar z; flgg := flg := t>>
       else v := cdr z;
      x := mkletarg klistt x;
      tensorletexprn!-do(x,u,v,b,flg);
      if flgg then frasc!* := mcond!* := nil
   end;   

symbolic procedure mkletarg u;
   if eqcar(caar u,'!*times) then cdaar u
    else list caar u;   

symbolic procedure tensorletexprn!-do(x,u,v,b,flg);
   begin scalar y,z;
      y := list(nil . (if mcond!* then mcond!* else t),v,nil);
      if null cdr x 
        then << x := car x;
                if eqcar(x,'expt) and numberp caddr x
                  then << rmsubs();
                          x := cadr x . caddr x;
                          ipowlis!* := riccir!-xadd!*(frasc!*,x . y,
                                                      ipowlis!*,b)>>
                 else if null flg
                        then << if atom x
                                  then if flagp(x,'used!*) then rmsubs()
                                        else nil
                                 else if 'used!* memq cddr fkern x
                                        then rmsubs();
                                tensorsetk1(x,v,b)>>
                       else if atom x then return errpri1 u
                       else << rmsubs();
                               put(car x,'opmtch,riccir!-xadd!*(frasc!*,
                                   cdr x . y,get(car x,'opmtch),b))>>>>
       else << rmsubs();
               imatch!* := riccir!-xadd!*(frasc!*,x . y,imatch!*,b)>>
   end;

symbolic procedure riccir!-xadd!*(frasc,u,v,b);
   % Adds replacement U to table V, with new rule at head.
   % Also checks boolean part for equality.
   begin scalar x;
      x := v;
      while x and not(car u=caar x and cadr u=cadar x) do x := cdr x;
      if x then <<v := delete(car x,v);
		 if b then riccir!-updoldrules(frasc,caddar x,cdadar x)>>;
      if b then v := u . v;
      return v
   end;

symbolic procedure riccir!-updoldrules(frasc,v,w);
   (if null u then nil
     else oldrules!* := append(
		 (if not atom v and numberp cdr v   % asymptotic case.
		    then list list(list('expt,car v,cdr v),0,nil,t)
		   else if atom car u
		    then list list(car u . car v,cadr v,nil,t)
		   else if x 
                          then (if car u neq y
			          then list list(car u,y,x,rsubla(x,w))
			         else nil) where y=rsubla(x,v)
                   % next step: in case of rules without notated free vars;
                   else (if z neq y
                           then list list(z,y,frasc,rsubla(frasc,w))
                          else nil) where z=rsubla(frasc,car u),
                                          y=rsubla(frasc,v)),
		 oldrules!*)
	   where x=caddr u)
    where u=newrule!*;

symbolic procedure find!-riccir!-cond(u,v,b);
   begin scalar x,y,z,w,r,s;
      << x := free!-and!-dummy2 movable!-msfindices uu;
         y := free!-and!-dummy2 msfindices uu;    
         z := free!-and!-dummy2 msfindices numr vv;
         w := xn(car x,
                 setdiff(car z,sfree movable!-msfindices numr vv));
         if symm!-setdiff(car z,car y) and numr vv
           then rerror('riccir!-letfn,1,
                  "The two sides are incompatible in the free indices");
         % All tensor expressions are checked. If a condition in general
         % sense is explicite written for a tensor expression  
         % (FOR ALL ... LET, FOR ALL ... SUCH THAT ... LET or RULE)
         % then there is no very much to do:
         %
         % If there are dummy indices on the right hand side
         % then hidden system indices are used instead of them:
         if (z := union(cdr z,sdummy msfindices denr vv)) 
           then << r := system!-dummy(z,indxlist!*);
                   v := where!-list!*(appinvs r,appinvs z,v)>>>>
           where uu=!*sf2msf (u := rsubla(frasc!*,u)), % Left hand side;
                 vv=msqsimp0 rsubla(frasc!*,v);        % Right hand side;
      % Condition is given by the user:
      if frasc!* 
        then << mcond!* := subla(list(0 . '(quote !0)),
                                 mcond!*);   % 0 => !0;
                return nil . subla(frasc!*,v)>>
      % There is no index on the left hand side;
       else if null(car y or cdr y) then return nil . subla(frasc!*,v); 
      if w then << % This is a relation with tensor on the left and 
                   % with symbol on the right hand side;
                   if null metric!* then << 
                     rederr "Metric tensor has not been introduced.";
                     rederr "Metric tensor is needed in the relation.">>;
                   r := system!-dummy(append(z,w),indxlist!*);
                   while z do << z := cdr z; r := cdr r>>;
                   s := for each j in pair(w,r) collect
                           (car j . if atom car j then list('minus,cdr j) 
                                     else cdr j);
                   r := for each j in s collect 
                           list(metric!*,indexabs car j,cdr j);
                   v := 'times .
                          where!-list!*(indicesinv for each j in s 
                                                      collect cdr j,
                                        w,v) . r>>;
      u := where!-list(upping car x,car x,u); 
      v := where!-list!*(upping car x,car x,v);
      % z => not movable dummy indices. cdr x => movable dummy indices;
      z := setdiff(cdr y,cdr x); 
      % r => not movable free indices;
      r := setdiff(car y,car x); 
      % s => movable and covariant not movable free indices;
      s := upping union(car x,setdiff(r,ups r));
      % r => contravariant not movable free indices;
      r := ups r;
      if s or r or z or cdr x 
        then set!-frasc!*!-mcond!*(s,r,append(z,cdr x));
      % where!-list instead of subla makes possible indices and objects
      % to have the same identifier;
      return where!-list!*(appinvs for each j in frasc!* collect cdr j,
                           appinvs for each j in frasc!* collect car j,
                           (u . v) . nil)
     end;

symbolic procedure msqsimp0 u;
   begin scalar x,!*let!*;
      x := subfg!* . !*sub2;
      subfg!* := nil;
      !*let!* := t;
      u := riccirsimp!* u;
      subfg!* := car x;
      !*sub2 := cdr x;
      return u
   end;

symbolic procedure system!-dummy(u,v);
   % Using them there are no need to remember which indexnames 
   % have been applied as dummy indices in the expressions; 
   if null u then nil
    else if null v 
     then begin scalar x;
             x := mkid('!&00,length indxlist!*);
             indxlist!* := ordn(x . indxlist!*);
             return x . system!-dummy(cdr u,v)
          end
    else car v . system!-dummy(cdr u,cdr v);

 symbolic procedure set!-frasc!*!-mcond!*(u,v,w);
   % U = movable and covariant not movable free indices;
   % V = contravariant not movable free indices;
   % W = dummy indices;
   begin scalar x,y,z;
      x := for each j in (z := append(u,append(v,w))) collect newvar j;
      y :=pair(z,x);
      frasc!* := y;
      frlis!* := union(x,frlis!*);
      x := nil;
      for each j in w do x := list('dummy,j) . x;
      for each j in v do x := list('contrav,j) . x;
      if x then << x := if cdr x then 'and . x else car x;
                   mcond!* := subla(y,formbool(x,nil,'algebraic))>>
   end;

symbolic procedure symm!-setdiff(u,v);
   if null u then v
    else if null v then u
    else if car u member v then symm!-setdiff(cdr u,delete(car u,v))
    else car u . symm!-setdiff(cdr u,v);

symbolic procedure tensorsetk1(u,v,b);
   begin scalar x;
      if eqcar(v,'!*sq)
        then v := '!*iq . !*msqnull!* . cddr u;
      if atom u
        then <<if null b
                 then <<if not get(u,'avalue)
                          then msgpri(nil,u,"not found",nil,nil)
                         else remprop(u,'avalue);
                        return nil>>
                else if (x := get(u,'avalue))
		 then if car x neq 'indexed!-quantity
                        then typerr(list(car x,u),'indexed!-quantity)
                       else put(u,'avalue,list(car x,v))
                else put(u,'avalue,list('indexed!-quantity,v));
               return v>>
       else if not atom car u
	then rerror('riccir!-letfn,2,
                    "Invalid syntax: improper assignment")
       else put(car u,'opmtch,riccir!-xadd!*(nil,cdr u .
                list(nil . (if mcond!* then mcond!* else t),v,nil),
                                    get(car u,'opmtch),b));
      return v
   end;

%********************************************************************;
%***** Possible CLEAR statements:                               *****;
%*****                                                          *****;
%*****  CLEAR SYMMETRY kernel-list;                             *****;
%*****      we can delete only all the symmetry property;       *****;
%*****  CLEAR METRIC TENSOR;                                    *****;
%*****      we keep the tensor but remove its metric tensor     *****; 
%*****         property (symmetry properties remain);           *****;
%*****  CLEAR CHRISTOFFEL SYMBOL;                               *****;
%*****  CLEAR LEVI!-CIVITA SYMBOL|TENSOR;                       *****;
%*****  CLEAR RIEMANN TENSOR;                                   *****;
%*****  CLEAR WEYL TENSOR;                                      *****;
%*****      effect is the same as at CLEAR METRIC;              *****;
%*****  CLEAR INDEX idlist;                                     *****;
%*****  CLEAR SCALAR idlist;                                    *****;
%*****      removes the scalar value of the id's;               *****;
%*****                                                          *****;
%*****  CLEAR list of kernels and/or ids;                       *****;
%*****                                                          *****;
%********************************************************************;

put('scalar,'clearfn,'clear!-scalar);
put('clear!-scalar,'formfn,'formrlis);

symbolic procedure clear!-scalar u;
   % 
   %   CLEAR SCALAR id1,id2,...,idn;
   %
   begin
      rmsubs();
      for each j in u do tensorsetk1(j,nil,nil)
   end;

endmodule;

end;
