%********************************************************************;
%*****              Modification (extension)  of REDUCE          ****;
%********************************************************************;

module riccir!-mods;

% Author: Jozsef Kadlecsik

global '(allrelations!* !*spec!-where!-constr!*
         mcond!*);

fluid '(!*resimp frasc!* newrule!* oldrules!*);

version!-dependent();

symbolic procedure assgneval u; 
   % modification : REVLISFN instead of REVLIS to avoid
   %                the simplification of -0 to 0 as index on the left hand
   %                side in the case of setk and indexed quantities;
   % Evaluate (possible) assignment statements and return results in 
   % a form that allows required printing of such assignments.
   begin scalar x,y;
   a: if atom u then go to b
       else if car u eq 'setq then x := ('setq . cadr u) . x
       else if car u eq 'setel 
        then x := ('setel . mkquote lispeval cadr u) . x
       else if car u eq 'setk
        then x := ('setk . mkquote 
                             if atom (y := lispeval cadr u) then y 
                              else (if z and (z := get(z,'revlisfn))
                                      then apply1(z,y)
                                     else revlis y) where z=getrtypeor y) . x
       else go to b;
      u := caddr u;
      go to a;
   b: u := mkquote lispeval u;
   c: if null x then return(lispeval u . u);
      u := list(caar x,cdar x,u);
      x := cdr x;
      go to c
   end;

% T is very 'popular' as index, therefore we need these modifications:

symbolic procedure forall u;
   begin scalar x,y;
      x := for each j in car u collect newvar j;
      y := pair(car u,x);
%     mcond!* := subla(y,cadr u);
      mcond!* := if cadr u eq 't then cadr u else subla(y,cadr u);
      % because t can play role as arbitrary variable too; 
      frasc!* := y;
      frlis!* := union(x,frlis!*);
      return lispeval caddr u
   end;

symbolic procedure newvar u;
   if not idp u then typerr(u,"free variable")
%    else if flagp(u,'reserved)
%     then typerr(list("Reserved variable",u),"free variable")
    else intern compress append(explode '!=,explode u);

% Extended (assignment) LET statement;

%                                 LET <expr> := <expr>; == <expr> := <expr>;
% FOR ALL <vars>                  LET <expr> := <expr>;
% FOR ALL <vars> SUCH THAT <cond> LET <expr> := <expr>;

symbolic procedure formlet1(u,vars,mode);
   'list . for each x in u collect
      if eqexpr x     
        then list('list,mkquote 'equal,form1(cadr x,vars,mode),
                   !*s2arg(form1(caddr x,vars,mode),vars),nil)
       else if eqcar(x,'setq)    % entry point for the assignment LET;
        then list('list,mkquote 'equal,form1(cadr x,vars,mode),
                   !*s2arg(form1(caddr x,vars,mode),vars),T)
       else form1(x,vars,mode);

symbolic procedure let1 u;
   begin scalar x;
      while u do
         <<if idp u then typerr(u,"rule list")
            else if eqcar(x :=listeval0 car u,'list)
	     then rule!-list(reverse cdr x,t)
            else if idp car u then typerr(u,"rule list")
            else if caar u memq '(replaceby)
             then if frasc!*
                    then rerror('riccir!-mods,1,
                       "=> invalid in FOR ALL statement")
                   else rule!-list(list car u,t)
            else if not(caar u eq 'equal) then typerr(car u,"rule list")
	    else if smemq('!~,car u)
	     then if frasc!* then typerr(car u,"rule")
                   else rule!-list(list car u,t)
            else let1!*(cadar u,caddar u,nil,t,car cdddar u);
           u := cdr u>>
   end;

symbolic procedure let1!*(u,v,w,z,flg);
   % AEVAL is possible instead of REVAL. Which one is better!?
   let2(u,if flg then reval v else v,w,z);

% Extended CLEAR statement;

put('clear,'stat,'clearstat);

symbolic procedure clearstat;
   begin scalar x,y;
      x := scan();
      if idp x and (y := get(x,'clearfn)) then x := scan()
       else y := get('clear,'clearfn);
      return if flagp!*!*(x,'delim) 
               then if flagp!*!*(y,'emptyarg) then y . nil
                     else symerr("CLEAR",t)
              else y . remcomma xread1 'lambda
  end;

put('clear,'clearfn,'clear);
put('clear,'formfn,'formgenclear);

symbolic procedure formgenclear(u,vars,mode);
   list(car u,formclear1(cdr u,vars,mode));

symbolic procedure clear1 u;
   begin scalar x;
      while u do
	 <<if flagp(x := car u,'share)
	     then if not flagp(x,'reserved) then set(x,x) else rsverr x
	    % if argument is an explicit list, clear each element.
	    else if eqcar(x,'list)
		   then u := nil . append(cdr x,cdr u)
	    % The following two cases allow for rules or the lhs of
	    % rules as arguments to CLEAR.
	    else if eqcar(x,'replaceby) or eqcar(x,'replacebysetq)
             then rule!-list(list x,nil)
	    else if smemq('!~,x)
	     then if eqcar(x,'equal) then rule!-list(list x,nil)
		   else rule!-list(list list('replaceby,x,nil),nil)
	    else << %if atom x then remprop(x,'rtypefn);  % see RICCIR!-INTRO;
                    let2(x,nil,nil,nil); let2(x,nil,t,nil)>>;
	   u := cdr u>>
   end;

% In order to use relations locally via a 'where' construction,
% these modifications are needed

% <expr> where <keyword>={<where-keyword> ....}

% !*spec!-where!-constr!* ::= (where!-keywords1 . where!-statfn1) . 
%                             (where!-keywords2 . where!-statfn2) .
%                              ... .
%                             nil
% where!-keywords ::=  where!-keyword1 . where!-keyword2 . ... nil

 
symbolic procedure xreadlist;   
   % Expects a list of expressions enclosed by {, }.
   % Used to allow expressions separated by ; - treated these as progn.
   begin scalar cursym,delim,lst,x;
        lst := !*spec!-where!-constr!*;
        while lst and null x do <<
           x := assoc(nxtsym!*,car lst);
           lst := cdr lst>>;
        if x then return eval cdr x;
        %if assoc(nxtsym!*,allrelations!*) then return relationstat();
	if scan() eq '!*rcbkt!* then <<scan(); return list 'list>>;
    a:	lst := aconc(lst,xread1 'group);
        cursym := cursym!*;
	if cursym eq '!*semicol!*
	  then symerr("Syntax error: semicolon in list",nil)
	 else if scan() eq '!*rcbkt!* and cursym eq '!*comma!*
	  then symerr("Syntax error: invalid comma in list",nil);
	if cursym eq '!*rcbkt!*
	  then return % if delim eq '!*semicol!*
		      %   then 'progn . lst else
		      'list . lst
	 else if null delim then delim := cursym;
        go to a
   end;

symbolic procedure evalletsub1(u,v);
   begin scalar x,w;
      x := car u;
      u := carx(cdr u,'simpletsub);
      if eqcar(x,'list) then x := cdr x else errach 'simpletsub;
      if null cdr x and eqcar(car x,'equal) 
         and (w := assoc('where,prop cadar x))
        then return evalpropsub(cadar x,caddar x,cdr w,list('aeval,mkquote u));
      w := evalletsub2({x,list('aeval,mkquote u)},v);
      return if errorp w then rerror('riccir!-mods,2,"Invalid simplification")
              else car w
   end;

symbolic procedure evalpropsub(fn,u,v,w);
   begin scalar x;
      v := u . v;
      put(fn,'where,v);
      x := errorset!*(w,nil);
      put(fn,'where,cdr v);
      return if errorp x then rerror('riccir!-mods,3,"Invalid simplification")
              else car x
   end;

% This function is borrowed from the 'next' Reduce version.

symbolic procedure evalletsub2(u,v);
  % car u   is an untagged list of rules or ruleset names,
  % cadr u  is an expression to be evaluated by errorset* with the
  %          rules activated locally, 
  % v should be nil unless the rules contain equations.
  % Returns the expression value corresponding to the
  % errorset protocol.
   begin scalar !*resimp,newrule!*,oldrules!*,x,y,z,z1;
      x:=car u; u:=cadr u;
      for each j in x do
         if v and eqcar(j,'replaceby) then z := j . z
          else if null v and eqcar(j,'equal)
           then z := ('replaceby . cdr j) . z
          else if (y := validrule j) or (y := validrule reval j)
           then (y:=reverse car y) and <<rule!-list(y,t); z1 := y . z1>>
          else typerr(car x,"rule list");
 %    z := reversip!* z;
      rule!-list(z,t);
      !*resimp := t;   % Since u may contain (*SQ ... T).
      u := errorset!*(u,nil);
      !*resimp := nil;
      for each j in z . z1 do rule!-list(j,nil);
      % Restore previous environment, if changed.
      for each j in oldrules!*
          do if atom cdar j
               then if idp cdar j
                      then if cdar j eq 'scalar
                             then let3(caar j,cadr j,nil,t,nil)
                            else typelet(caar j,cadr j,nil,t,cdar j)
                     else nil
              else rule!*(car j,cadr j,caddr j,cadddr j,'old);
      return u
   end;

endmodule;

end;
