%*********************************************************************;
%*****              Modified standard quotient                    ****;
%*********************************************************************;

module riccir!-msq;

% Author: Jozsef Kadlecsik

%           The structure of modified standard quotient 
%
%       msq ::=   msf . msf             (numr . denr)
%
%       msf ::=   nil  |  mlt . msf     (lt . red)
%
%       mlt ::=   mtv . sf              (lpow . lc)   (tpow . tc)
%
%       mtv ::=   1    |  msk
%
%       msk ::=   (t1 t2 ... )
%
%       t   ::=   iq | siq . n  
%

global '(!*msfone!* !*msqnull!* !*relation!* alliqlist!*  
         !*resubs !*sqvar!* simpcount!* simplimit!*);

fluid '(!*let!*
        !*resimp alglist!* kord!*);

!*msfone!* := 1 .* 1 .+ nil;

!*msqnull!* := nil ./ (1 .* 1 .+ nil);

symbolic procedure msfonep u;
   u = !*msfone!*;

% Smacro procedures

smacro procedure !*k2msf u;
   !*k2msk u .* 1 .+ nil;

smacro procedure !*k2msq u;
   (!*k2msk u .* 1 .+ nil) ./ !*msfone!*;

smacro procedure !*msk2msf u;
   u .* 1 .+ nil;

smacro procedure !*msk2msq u;
   (u .* 1 .+ nil) ./ !*msfone!*;

smacro procedure !*dsf2msf u;
   1 .* u .+ nil;

smacro procedure !*dsf2msq u;
   (1 .* u .+ nil) ./ !*msfone!*;

smacro procedure !*dsq2msq u;
   (1 .* numr u .+ nil) ./ (1 .* denr u .+ nil);

smacro procedure negmsf u;
   multmsfsf(u,-1);

smacro procedure negmsq u;
   multmsqsq(u,(-1) ./ 1);

symbolic procedure !*k2msk u;
   % U : indexed quantity from a sq;
   if eqcar(u,'!*times) then re!*expt cdr u 
    else if pairp u and indexedp u then list u
    else list(u . 1); 

symbolic procedure re!*expt u;
   if null u then nil
    else (if listp car u and indexedp car u then car u
           else if eqcar(car u,'expt) then cadar u . caddar u
           else car u . 1) . re!*expt cdr u;

% Evaluation function

put('indexed!-quantity,'tag,'!*iq);

put('indexed!-quantity,'rtypefn,'(lambda (x) 'indexed!-quantity));

put('indexed!-quantity,'evfn,'iqevfn);

put('!*iq,'rtypefn,'(lambda (x) 'indexed!-quantity));

put('!*iq,'simpfn,'simp!*iq);

flag('(!*iq),'full);

put('!*iq,'prifn,'!*iqprint);

symbolic procedure iqevfn(u,v);
   (if null v then x else prepsqxx !*msq2sq cadr x) 
      where x=riccirsimpop u;

% Auxiliary function to handle the problem with SQRT (not complete);

put('!*times,'rtype,'indexed!-quantity);

put('!*times,'simpfn,'simp!*times);

flag('(!*times),'full);

put('!*times,'prifn,'!*timesprt);

symbolic procedure simp!*times u; !*k2q u;

% Simplification function

symbolic procedure riccirsimpop u;
   << refresh!-alliqlist();
      mk!*iq riccirsimp u>>;

symbolic procedure refresh!-alliqlist;
   % alliqlist!* = Ordered list of user defined symbol and tensor ids;
   alliqlist!* := append(kord!*,setdiff(append(ordn tensorlist!*,
                                               ordn symbollist!*),
                                        kord!*));

symbolic procedure mk!*iq u;
   '!*iq . u . if !*resubs then !*sqvar!* else list nil;

symbolic procedure simp!*iq u;
   !*msq2sq riccirsimp u;

% Main operators to introduce the msq structure;

symbolic procedure riccirsimp u;
   begin scalar x;
      x := get('relation0,'where);
      return call!-algs(mp3msq riccirsimp!* u,
                        (x and car x) or !*relation!*)
   end;
 
symbolic procedure riccirsimp!* u;
   begin scalar x,alglist!*;
      if simpcount!* > simplimit!*
        then << simpcount!* := 0;
                rerror('alg,12,"Simplification recursion too deep")>>
       else simpcount!* := simpcount!* + 1;
      u := if atom u then if indexedp!* u then riccirsimpatom u
                           else if numr(x := simp!* u) then !*dsq2msq x
                           else !*msqnull!* 
            else if x := get(car u,'riccirfn)
                   then if flagp(car u,'full) then apply1(x,u)
                         else apply1(x,cdr u)
            else if car u eq '!*sq 
                   then if !*let!* then remsq !*sq2msq cadr u
                         else if caddr u and null !*resimp then msqsimp0 u
                         else resimpmsq newmsqdummy !*sq2msq cadr u
            else if car u eq '!*iq 
                   then if !*let!* then remsq cadr u
                         else if caddr u and null !*resimp then msqsimp0 u 
                         else resimpmsq newmsqdummy cadr u
             else if car u eq 'plus 
                   then << x := riccirsimp!* cadr u;
                           for each j in cddr u do
                              x := addmsq(riccirsimp!* j,x); x>>
            else if car u eq 'minus then negmsq riccirsimp!* cadr u
            else if car u eq 'difference 
                   then addmsq(riccirsimp!* cadr u,
                               negmsq riccirsimp!* caddr u)
            else if car u eq 'times or car u eq '!*times
                   then << x := riccirsimp!* cadr u;
                           for each j in cddr u do
                              x := multmsq(riccirsimp!* j,x); x>>
            else if car u eq 'quotient 
                   then multmsq(riccirsimp!* cadr u,
                                riccirsimp!*('recip . cddr u))
            else if car u eq 'expt then riccirsimpexpt u
            else if x := get(car u,'simpfn)
                   then !*sq2msq (if flagp(car u,'full) or x eq 'simpiden
                                    then apply1(x,y)
                                   else apply1(x,cdr y)) 
                                       where y=callriccirsimp u
            else if numr(x := simp!* u)
                   then !*dsq2msq x 
            else !*msqnull!*;
      simpcount!* := simpcount!* - 1;
      return u
   end;
                         

symbolic procedure riccirsimpexpt u;
   begin scalar v,x,n;
      v := simpexpon carx(cddr u,'expt);
      u := riccirsimp cadr u;
      if sfree (x := msfindices numr u)
        then rerror('riccir!-msq,1,
                    "Free index in the argument of EXPT");
      return if x and numberp (n := numr v) and fixp n and denr v = 1
               then if n > 0 then exptmsq(u,n)
                     else revpr exptmsq(u,-n)
              else !*sq2msq simpexpt list(mk!*sq !*msq2sq u,mk!*sq v)
   end;

symbolic procedure callriccirsimp u;
   begin scalar x,y;
      x := car u;
      return x . for each j in cdr u collect << 
            y := riccirsimp j;
            if sfree msfindices numr y
              then rerror('riccir!-msq,2, 
                    list("Free index in the argument of",x));
            mk!*sq !*msq2sq y>>
   end;

% Resimplification function;

% In every term;

symbolic procedure resimpmsq u;
   multmsq(resimpmsf numr u,revpr resimpmsf denr u);

symbolic procedure resimpmsf u;
    if null u then nil ./ nil
     else addmsq(if lpow u = 1 then x
                  else multmsq(resimpmsk lpow u,x),resimpmsf red u)
                      where x=resimpsf2msq lc u;

symbolic procedure resimpmsk u;
   riccirsimp!*('times . mk!*expt u);

symbolic procedure mk!*expt u;
   for each j in u collect
      if numberp cdr j
        then if cdr j = 1 then car j
              else list('expt,car j,cdr j)
       else j;

symbolic procedure resimpsf2msq u;
   if null u then !*msqnull!*
    else if domainp u then !*dsf2msq u
    else addmsq(multmsq(exptmsq(riccirsimp!* mvar u,ldeg u),
                        resimpsf2msq lc u),
                resimpsf2msq red u);

% Only in the tensorial terms;

symbolic procedure remsq u;
   multmsq(remsf numr u,revpr remsf denr u);

symbolic procedure remsf u;
   if null u or lpow u = 1 then u ./ !*msfone!*
    else addmsq(multmsqsq(resimpmsk lpow u,lc u ./ 1),remsf red u);

% Conversion functions;

symbolic procedure !*sq2msq u;
   !*sf2msf numr u ./ !*sf2msf denr u;

symbolic procedure !*sf2msf u;
   if null u then nil
    else if domainp u then 1 .* u .+ nil
    else addmsf(if sfp mvar u and sfindexedp mvar u
                  then multmsf(exptmsf(!*sf2msf mvar u,ldeg u),
                               !*sf2msf lc u)
                 else if null sfp mvar u and tensor!-exprp mvar u
                  then multmsf(exptmsf(!*k2msf mvar u,ldeg u),
                               !*sf2msf lc u)
                 else multmsfsf(!*sf2msf lc u,!*p2f lpow u),
                !*sf2msf red u);

symbolic procedure sfindexedp u;
   (if sfp mvar u then sfindexedp mvar u
     else tensor!-exprp mvar u)
   or (null domainp lc u and sfindexedp lc u)
   or (red u and null domainp red u and sfindexedp red u);

symbolic procedure !*msq2sq u;
   multsq(!*msf2sf numr u ./ 1,1 ./ !*msf2sf 
          if !*let!* then mk!-dummy0(denr u,sdummy msfindices denr u,
                                     append(car x,cdr x)) 
                            where x=free!-and!-dummy2 msfindices numr u
           else denr u);

symbolic procedure !*msf2sf u;
   if null u then nil
    else if lpow u = 1 then lc u
    else addf(multf(!*msk2sf lpow u,lc u),!*msf2sf red u);

symbolic procedure !*msk2sf u;
   if !*let!* 
     then !*k2f (if cdr x then '!*times . x else car x) where x=mk!*expt u
    else !*msk2sf0(u,nil);

symbolic procedure !*msk2sf0(u,v);
   if null u then nil
    else if numberp cdar u 
     then car u .* (if y then y else 1) .+ nil where y=!*msk2sf0(cdr u,v)
    else begin scalar x;
            x := buildmonom(car u,cdr u);   % (free . dummy) . msk;
            u := setdiff(u,cdr x);
            x := reindexsimp(x,v);          % (free . dummy) . msk;
            v := append(upping caar x,append(cdar x,v));
            return if caar x
                     then y .** 1 .* (if null u then 1
                                       else !*msk2sf0(u,v)) .+ nil
                           where y=if cddr x then '!*times . cdr x else cadr x
                    else !*msk2sf1(u,cdr x,v,cdar x,1)
         end;

symbolic procedure !*msk2sf1(u,v,w,z,n);
   (if null u then y .** n .* 1 .+ nil
     else (if x 
             then if (x := setdiff(u,car x)) then !*msk2sf1(x,v,w,z,n + 1)
                   else y .** (n + 1) .* 1 .+ nil
            else y .** n .* !*msk2sf0(u,w) .+ nil)
                where x=quotxn0(u,(nil . z) . v))
        where y=if cdr v then '!*times . v else car v;

endmodule;

end;
