%********************************************************************;
%*****      Elementary operations with the modified sq's         ****;
%********************************************************************;

module riccir!-msqalg;

% Author: Jozsef Kadlecsik

global '(!*msfone!* !*msqnull!*);

fluid '(!*tmpremult);

% Addition

symbolic procedure addmsq(u,v);
   % U and V are modified standard quotients.
   % Value is canonical sum of U and V;
   if null numr u then v
    else if null numr v then u
    else if msfonep denr u and msfonep denr v
           then addmsf(numr u,numr v) ./ denr u
    else begin scalar x,y,z;
        x := gcdmsf(denr u,denr v);
        z := quotmsf(denr u,x) ./ quotmsf(denr v,x);
        y := addmsf(multmsf(denr z,numr u),
                    multmsf(numr z,numr v));
        if null y then return !*msqnull!*; 
        z := multmsf(denr u,denr z);
        if msfonep x then return y ./ z;     % ONEP
        x := gcdmsf(y,x);
        return if msfonep x then y ./ z
                else quotmsf(y,x) ./ quotmsf(z,x)
    end;

symbolic procedure addmsf(u,v);
   % U and V are modified standard forms. 
   % Value is modified standard form for U+V;
   if u and v and symm!-setdiff(sfree msfindices u,sfree msfindices v)
     then rederr "Different type of objects cannot be added"
    else addmsf0(u,v);

symbolic procedure addmsf0(u,v);
   if null u then v
    else if null v then u
    else if lpow u = 1 then addmsfm(u,v)
    else if lpow v = 1 then addmsfm(v,u)
    else if lpow u = lpow v 
           then (if null x then y else lpow u .* x .+ y)
                   where x=addf(lc u,lc v),y=addmsf0(red u,red v)
    else if riccir!-ordp(lpow u,lpow v) then lt u .+ addmsf0(red u,v)
    else lt v .+ addmsf0(u,red v);

symbolic procedure addmsfm(u,v);
   % U is a domain modified standard form, V a modified standard form.
   % Value is a modified standard form for U+V;
   if null v then u
    else if lpow v = 1 
     then (if x then 1 .* x .+ nil else nil) where x=addf(lc u,lc v)
    else lt v .+ addmsfm(u,red v);

% Multiplication

symbolic procedure multmsq(u,v);
   % U and V are modified standard quotients.
   % Value is product of U and V;
   if null numr u or null numr v then !*msqnull!*
    else begin scalar x,y,!*tmpremult;
        if msfonep denr u and msfonep denr v then go to a;
        x := gcdmsf(numr u,denr v);
        y := gcdmsf(numr v,denr u);
        u := multmsf(quotmsf(numr u,x),quotmsf(numr v,y)) ./
               multmsf(quotmsf(denr u,y),quotmsf(denr v,x));
        go to b;
     a: u := multmsf(numr u,numr v) ./ denr u;
     b: return if !*tmpremult
        % !*TMPREMULT is T, if a metric tensor was applied; 
                 then multmsq(remsf numr u,revpr remsf denr u)
                else u
         end;

symbolic procedure multmsf(u,v);
   % U and V are modified standard forms.
   % Value is modified standard form for U*V;
   if null u or null v then nil
    else if lpow u = 1 then multmsfsf(v,lc u)
    else if lpow v = 1 then multmsfsf(u,lc v)
    else addmsf(addmsf(multmst(lt u,lt v),multmsf(red u,v)),
                multmsf(lt u .+ nil,red v));

symbolic procedure multmst(u,v);
   % U and V are modified standard terms.
   % Value is modified standard form for U*V;
   (if null x then nil
     else multmsfsf(x,multf(tc u,tc v))) where x=multmk2(tpow u,tpow v);

symbolic procedure multmsqsq(u,v);
   multmsq(u,!*dsq2msq v);

symbolic procedure multmsfsf(u,v);
   if null(u and v) then nil
    else multmsfsf0(u,v);

symbolic procedure multmsfsf0(u,v);
   if null u then nil
    else lpow u .* multf(lc u,v) .+ multmsfsf0(red u,v);

symbolic procedure exptmsq(u,n);
   begin scalar x;
      x := u;
      while (n := n-1) > 0 do x := multmsq(u,x);
      return x
   end;

symbolic procedure exptmsf(u,n);
   begin scalar x;
      x := u;
      while (n := n-1) > 0 do x := multmsf(u,x);
      return x
   end;

% Quotient

symbolic procedure quotmsf(p,q);
   % P and Q are modified standard forms
   % Value is the quotient of P and Q if it exists or NIL;
   if null p then nil
    else if p=q then !*msfone!*
    else if msfonep q then p
    else if lpow p = 1 or lpow q = 1
     then (if lpow p = 1 
             then if lpow q = 1 
                    then if x then !*dsf2msf x 
                          else nil
                   else nil
            else if x 
                   then if red p
                          then (if y then lpow p .* x .+ y
                                 else nil) where y=quotmsf(red p,q)
                         else lpow p .* x .+ nil
            else nil) where x=quotf(lc p,lc q)
    else (if x
            then << p := addmsf(p,multmsf(x,negmsf q));
                    if null p then x
                     else if null (p := quotmsf(p,q)) then nil
                     else addmsf(x,p)>>
           else nil) where x=quotmst(lt p,lt q);

symbolic procedure quotmst(u,v);
   (if x and y
      then if car y then mk!-dummy car y .* x .+ nil
            else 1 .* x .+ nil
     else nil) where y=quotxn(tpow u,tpow v),x=quotf(tc u,tc v);

symbolic procedure quotxn(u,v);
   if null v then list u
    else if null u then nil
    else if numberp cdar v
           then (if x
                   then if cdar v > cdr x then nil
                         else if cdar v = cdr x 
                          then quotxn(delete(x,u),cdr v)
                         else quotxn(repasc(car x,cdr x - cdar v,u),cdr v)
                  else nil) where x=assoc(caar v,u)
    else (if x then quotxn(setdiff(u,car x),setdiff(v,cdr x))
           else nil) where x=quotxn0(u,buildmonom(car v,cdr v));

symbolic procedure quotxn0(u,v);
   if null u or numberp cdar u then nil
    else (if caar x = caar v                   % it is needed in GCDMSF;
             and length cdar x = length cdar v 
             and cdr x = where!-dummy(cdar x,cdar v,cdr v)
            then cdr x . cdr v
           else quotxn0(setdiff(u,cdr x),v)) where x=buildmonom(car u,cdr u); 

% Greatest commom divisor;

symbolic procedure gcdmsf(u,v);
   dummygcdmsf gcdmsf1(u,v);

symbolic procedure gcdmsf1(u,v);
   if null u then v
    else if null v then u
    else if msfonep u or msfonep v then !*msfone!*  
    else if lpow u = 1 then gcdmsd(lc u,v)
    else if lpow v = 1 then gcdmsd(lc v,u) 
    else if quotmsf(u,v) then v
    else if quotmsf(v,u) then u
    else gcdmsf2(u,v);

symbolic procedure gcdmsf2(u,v);
   begin scalar x,y,z;
      x := comfacmsf u;
      y := comfacmsf v;
      z := gcdmsf1(cdr x,cdr y);
      u := quotmsf(u,comfac!-to!-msf2 x);
      v := quotmsf(v,comfac!-to!-msf2 y);
      z := multmsf(gcdmsk(u,v),z);
      return if car x and car y 
               then multmsf(gcdmsk(!*msk2msf car x,!*msk2msf car y),z)
              else z
   end;

symbolic procedure gcdmsd(u,v);
   if null v then !*dsf2msf u
    else if lpow v = 1 then !*dsf2msf gcdf(lc v,u)
    else gcdmsd(gcdf(lc v,u),red v);

symbolic procedure gcdmsk(u,v);
   if u = v then u
    else if msfonep u or msfonep v then !*msfone!*
    else if quotmsf(u,v) then v
    else if quotmsf(v,u) then u
    else !*msfone!*;

symbolic procedure comfacmsf p;
   % P is a non-atomic modified standard form.
   % CAR of result is the longest contracted iq. of leading kernel in
   % every term in P (or NIL). CDR is gcd of all coefficients of
   % iq of leading kernel.
   begin scalar x,y,z;
      p := tensorfac p;
      if null red p then return lpow p . fac!-to!-msf lc p;
      x := lc p;
      y := lpow p;
   a: p := red p;
      if lpow p = 1 then return nil . gcdmsf!-fac(x,p)
       else if null (y := facmult(y,lpow p)) then return nil . gcdmsf!-fac(x,p)
       else if null red p then return caar y . gcdmsf!-fac(lc p,x)
       else x := tensorfac gcdmsf!-fac(lc p,x);
      go to a
   end;

symbolic procedure facmult(u,v);
   quotxn!*(u,nil,u,v,sdummy argindices u,sdummy argindices v);

symbolic procedure gcdmsf!-fac(u,v);
   gcdmsf(fac!-to!-msf u,fac!-to!-msf v);

symbolic procedure comfac!-to!-msf2 u;
   if null car u then cdr u else multmsf(car u .* 1 .+ nil,cdr u);

symbolic procedure fac!-to!-msf u;
   if null u or lpow u = 1 then u
    else addmsf(multmsf(lpow u .* 1 .+ nil,fac!-to!-msf lc u),
                fac!-to!-msf red u);

symbolic procedure dummygcdmsf u;
   if null sfree msfindices u then u
    else begin scalar x,y,z,v,w; integer n;
            u := comfacmsf u;
            x := sfree msfindices cdr u;
            w := scalpart car u;
            v := setdiff(car u,w);
            n := length v;
            while null y and n neq 0 do <<
              z := for each j in comb(v,n) collect j . sfree argindices j;
              while null y and z do
                if sfree append(x,cdar z) then z := cdr z
                 else y := caar z;
              n := n - 1>>;
            return if null y
                     then if null w then !*msfone!*
                           else w .* 1 .+ nil
                    else multmsf(append(y,w) .* 1 .+ nil,cdr u)
         end; 

% Factorization;

symbolic procedure tensorfac u;
   if null u then nil
    else if lpow u = 1 then u
    else (caar x .* tensorfac cdar x .+ tensorfac cdr x) 
           where x=tensorfac0(lpow u .* !*dsf2msf lc u,red u,nil,
                              sdummy argindices lpow u);

symbolic procedure tensorfac0(u,v,w,z);
   if null v then u . w
    else if lpow v = 1 then u . addmsf(v,w)
    else (if x then tensorfac0(mktfacform(x,tc u,!*dsf2msf lc v),red v,w,z)
           else tensorfac0(u,red v,addmsf(lt v .+ nil,w),z))
             where x=quotxn!*(tpow u,nil,tpow u,lpow v,z,
                              sdummy argindices lpow v);

symbolic procedure quotxn!*(u,w,a,b,z1,z2);
   if null u 
     then if w then reverse w . (a . b) else nil
    else if numberp cdar u
           then (if x
                   then if cdar u = cdr x 
                          then quotxn!*(cdr u,x . w,
                                        delete(x,a),delete(x,b),z1,z2)
                         else if cdar u > cdr x
                          then quotxn!*(cdr u,
                                        (car x . (cdar u - cdr x)) . w,
                                        repasc(car x,cdar u - cdr x,a),
                                        delete(x,b),z1,z2)
                         else quotxn!*(cdr u,
                                       (car x . (cdr x - cdar u)) . w,
                                       delete(x,a),
                                       repasc(car x,cdr x - cdar u,b),z1,z2)
                  else quotxn!*(cdr u,w,a,b,z1,z2)) where x=assoc(caar u,b)
    else (if x then quotxn!*(cdr u,
                             car u . w,
                             delete(car u,a),
                             where!-dummy(caar x,cdar x,delete(cdr x,b)),z1,z2)
            else quotxn!*(cdr u,w,a,b,z1,z2)) 
               where x=findsame(car u,b,xn(indices car u,z1),z2);

symbolic procedure findsame(u,v,z1,z);
   if null v or numberp cdar v then nil
    else (if sameposl2(z1,z2)
             and u = where!-dummy(z1,z2,car v) 
            then (append(z1,z2) . append(z2,z1)) . car v
           else findsame(u,cdr v,z1,z)) 
               where z2=xn(indices car v,z);
 
symbolic procedure mktfacform(x,u,v);
   car x .* addmsf(mktfacform0(cadr x,u),mktfacform0(cddr x,v));

symbolic procedure mktfacform0(x,u);
   if x then multmsf(!*msk2msf x,u) else u;

symbolic procedure sameposl2(u,v);
   (u and v and samepos(car u,car v) and sameposl2(cdr u,cdr v))
   or null(u or v);

endmodule;

end;
