%*********************************************************************;
%*****                 Modified multiplication                    ****;
%*********************************************************************;

module riccir!-mult;

global '(diffindices!* indexlist!* indexnam!* kronecker!* metric!*);

fluid '(!*metric !*symmsign !*tmpremult dummy!* free!*);

symbolic procedure multmk2(u,v);
   begin scalar !*symmsign;
      return if null car (v := temp!-dummy(u,v))
               then cdr v .* 1 .+ nil
              else if null (u := indexsimp car v) then nil
              else append(mktkern u,cdr v) 
                       .* (if !*symmsign then -1 else 1) .+ nil
                          
      end;

symbolic procedure indexsimp u;
   begin scalar x;
      % A : checking the symmetry...
      x := argsymm u;
      if symm!-antisymm x then return nil;
      % B : ordering the tensors and checking the symmetry again...
      if null(u := mkmonom(u,appinvs sdummy argindices u,cdar x . cddr x)) 
        then return nil;
      % C: applying the metric tensors...
      if kronecker!* or metric!* 
        then u := apply!-metric(u,u);
      % If *tmpremult is T then resimplification will be needed
      % because of the index-substitutions from the applied metric tensor(s);
      return if !*tmpremult then u
              % D: generating the new dummy indices 
              %    and applying the symmetries ...
              else mk!-dummy apply!-symm(u,argsymm u,truesdummy argindices u)
   end;

symbolic procedure truesdummy u;
   setdiff(u,sfree u);

symbolic procedure mkmonom(u,v,w);
   unpack!-monoms(monom!-ordn monom mkmonom0(u,v),w);

symbolic procedure mkmonom0(u,v);
   for each j in pair(u,for each k in u collect allindices k)
      collect (car j . (setdiff(cdr j,x) . nil) . (x . nil))
                  where x=xn(cdr j,v);

symbolic procedure unpack!-monoms(u,w);
   begin scalar x,y,z;
      while u and null z do <<
         y := for each j in car u collect car j;
         z := mult!-symm(y,car w,cdr w);
         x := append(x,y);
         u := cdr u>>;
      return if z then nil else x
   end; 
         
symbolic procedure monom u;
   % U : list of (tensor . (free . nil) . (dummy . nil))
   if null u then nil
    else inmonom!-ordn x . monom setdiff(u,x) 
           where x=car u . monom0(indicesinv caddar u,cdr u,nil,nil);
           
symbolic procedure monom0(u,v,w,bool);
   if null v 
     then if bool then monom0(u,w,nil,nil)
           else nil
    else if xnp(u,caddar v) 
      then car v . monom0(sfree append(u,indicesinv caddar v),cdr v,w,T)
    else monom0(u,cdr v,car v . w,bool);

symbolic procedure reindexsimp(u,v);
   % renames the dummy indices in *MSK2SF;
   ((caar u . x) . where!-dummy(x,cdar u,cdr u))
        where x=choose!-dummy(cdar u,indexlist!*,union(upping caar u,v));

symbolic procedure mk!-dummy u;
   % renames the dummy indices in MULTMK;
   mk!-dummy0(u,cdr x,car x) where x=free!-and!-dummy2 argindices u;
   
symbolic procedure mk!-dummy!* u; 
   % renames the dummy indices in MKMSQ;
   mk!-dummy0(u,cdr x,car x) where x=free!-and!-dummy2 indices u;
   
symbolic procedure newmsqdummy u;
   % renames the dummy indices in RICCIRSIMP*;
   newmsfdummy numr u ./ newmsfdummy denr u;

symbolic procedure newmsfdummy u;
    mk!-dummy0(u,cdr x,car x) where x=free!-and!-dummy2 msfindices u;

symbolic procedure mk!-dummy0(u,v,w); 
   % V : dummy indices;
   % W : free indices;
   where!-dummy(choose!-dummy(v,indexlist!*,upping w),v,u);
 
symbolic procedure temp!-dummy(u,v);
   % Temporary dummy indices to avoid index conflict in INDEXSIMP
   begin scalar x,y,z,free!*,dummy!*;
      x := free!-and!-dummy2 argindices u;
      y := free!-and!-dummy2 argindices v;
      if xnp(car x,car y) then rerror('riccir!-mult,1,
        list("An index occurs twice in cov. or contrav. positions",
             " of a multiplication"));
      free!* := (setdiff(union(a,b),z := xn(a,b)))
                  where a=upping car x,b=upping car y;
      dummy!* := union(z,union(cdr x,cdr y));
      % New dummy indices instead of those ones, which occur in
      % the other expression as free indices...
      u := temp!-dummy0(u,cdr x,car y);
      v := temp!-dummy0(v,cdr y,car x);
      % New dummy indices instead of those ones, which occur in
      % the other expression as dummy indices...
      if (z := xn(cdr x,cdr y))
        then << y := choose!-dummy(z,setdiff(indexlist!*,dummy!*),free!*);
                v := where!-dummy(y,z,v)>>;
      return monappend(u,v)
   end;

symbolic procedure monappend(u,v);
   (append(setdiff(u,a),setdiff(v,b)) . ordscal(a,b))
       where a=scalpart u,b=scalpart v;

symbolic procedure scalpart u;
   if null u or numberp cdar u then u
    else scalpart cdr u;

symbolic procedure ordscal(u,v);
   if null u then v
    else if null v then u
    else if caar u = caar v 
           then (caar u . (cdar u + cdar v)) . ordscal(cdr u,cdr v)
    else if ordp(caar u,caar v) then car u . ordscal(cdr u,v)
    else car v . ordscal(u,cdr v);

symbolic procedure temp!-dummy0(u,d1,f2);
   begin scalar x,y;
      if null (x := xn(d1,upping f2)) then return u;
      y := choose!-dummy(x,setdiff(indexlist!*,dummy!*),free!*);
      dummy!* := append(dummy!*,y);
      return where!-dummy(y,x,u)
   end;

symbolic procedure choose!-dummy(v,w,z);
   % diffindices!* will may be used as free indices;
   choose!-dummy0(v,setdiff(w,diffindices!*),z);

symbolic procedure choose!-dummy0(v,w,z);
   if null v then nil
    else if null w then gen!-dummy(0) . choose!-dummy0(cdr v,w,z)
    else if car w memq z then choose!-dummy0(v,cdr w,z)
    else car w . choose!-dummy0(cdr v,cdr w,z);
    
symbolic procedure gen!-dummy n;
   (if x memq indexlist!* then gen!-dummy(n+1)
     else << msgpri(nil,x,"introduced as abstract index.",nil,nil);
             indexlist!* := ordn(x . indexlist!*);
             x>>) 
       where x=mkid(indexnam!*,n);

symbolic procedure apply!-metric(u,v);
   if null u 
     then if !*metric 
            then apply!-metric!*(v,appinvs sdummy movable!-argindices v)
           else v
    else if kroneckerp car u                  % Kronecker delta;
     then apply!-metric0(u,v,xnmetrind(u,argindices v))
    else if metricp car u and cadar u neq indexinv caddar u 
     then if not samepos(cadar u,caddar u)    % Kronecker delta;
            then apply!-metric0(u,v,xnmetrind(u,argindices v))
           else if !*metric                   % metric tensor;
            then apply!-metric0(u,v,xnmetrind(u,movable!-argindices v))
           else apply!-metric(cdr u,v)
    else apply!-metric(cdr u,v);

symbolic procedure xnmetrind(u,v);
   xn(cdar u,appinvs sdummy v);

symbolic procedure apply!-metric0(u,v,w);
   if w then << !*tmpremult := t;
                v := apply!-metric1(cdar u,w,delete(car u,v));
                apply!-metric(v,v)>>
    else apply!-metric(cdr u,v);

symbolic procedure apply!-metric1(u,v,w);
   indexsubst(if car v = car u then cadr u else car u,indexinv car v,w);

symbolic procedure apply!-metric!*(u,w);
   if null u then nil
    else begin scalar x,y;
            if null (x := sfree movable!-indices car u)
              then return car u . apply!-metric!*(cdr u,w);
            y := ups xn(x,w);
            w := setdiff(w,appinvs x);
            return if y then << y := appinvs y;
                                u := where!-list(y,indicesinv y,u);
                                car u . apply!-metric!*(cdr u,w)>>
                    else car u . apply!-metric!*(cdr u,w)
         end;
                           
symbolic procedure mult!-symm(u,v,w);
   (v and (mult!-symm0(u,car v,length caar v) 
           or mult!-symm(u,cdr v,w)))
   or mult!-symm0(u,w,1);

symbolic procedure mult!-symm0(u,v,n);
   (if n=1 then symm!-antisymm1(x,v)
     else symm!-antisymm2(x,v))
       where x=find!-mult!-symm(nindx!-expr(u,n),n);

symbolic procedure nindx!-expr(u,n);
   if null u then nil
    else (if caar x and length caar x = n then x . nindx!-expr(cdr u,n)
            else nindx!-expr(cdr u,n)) where x=buildmonom(car u,cdr u);

symbolic procedure buildmonom(u,v);
   (free!-and!-dummy2 argindices x . x) 
      where x=u . buildmonom0(sfree allindices u,v);

symbolic procedure buildmonom0(u,v);
   if null v or numberp cdar v then nil
    else (if sdummy(x := append(x,u)) then car v . buildmonom0(sfree x,cdr v)
           else nil) where x=sfree allindices car v;

symbolic procedure find!-mult!-symm(u,n);
   if null u then nil
    else (if x 
            then if cdar x then car x . find!-mult!-symm(cdr x,n)
           else find!-mult!-symm(cdr u,n)) 
             where x=mkmultsymm(car u,cdr u,
                                if n=1 then caaar u else list caaar u,nil,n); 

symbolic procedure mkmultsymm(u,v,w,z,n);
   if null v then w . z
    else if length cdar u = length cdaar v
            and cdr u = where!-list(caar u,caaar v,
                                    where!-dummy(cdar u,cdaar v,cdar v))
           then mkmultsymm(u,cdr v,
                           if n=1 then caaaar v . w else caaar v . w,z,n)
    else mkmultsymm(u,cdr v,w,car v . z,n);

symbolic procedure symm!-antisymm u;
   symm!-antisymm2(caar u,cdar u) or symm!-antisymm1(cadr u,cddr u);

symbolic procedure symm!-antisymm1(u,v);
   % U : list of lists of 1st symmetric indices,
   % V : list of lists of 1st antisymmetric indices.
   % Returns T, if there is a dummy index in the antisymmetric,
   % or two dummy indices in the symm. and antisymm. indices;
   v and (sdummy symbolic!-indices car v  % means a metric tensor;
          % If there is a dummy index in car v, it must come from tensor;
          or repeats car v                % diagonal element;
          or symm!-antisymm10(u,car v) 
          or symm!-antisymm1(u,cdr v));

symbolic procedure symm!-antisymm10(u,v);
   u and (length sdummy append(car u,v) geq 2  
          or symm!-antisymm10(cdr u,v));

symbolic procedure symm!-antisymm2(u,v);
   % U : list of 2nd symmetric indices,
   % V : list of 2nd antisymmetric indices.
   % Returns T, if there is a dummy index in the antisymmetric,
   % or two dummy indices in the symm. and antisymm. indices;
   v and (allsdummy v                 % means a metric tensor;
          % If there is a dummy index in car v, it must come from tensor;
          or repeats v                % diagonal element;
          or symm!-antisymm20(u,v));

symbolic procedure symm!-antisymm20(u,v);
   u and ((x and allsdummy2(delete(x,car u),v) where x=allsdummy2(car u,v))
          or symm!-antisymm20(cdr u,v));

symbolic procedure allsdummy u;
   cdr u and (indicesinv car u member cdr u or allsdummy cdr u);

symbolic procedure allsdummy2(u,v);
   u and ((indicesinv car u member v and car u) or allsdummy2(cdr u,v));

symbolic procedure apply!-symm(u,s,w);
   begin integer n;
      n := 4;
      u := u . w;
      s := caar s . cdar s . cadr s . cddr s . nil;
      while n neq 0 do <<
        u := apply!-symm0(car u,s,cdr u,n);
        s := cdr u;
        u := car u;
        n := n - 1>>;
      return tkern car u
end;

symbolic procedure apply!-symm0(u,s,w,n);
   % orders the indices using the symmetry properties of U;
   % S : symmetries;
   % W : symbolic dummy indices;
   begin scalar x,y,z,a,b,bool;
      bool := n = 4 or n = 3;
      a := car s;
      s := cdr s;
      while a do <<                
        b := car a;
        x := contr!-symm(b,w,bool);
        u := where!-list(appends!*(y := indxordn(b,w,bool),
                                   z := indxordn(x,w,bool),bool),
                         appends!*(b,x,bool),u);
        if n = 3 or n = 1
          then !*symmsign := setsymmsign(list(not tpermp(y,b),
                                              not tpermp(z,x)),
                                         !*symmsign);
%        z := pair(appends!*(b,x,bool),appends!*(y,z,bool));
%        w := indexsubla!*(z,w);
%        s := indexsubla!*(z,s);
%        a := indexsubla!*(z,cdr a)>>;
        z := pair(b := appends!*(b,x,bool),y := appends!*(y,z,bool));
        s := indexsubla!*(z,s);
        a := indexsubla!*(z,cdr a);
        w := indexsubla!*(pair(xn(b,w),xn(y,w)),w)>>;
      return (u . w) . s
   end; 

symbolic procedure appends!*(u,v,bool);
   (if bool then appends x else x) where x=append(u,v);

symbolic procedure contr!-symm(u,v,bool);
   (if x and cdr x then x else nil) 
      where x=if bool then contr!-symm0(notsubmemx u,v)
               else xn(v,indicesinv sfree u);

symbolic procedure contr!-symm0(u,v);
   if null v then nil
    else (if x 
            then (if length x = length y 
                    then y . contr!-symm0(delete(x,u),setdiff(v,y))
                   else contr!-symm0(u,cdr v)) where y=xn(v,indicesinv x)
           else contr!-symm0(u,cdr v)) where x=submem(indexinv car v,u);

symbolic procedure submem(u,v);
   v and ((u member car v and car v) or submem(u,cdr v));   

symbolic procedure notsubmemx u;
   if null u or null cdr u then u
    else (if x then notsubmemx delete(x,cdr u)
           else car u . notsubmemx cdr u) 
              where x=submemx(indicesinv car u,cdr u);

symbolic procedure submemx(u,v);
   v and ((xn(u,car v) and car v) or submem(u,cdr v));   

symbolic procedure indexsubla!*(u,v);
   if null v then nil
    else if atom v or eqcar(v,'minus)
     then (if x then cdr x else v) where x=assoc(v,u)
    else indexsubla!*(u,car v) . indexsubla!*(u,cdr v);
 
endmodule;

end;
