%********************************************************************;
%*****                  Order functions                          ****;
%********************************************************************;

module riccir!-ord;

% Author: Jozsef Kadlecsik

global '(alliqlist!*);

% Ordering in monoms;
% structures: original    - (tensor . (free . nil)
%                                   . (dummy . nil))
%             firstweight - (tensor . (free . nameweight . indexweight)
%                                   . (dummy . nil) 
%             secweight   - (tensor . (free . nameweight . indexweight)
%                                   . (dummy . weight!-value)) 

symbolic procedure inmonom!-ordn u;
   if null u then nil
    else (if null cdr u 
            then list(car x . cadr x . (caddr x . t))
           else inmonom!-ordn0(u,cdr u,list x . cdr u))
               where x=firstweight car u;

symbolic procedure inmonom!-ordn0(u,v,w);
   if null v then check!-weights(car w,cdr w)
    else inmonom!-ordn0(u,cdr v,
           weight!-namep(list firstweight car v . delete(car v,u),w));

symbolic procedure weight!-namep(u,v);
   (if x = 0 then weight!-namep0(u,nil,v,nil)
     else if x then u else v) where x=firstweightp(cadaar u,cadaar v);

symbolic procedure firstweight u;
   car u . (caadr u . nameweight car u . indexweight car u) . cddr u;

symbolic procedure firstweightp(u,v);
   if car u = car v
     then (if x=0
             then numrlist!-ordp(cddr u,cddr v)
            else x) where x=numrlist!-ordp(cadr u,cadr v)
    else ordp(ordn car u,ordn car v);

symbolic procedure weight!-namep0(u,a,v,b);
   if null car u 
     then (if ordp(for each j in x collect car j,
                   for each j in y collect car j) then x . cdr u
            else y . cdr v) where x=reverse a,y=reverse b
    else if null cdddr caar u then weight!-namep0(nextweight u,a,v,b)
    else if null cdddr caar v then weight!-namep0(u,a,nextweight v,b)
    else (if x=0 then weight!-namep0(cdar u . cdr u,caar u . a,
                                     cdar v . cdr v,caar v . b)
           else if x then append(reverse a,car u) . cdr u
           else append(reverse b,car v) . cdr v)
               where x=secweightp(caar u,caar v);

symbolic procedure secweightp(u,v);
   (if x=0
      then if length cdddr u = length cdddr v
             then numrpair!-ordp(cdddr u,cdddr v)
            else length cdddr u > length cdddr v
     else x) where x=firstweightp(cadr u,cadr v);

symbolic procedure nextweight u;
   mknextweight(car x,cadr x . caddr x,caddr x,
                caar y,cdar y,cadr y,cddr y,cdar u,cdr u,nil)
       where y=getsymm caaar u,x=caar u;

symbolic procedure mknextweight(u,v,w,s2,a2,s1,a1,b,x,z);
   % U : tensor;
   % V : (free . nameweight . indexweight) . dummy;
   % W : dummy;
   % S2 : list of 2nd symmetric indices;
   % A2 : list of 2nd antisymmetric indices;
   % S1 : list of 1st symmetric indices;
   % A1 : list of 1st antisymmetric indices;
   % B : already separated terms;
   % X : other multiplicative tensors;
   % Z : list of list of ordered multiplicative tensors;
   begin scalar as,r;
      return if null(w and x)
               then if null (as := append(revappend z,b))
                      then list(u . car v . (cdr v . t))  . x
                     else ((u . car v  
                                . (cdr v . weight!-value(as,indicesinv cdr v))) 
                                  . as) . x 
              else if (s2 and memsymm2(w,s2)) or (a2 and memsymm2(w,a2))
               then inmonom!-ordn2(u,v,w,s2,a2,s1,a1,b,x,z)
              else if (s1 and (as := memsymm1(car w,s1))) or
                      (a1 and (as := memsymm1(car w,s1)))
               then << r := inmonom!-ordn1(x,xn(as,w));
                       mknextweight(u,v,setdiff(w,as),
                                    s2,a2,s1,a1,b,cdr r,car r . z)>>
              else if (as := nextcontr(x,indexinv car w)) 
               then mknextweight(u,v,cdr w,s2,a2,s1,a1,b,
                                 delete(as,x),list firstweight as . z)
              else mknextweight(u,v,cdr w,s2,a2,s1,a1,b,x,z) 
      end;

symbolic procedure inmonom!-ordn1(u,v); 
   begin scalar x,y,z;
      while v do <<
        if (x := nextcontr(u,indexinv car v))
          then << y := allnextcontr(u := delete(x,u),indicesinv caddr x,nil);
                  u := cdr y;
                  z := check!-weights(list firstweight x,car y) . z>>;
        v := cdr v>>;
      return appends monom!-ordn z . u
   end;

symbolic procedure check!-weights(u,v);
   if null u
     then if v then inmonom!-ordn v else nil
    else if null cdddar u then caar x . check!-weights(cdar x,cdr x)
                             where x=nextweight(u . v)
    else car u . check!-weights(cdr u,v);

symbolic procedure memsymm1(u,v);
   if null v then nil
    else if u member car v then car v
    else memsymm1(u,cdr v);

symbolic procedure nextcontr(u,v);
   if null u then nil
    else if v member caddar u then car u 
    else nextcontr(cdr u,v);

symbolic procedure allnextcontr(u,v,w);
   if null (u and v) then reverse w . u
    else (if x then allnextcontr(delete(x,u),cdr v,x . w)
           else allnextcontr(u,cdr v,w)) where x=nextcontr(u,car v);

symbolic procedure revappend u;
   if null u then nil
    else if null cdr u then car u
    else append(revappend cdr u,car u);

symbolic procedure inmonom!-ordn2(u,v,w,s2,a2,s1,a1,b,x,z); 
   begin scalar y,y1,y2,y3,r;
      y1 := memsymm2(w,s2);
      y2 := memsymm2(w,a2);
      while w do 
        if y1 and (y := memsymm1(car w,y1)) 
          then << y3 := mknextweight(u,v,y,nil,nil,s1,a1,nil,x,nil);
                  if cdar y3 
                    then << x := cdr (y3 :=  check!-weights2(cdar y3,cdr y3));
                            r := (if null setdiff(y,w) then 's . car y3
                                   else car y3) . r>>;
                  w := setdiff(w,y)>>
         else if y2 and (y := memsymm1(car w,y2))
          then << y3 := mknextweight(u,v,y,nil,nil,s1,a1,nil,x,nil);
                  if cdar y3 
                    then << x := cdr (y3 :=  check!-weights2(cdar y3,cdr y3));
                            r := (if null setdiff(y,w) then 'a . car y3
                                   else car y3) . r>>;
                  w := setdiff(w,y)>>
         else << y3 := mknextweight(u,v,list car w,nil,nil,nil,nil,nil,x,nil);
                 if cdar y3
                   then << x := cdr (y3 := check!-weights2(cdar y3,cdr y3));
                           r := car y3 . r>>;
                 w := cdr w>>;
      return if null (r := append(revappend!* reverse r,b))
               then list(u . car v . (cdr v . t))  . x
              else ((u . car v . (cdr v . weight!-value(r,indicesinv cdr v))) 
                                  . r) . x 
   end;

symbolic procedure check!-weights2(u,v);
   begin scalar x,y;
      while u do
        if null cdddar u 
          then << y := nextweight(u . v);
                  x := caar y . x;
                  u := cdar y;
                  v := cdr y>>
         else << x := car u . x;
                 u := cdr u>>;
      return reverse x . v
   end;

symbolic procedure memsymm2(u,v);
   if null v then nil
    else if memsymm2p(u,car v,nil) then car v
    else memsymm2(u,cdr v);

symbolic procedure memsymm2p(u,v,bool);
   v and ((xn(u,car v) = car v and (bool or memsymm2p(u,cdr v,t)))
          or memsymm2p(u,cdr v,bool));

symbolic procedure revappend!* u;
    appends revappend2(monom!-ordn coll!-symms(u,'a),'a,
                       revappend2(monom!-ordn coll!-symms(u,'s),'s,u));

symbolic procedure coll!-symms(u,s);
   if null u then nil
    else if caar u eq s then cdar u . coll!-symms(cdr u,s)
    else coll!-symms(cdr u,s);

symbolic procedure revappend2(u,s,v);
   if null u then v
    else if caar v eq s then car u . revappend2(cdr u,s,cdr v)
    else car v . revappend2(u,s,cdr v);

symbolic procedure weight!-value(u,v);
   (if null x then t
     else weight!-value0(v,for each j in x collect caddr j,0))
        where x=car allnextcontr(u,v,nil);

symbolic procedure weight!-value0(u,v,n);
   if null(u and v) then nil
    else if xnp(u,car v)
           then if car u member car v 
                  then (n . nth!?(car u,car v,0)) . 
                           weight!-value0(cdr u,v,n + 1)
                 else weight!-value0(cdr u,v,n + 1)
    else weight!-value0(u,cdr v,n);

symbolic procedure nameweight u;
   % It is needed because of the KORD*;
   if atom u then length memq(u,alliqlist!*)
    else if diffopp u 
     then length memq(car u,'(d ld partdiff)) . nameweight cadr u
    else length memq(car u,alliqlist!*);

symbolic procedure indexweight u;
   if atom u then 0
    else if diffopp u then length cddr u . indexweight cadr u
    else length cdr u;

symbolic procedure numrlist!-ordp(u,v);
   if atom u
     then if atom v
            then if u = v then 0
                  else u > v
           else nil
    else if atom v then t
    else if car u = car v then numrlist!-ordp(cdr u,cdr v)
    else car u > car v;

symbolic procedure numrpair!-ordp(u,v);
   if atom u 
     then if atom v then 0 else nil
    else if atom v then t
    else if caar u = caar v
           then if cdar u = cdar v then numrpair!-ordp(cdr u,cdr v)
                 else cdar u > cdar v
          else caar u > caar v;

% Ordering between monoms ;

symbolic procedure monom!-ordn u;
   % ordering the list of monoms;
   if null u or null cdr u then u
    else if null cddr u then monom!-ord2(car u,cadr u)
    else monom!-ordad(car u,monom!-ordn cdr u);

symbolic procedure monom!-ordad(a,u);
   if null u then list a
    else if monom!-ordp(car u,a) then car u . monom!-ordad(a,cdr u)
    else a . u;

symbolic procedure monom!-ord2(u,v);
   if monom!-ordp(u,v) then list(u,v) else list(v,u);

symbolic procedure monom!-ordp(u,v);
   (if z = 0 then ordp(for each j in u collect car j,
                       for each j in v collect car j)
     else z) where z=monomweightp(u,v);

symbolic procedure monomweightp(u,v);
   if null u 
     then if null v then 0 else nil
    else if null v then t
    else (if x=0 then monomweightp(cdr u,cdr v)
           else x) where x=secweightp(car u,car v);

% Ordering at addition;

symbolic procedure riccir!-ordp(u,v);
   if null u then nil
    else if null v then t
    else if numberp cdar u
           then if numberp cdar v 
                  then if caar u = caar v
                         then if cdar u = cdar v then riccir!-ordp(cdr u,cdr v)
                               else cdar u > cdar v
                        else ordp(caar u,caar v)
                 else nil
    else if numberp cdar v then t
    else if car u = car v then riccir!-ordp(cdr u,cdr v)
    else ((if x = y then riccir!-ordp(a,b)
            else (if z = 0 then ordp(u,v)
                   else z) where z=monomweightp!*(mkwnarg x,mkwnarg y))
               where x=setdiff(u,a),y=setdiff(v,b))
            where a=scalpart u,b=scalpart v;

symbolic procedure mkwnarg u;
   for each j in mkmonom0(u,appinvs sdummy argindices u)
      collect firstweight j;

symbolic procedure mkwnarg2(u,v);
   car u . cadr u . (caddr u . weight!-value(v,indicesinv caddr u));

symbolic procedure monomweightp!*(u,v);
   if null u 
     then if null v then 0 else nil
    else if null v then t
    else (if x=0 then monomweightp!*(cdr u,cdr v)
           else x) where x=secweightp(mkwnarg2(car u,cdr u),
                                      mkwnarg2(car v,cdr v));


% Ordering of indices;

symbolic procedure indxordn(u,w,bool);
   if null u or null cdr u then u
    else if null cddr u then indxord2(car u,cadr u,w,bool)
    else indxordad(car u,indxordn(cdr u,w,bool),w,bool);

symbolic procedure indxordad(a,u,w,bool);
   if null u then list a
    else if indxordp(car u,a,w,bool) then car u . indxordad(a,cdr u,w,bool)
    else a . u;

symbolic procedure indxord2(u,v,w,bool);
   if indxordp(u,v,w,bool) then list(u,v) else list(v,u);

symbolic procedure indxordp(u,v,w,bool);
   (if x or y then indxordp0(x,y,bool,t)
     else indxordp0(u,v,bool,nil)) 
       where x=pick!-free(u,w,bool),y=pick!-free(v,w,bool);

symbolic procedure pick!-free(u,w,bool);
   if bool then setdiff(u,w)
    else if u member w then u else nil;

symbolic procedure indxordp0(u,v,bool1,bool2);
   if bool1 
     then if bool2 or sameposl2(u,v) then ordp(ordn u,ordn v)
           else firstup2(u,v)
    else if bool2 or samepos(u,v) then ordp(u,v) else atom u;

symbolic procedure firstup2(u,v);
   if null u then v
    else if null v then t
    else if samepos(car u,car v) then firstup2(cdr u,cdr v) 
    else atom car u;     

endmodule;

end;
