%********************************************************************;
%*****                 Controlling the relations                 ****;
%********************************************************************;

module riccir!-rels;

% Author: Jozsef Kadlecsik

global '(!*relation!* allrelations!* indexlist!* spec!-where!-constr!*
         stcheckfn!* symbollist!* tensorlist!* 
         cursym!* lnnum!* nxtsym!*);

fluid '(!*levi!-civita !*metric
        orig!*);

% Switches in RICCIR

% !*metric       = application of metric tensor during the evaluation;
% !*levi!-civita = tranformation of multiplied Levi-Civita's to 
%                  metric tensors;

% Possible relation for RICCIRSIMP, controlling 
% the appropriate relation in the final recursion.
                        
allrelations!* :=                               % Internal codes: 
   '((partial (partial covariant))              % part!-cov
     (covariant (covariant partial)             % cov!-part
                (covariant lie))                % cov!-lie
     (lie (lie covariant))                      % lie!-cov 
     (riemann (riemann ricci)                   % 1  
              (riemann weyl)                    % 6
              (riemann definition)              % 7
              (riemann secondcov))              % riemann2
     (ricci (ricci riemann)                     % 2
            (ricci invcurv))                    % 3 
     (invcurv (invcurv ricci))                  % 4
     (weyl (weyl riemann))                      % 5 
     (definition (definition riemann)           % riemann1
                 (definition christoffel))      % christoffel
     (christoffel (christoffel definition))     % 8
     (bianchy1 nil)                             % bianchy1
     (bianchy2 nil)                             % bianchy2
     (bianchyc)                                 % 9 
     (jacobi nil)                               % jacobi
     (ordercov));                               % ordercov


spec!-where!-constr!* := (allrelations!* . 'relationstat) . 
                          spec!-where!-constr!*;

stcheckfn!* := '((bianchy1 chkbian1)
                 (bianchy2 chkbian2)
                 (bianchyc chkbianc)
                 (jacobi chkjacobi)
                 (ordercov chkordercov));

lnnum!* := 0;

switch metric,levi!-civita;

deflist('((metric ((T (rmsubs))))
          (levi!-civita ((T (rmsubs))))
         ),'simpfg);

% Controlling the inside relations;

put('relation,'stat,'relationstat);
put('relation,'formfn,'formgenclear);
put('relation0,'formfn,'formgenclear);
put('relation0,'where,nil);   % it makes possible to use relations locally;

%symbolic procedure relationstat;
%   begin scalar x,y,z,fn,w;
%      if cursym!* eq '!*lcbkt!*
%        then << fn := 'list; 
%                if (x := scan()) eq '!*rcbkt!* then symerr('relation,t)>>
%       else if (x := scan()) eq 'equal 
%        then << fn := 'relation0;
%                if (x := scan()) neq '!*lcbkt!*
%                  then if scan() eq '!*semicol!*
%                         then return list(fn,x)
%                        else symerr('relation,t)
%                 else x := scan()>>
%       else fn := 'relation;  
%   a: if scan() memq '(!*comma!* !*semicol!*) then z := x . z
%       else if x eq '!*lcbkt!*
%        then if w then symerr('relation,t)
%              else << x := cursym!*; w := t; go to a>>
%       else if cursym!* neq 'replaceby then << z := x . z; go to b>>
%       else if (y := scan()) eq '!*lcbkt!*
%        then z := list('replaceby,x,xreadlist()) . z
%       else if x memq '(bianchy1 bianchy2 bianchyc jacobi ordercov)
%        then z := list('replaceby,x,xread1 t) . z
%       else if (x := list(x,y)) and (y := scan()) eq '!*lcbkt!*
%        then z := list('replaceby,car x,list(cadr x,xreadlist())) . z
%       else if flagp!*!*(y,'delim) or y eq '!*comma!*
%        then z := ('replaceby . x) . z
%       else z := list('replaceby,car x,list(cadr x,xread1 t)) . z;
%   b: if (y := cursym!*) eq '!*comma!* then << x := scan(); go to a>>
%       else if y eq '!*rcbkt!* and w 
%        then << w := nil; scan(); go to b>>
%       else if (y eq '!*semicol!* and fn eq 'relation)
%         or (y eq '!*rcbkt!* and fn memq '(relation0 list) 
%             and scan() eq '!*semicol!*)
%        then return fn . reverse z 
%       else symerr('relation,t)
%   end;

symbolic procedure relationstat;
   begin scalar x,y,z,fn,w;
      if cursym!* eq '!*lcbkt!*
        then << fn := 'list; 
                if (x := scan()) eq '!*rcbkt!* then symerr('relation,t)>>
       else if (x := scan()) eq 'equal 
        then << fn := 'relation0;
                if (x := scan()) neq '!*lcbkt!*
                  then if scan() eq '!*semicol!*
                         then return list(fn,x)
                        else symerr('relation,t)
                 else x := scan()>>
       else fn := 'relation;  
   a: if scan() memq '(!*comma!* !*semicol!*) then z := x . z
       else if x eq '!*lcbkt!*
        then if w then symerr('relation,t)
              else << x := cursym!*; w := t; go to a>>
       else if cursym!* neq 'replaceby then << z := x . z; go to b>>
       else if (y := scan()) eq '!*lcbkt!*
        then z := list('replaceby,x,xreadlist()) . z
       else if x memq '(bianchy1 bianchy2 bianchyc jacobi ordercov)
        then z := list('replaceby,x,xread1 t) . z
       else if flagp!*!*(y,'delim) or y eq '!*comma!*
        then z := ('replaceby . x) . z
       else << z := list('replaceby,x,y) . z; scan()>>;
   b: if (y := cursym!*) eq '!*comma!* then << x := scan(); go to a>>
       else if y eq '!*rcbkt!* and w 
        then << w := nil; scan(); go to b>>
       else if (y eq '!*semicol!* and fn eq 'relation) 
               or (y eq '!*rcbkt!* and fn memq '(relation0 list) and scan())
        then return fn . reverse z 
       else symerr('relation,t)
   end;

symbolic procedure relation u;
   %
   %   RELATION rel1,rel2,...,reln;    
   %
   %         where rel is 
   %
   %  NIL 
   % 
   %  BIANCHY1 => {i,{j,k,l}}              | {..,..}
   %  BIANCHY2 => {i,j,{k,l,m}}            | {..,..}
   %  BIANCHC | BIANCHYC {i,j,k,l} | {i,j} | {..,..}
   %  JACOBI   => ld(a(i),b(i),c(i))       | {..,..}
   %  ORDERCOV  |
   %   ORDERCOV => d(a(i,j),-k,-l) | @(a(i,j),-k,-l)   | {..,..}
   %
   %  PARTIAL     => COVARIANT   COVARIANT   => PARTIAL 
   %  COVARIANT   => LIE         LIE         => COVARIANT 
   % 
   %  RIEMANN     => RICCI       RICCI       => RIEMANN 
   %  RICCI       => INVCURV     INVCURV     => RICCI 
   %  WEYL        => RIEMANN     RIEMANN     => WEYL
   % 
   %  RIEMANN     => DEFINITION  DEFINITION  => RIEMANN 
   %  CHRISTOFFEL => DEFINITION  DEFINITION  => CHRISTOFFEL 
   %  RIEMANN     => SECONDCOV   
   %                        
   begin
      rmsubs();
      !*relation!* := relation!*(!*relation!*,u)
   end;

symbolic procedure relation0 u;
   list('equal,'relation0,relation!*(nil,u));

symbolic procedure relation!*(u,v);
   begin 
      for each j in stored!-comm v do
        if null j then u := nil
         else if atom j and j memq '(bianchyc ordercov)
          then u := append!-last(u,j)
         else if car j memq '(bianchy1 bianchy2 bianchyc jacobi ordercov)
          then u := append!-last(u,check!-rel(car j,cadr j))
         else if member(j,assoc(car j,allrelations!*))
          then u := append!-last(u,j)
         else errpri2(j,'hold);
      return u
   end;

symbolic procedure stored!-comm u;
   if null u then nil
    else if idp car u
           then (if eqcar(x,'list) then append(stored!-comm cdr x,
                                               stored!-comm cdr u)
                  else car u . stored!-comm cdr u) where x=listeval0 car u
    else cdar u . stored!-comm cdr u;

symbolic procedure append!-last(u,v);
   if null u then list v else car u . append!-last(cdr u,v);

symbolic procedure check!-rel(u,v);
   if null v then u
    else u . if u memq '(bianchy1 bianchy2) 
               then for each j in if list!-lev v = 3
                                    then cdr v else list v
                       collect (if x then x
                                 else errpri2(j,'hold))
                                    where x=apply1(cadr assoc(u,stcheckfn!*),j)
              else for each j in if eqcar(v,'list) then cdr v else list v
                      collect if apply1(cadr assoc(u,stcheckfn!*),j) then j
                               else errpri2(j,'hold);

symbolic procedure list!-lev u;
   if atom u then 0
    else if car u eq 'list then 1 + list!-lev cdr u
    else if eqcar(car u,'list) then list!-lev car u
    else list!-lev cdr u;
  
symbolic procedure chkbian1 u;
   eqcar(u,'list) and (u := cdr u) and null cddr u
   and ((genindexp car u and eqcar(cadr u,'list))
        or (eqcar(car u,'list) and genindexp cadr u))
   and genindexlistp u; 

symbolic procedure chkbian2 u;
   eqcar(u,'list) and (u := cdr u) and null cdddr u
   and ((genindexp car u and genindexp cadr u and eqcar(caddr u,'list))
        or (eqcar(car u,'list) and genindexp cadr u and genindexp caddr u))
   and genindexlistp u; 

symbolic procedure genindexp u;
   atom u or (car u memq '(minus !~) and null cddr u 
              and (atom cadr u 
                   or (car u eq 'minus and caadr u eq '!~ 
                       and null cddadr u and atom cadadr u)));

symbolic procedure genindexlistp u;
   if null u then nil
    else ((if eqcar(x,'list) 
             then if length x = 4 and genindexp cadr x 
                     and genindexp caddr x and genindexp cadddr x
                    then cdr x 
                   else errpri2(x,'hold)
            else x) . genindexlistp cdr u) where x=car u;

symbolic procedure chkbianc u;
   eqcar(u,'d) and null cdddr u and (riemannp cadr u or riccip cadr u)
   and indexinv caddr u member symbolic!-indices cdadr u;

symbolic procedure chkjacobi u;
   eqcar(u,'ld) and cdddr u and null cddddr u 
   and lieindexlist cdr u and simplelieargl cdr u;

symbolic procedure simplelieargl u;
   null u or (simpleliearg car u and simplelieargl cdr u);

symbolic procedure simpleliearg u;
   not atom u and (if car u eq 'times then onlyindexedp cdr u
                    else indexedp u);

symbolic procedure onlyindexedp u;
   null u or (pairp car u and indexedp car u and onlyindexedp cdr u);

symbolic procedure chkordercov u;
   (eqcar(u,'partdiff) or car u eq 'd) and cdddr u;

put('relation,'clearfn,'clear!-relation);
put('clear!-relation,'stat,'relationstat);
put('clear!-relation,'formfn,'formgenclear);

symbolic procedure clear!-relation u;
   %
   %   CLEAR RELATION rel1,rel2,...,reln;
   %
   begin
      rmsubs();
      u := for each j in stored!-comm u collect 
             if null j then j
              else if atom j and j memq '(bianchyc ordercov) then j
              else if car j memq '(bianchy1 bianchy2 bianchyc jacobi ordercov)
               then check!-rel(car j,cadr j)
              else if member(first!-two j,assoc(car j,allrelations!*))
               then check!-rel(first!-two j,if null cddr j then nil
                                             else caddr j)
              else typerr(j,'hold);
      !*relation!* := clear!-relation0(!*relation!*,u)
   end;

symbolic procedure clear!-relation0(u,v);
   if null u or null v then nil
    else if car u = car v then clear!-relation0(cdr u,cdr v)
    else if pairp car v and eqcar(car u,caar v)
            and caar v memq '(bianchy1 bianchy2 bianchyc jacobi ordercov)
     then (if x then (caar v . x) . clear!-relation0(cdr u,cdr v)
            else clear!-relation0(cdr u,cdr v)) where x=setdiff(cdar u,cdar v)
    else if pairp car u and pairp car v and first!-two car u = first!-two car v
     then (if x then append(first!-two car v,x) . clear!-relation0(cdr u,cdr v)
            else clear!-relation0(cdr u,cdr v)) where x=setdiff(cddar u,cddar v)
    else car u . clear!-relation0(cdr u,v);

% Status messages about the switches and the relationations;

put('status,'stat,'endstat);
put('status,'psopfn,'status!-prt);

symbolic procedure status;
   %
   %   STATUS;
   %
   begin scalar x,y;
      lnnum!* := lnnum!* + 1;
      orig!* := 0;
      status!-prt0("Tensor:",ordn tensorlist!*,2,t);
      status!-prt0("Symbol:",ordn symbollist!*,2,t);
      status!-prt0("Index:",indexlist!*,3,t);
       x := list(!*metric . 'METRIC,
                !*levi!-civita . 'LEVI!-CIVITA);
      for each j in x do 
         if car j then y := j . y;
      status!-prt0("Switched ON:",for each j in reverse y collect cdr j,4,t);
      status!-prt0("Used RELATION:",!*relation!*,2,nil);
   end;
  
symbolic procedure status!-prt0(x,y,n,bool);
   if y then << newlinepos 2; prin2 x; 
                stprint(y,n,bool);
                newlinepos 0; y>>;

symbolic procedure stprint(u,n,bool);
   begin integer m,orig;
      mkspaces n;
      m := posn();
      if bool
        then for each j in u do
               if flatsizec j < (linelength nil - posn())
                 then print!+space j
                else << newlinepos m; print!+space j>>
       else << print!+line car u;
               orig := orig!*;
               orig!* := m;
               for each j in cdr u do print!+line j;
               orig!* := orig>>
   end;

symbolic procedure print!+space u;
   << prin2 u; prin2 '! >>;

symbolic procedure print!+line u;
   begin scalar x;
      if pairp u then <<
        if car u memq '(bianchy1 bianchy2 bianchyc jacobi ordercov)
          then << x := list car u; 
                  u := if car u memq '(bianchy1 bianchy2)
                         then for each j in cdr u
                                 collect 'list . origsntx j
                        else cdr u>>
         else << x := first!-two u; 
                 u := cddr u>>;
        u := append(x,if u and cdr u then list('list . u) else u);   
        maprin('replaceby . u);
        terpri!* t>>
       else << prin2 u; terpri()>>
   end;

symbolic procedure origsntx u;
   if null u then nil
    else if listp car u and length car u = 3 then ('list . car u) . cdr u
    else car u . origsntx cdr u; 

symbolic procedure newlinepos n;
   % Begins a fresh line at posn N;
   begin
      lnnum!* := lnnum!* + 1;
      terpri();
      mkspaces(orig!* + n);
   end;

symbolic procedure mkspaces n; for i:=1:n do prin2 '!  ;

endmodule;

end;
