%********************************************************************;
%******                     Reversed relations               ********;
%********************************************************************;

module riccir!-revrel;

% Author: Jozsef Kadlecsik

global '(christoffel!* metric!* riemann!*);

fluid '(!*currten);

symbolic procedure gdiff!-ten u;                           
   diffopp!* u and (apply1(!*currten,u) or gdiff!-ten cadr u);

symbolic procedure no!-mult!-ten(u,v); 'one;               

symbolic procedure revrelindexsimp u;
   numr msqsimp0 ('times . mk!*expt u);                

% Reversed relation:  DEFINITION => RIEMANN;

% !*currtenfn := 'gdiff!-ten;                                 % !*currtenfn

% !*multtenfn := 'no!-mult!-ten;                              % !*multtenfn

symbolic procedure partd!-chris u;                            % !*currten
   if eqcar(u,'partdiff)
      and (pairp x and christoffelp x and atom cadr x
           and pairp caddr x and pairp caddr x where x=cadr u)
     then u else nil;

symbolic procedure right!-riemdef(u,ug,u1,uc,v,vg,v1,vc);     % !*righttenfn
   % Returns      u
   %         ( GAM         . (vz))   . (<ops> . sf))
   %                vz,w...
   begin scalar x,y,z,z1,w;
      x := gdiff!-ten u;
      y := cddr x; 
      z := list(cddr cadr x,reverse cddr cadr x);
      while null w and y do <<
         z1 := z;
         while null w and z1 do
            if null addmsf(multmsfsf(mkpairten(u1,list(cadar z1,car y)),uc),
                           v1 .* vc .+ nil)
              then w := car z1 . car y
             else z1 := cdr z1;
         y := cdr y>>;
      return if null w then nil
              else (mkrightten(u,car x . (caadr x . cadadr x . car w) .
                                 (cdr w . delete(cdr w,cddr x))) . car w) 
                        . (delete(u,u1) . uc)
   end;

symbolic procedure mkpairten(u,v);
   revrelindexsimp where!-list(reverse v,v,u);

symbolic procedure mkrightten(u,v);
   if apply1(!*currten,u) then v
    else car u . mkrightten(cadr u,v) . cddr u;

symbolic procedure revrel!-riemdef(u,v);                      % !*revrelfn
   revrel!-riemdef0(cadr u,cddr u,v);

symbolic procedure revrel!-riemdef0(u,v,w);
   ((if cdr v then ('partdiff . x . cdr v) . ('partdiff . y . cdr v) 
      else x . y) where y=list('minus,riemann!-def cdr x))
        where x=list(car riemann!*,cadr u,car w,car v,cadr w);

% Reversed relation: DEFINITION => CHRISTOFFEL

% !*currtenfn := 'gdiff!-ten;                                 % !*currtenfn

symbolic procedure mult!-metr(u,v);                           % !*multtenfn
   (if x then u . x else nil)
      where x=mult!-metr0(indicesinv metr!-ind u,v);

symbolic procedure metr!-ind u;
   if partd!-metr u then cdadr u else metr!-ind cadr u;

symbolic procedure mult!-metr0(u,v);
   if null v then nil
    else if metricp car v and xnp(u,cdar v) 
           then car v . mult!-metr0(u,cdr v)
    else mult!-metr0(u,cdr v);

flag('(mult!-metr),'more);

symbolic procedure partd!-metr u;                             % !*currten
   if eqcar(u,'partdiff)
      and (pairp x and metricp x 
           and pairp cadr x and pairp caddr x where x=cadr u)
     then u else nil;

symbolic procedure right!-chrsdef(u,ug,u1,uc,v,vg,v1,vc);     % !*righttenfn
   % Returns
   %          ( G        . (i v w) ) . (<ops> . sf)
   %             nv,w...
   begin scalar x,y,y1,z,z1,w;
      x := gdiff!-ten u;
      y := y1 := cddr x; 
      z := list(cdr cadr x,reverse cdr cadr x);
      while null w and y do <<
         z1 := z;
         while null w and z1 do
            if null addmsf(multmsfsf(mkpairten(u1,list(cadar z1,car y1)),uc),
                           v1 .* negf vc .+ nil)
              then w := car z1 . car y1
             else z1 := cdr z1;
         y1 := cdr y1>>;
      return if null w then nil
              else (mkrightten(u,car x . (caadr x . car w) .
                                 (cdr w . delete(cdr w,cddr x))) 
                     . list(car delete(indicesinv car w,cdr ug),
                            cadar w,cdr w)) . (delete(u,delete(ug,u1)) . uc)
   end;

symbolic procedure revrel!-chrisdef(u,v);                     % !*revrelfn
   revrel!-chrisdef0(cadr u,cddr u,v);

symbolic procedure revrel!-chrisdef0(u,v,w);
   (if cdr v then ('partdiff . x . cdr v) . ('partdiff . y . cdr v) 
     else x . y) where y=revrel!-chrsdef0 w,x=revrel!-chrsdef1 w;

symbolic procedure revrel!-chrsdef0 u;
   if atom car u 
     then list('minus,list('times,2,
                           christoffel!-def(car u,cadr u,caddr u)))
    else list('times,2,list(metric!*,car u,'(minus !&008)),
                     list('minus,christoffel!-def('!&008,cadr u,caddr u)));
  
symbolic procedure revrel!-chrsdef1 u;
   list('times,2,
        if atom car u then christoffel!* . u
         else list('times,list(metric!*,car u,'(minus !&008)),
                          list(christoffel!*,'!&008,cadr u,caddr u)));
 
% Reversed relation: RIEMANN => SECONDCOV

symbolic procedure riem!-ten u;                               % !*currtenfn
   apply1(!*currten,u) or (diffopp!* u and riem!-ten cadr u);

symbolic procedure mult!-rten(u,v);                           % !*multtenfn
   u . for each j in reverse ordn mult!-rten0(indicesinv riem!-ind u,v)
          collect cdr j;

symbolic procedure riem!-ind u;
   if diffopp u then riem!-ind cadr u else cdr u;

symbolic procedure mult!-rten0(v,w);
   if null w then nil
    else (if x 
            then append(for each j in x collect y . (car w . j),
                        mult!-rten0(v,cdr w)) where y=diffoprank car w
           else mult!-rten0(v,cdr w)) where x=xn(indices car w,v);

symbolic procedure diffoprank u;
   if atom u or not diffopp u then 0
    else length cddr u + diffoprank cadr u;

flag('(mult!-rten),'more);

symbolic procedure riemp u;                                   % !*currten
   if pairp u and riemannp u then u else nil;

% !*righttenfn := nil;                                        % !*righttenfn

symbolic procedure revrel!-riemdef2(u,v);                     % !*revrelfn
   ((list('times,cadr riemann!*,
                 list('plus,'d . y . reverse cddr x,
                            list('minus,'d . y . cddr x)))
    . list('times,car x,riemanndef2(y,cddr x))) 
               where y=indexsubst(cadr x,cdr v,car v))
         where x=which!-indx(cdr u,indexinv cdr v);

symbolic procedure which!-indx(u,v);
   if v member first!-two u
     then if atom v
            then if car u = v then (-1) . cdr u
                  else 1 . (car u . cddr u)
           else if car u = v then 1 . cdr u
                 else (-1) . (car u . cddr u)
    else which!-indx(append(cddr u,first!-two u),v);  

endmodule;

end;
