%********************************************************************;
%*****          Simplification of tensors and arguments          ****;
%********************************************************************;

module riccir!-simp;

% Author: Jozsef Kadlecsik
 
global '(!*msfone!* !*msqnull!* christoffel!* metric!* riemann!* 
         stdim!* weyl!* 
         !*resubs);

fluid '(!*curralg!* !*metric !*symmsign
        !*nosubs);

% !*symmsign  := nil;      %to indicate that there is need for sign
                           %exchange after using the symmetry prop's;

symbolic procedure riccir!-revlis u;
   if atom u then reval u 
    else car u . (
     if diffopp u then riccir!-revlis cadr u . revalindexlist cddr u
      else if tensorp u or symbolp u then revalindexlist cdr u
      else for each j in cdr u collect riccir!-revlis j); 

symbolic procedure revalindexlist u;
   for each j in u collect
      (if atom x then x 
        else if car x eq 'minus and atom cadr x then x 
        else typerr(j,"index")) where x=revalind j;

symbolic procedure revalind u;  % originally from EXCALC
   % to avoid the simplification of -0 to 0;
   begin scalar x,y;
      x := subfg!*;
      subfg!* := nil;
      if null numr(y := riccirsimp!* subla(list(0 . '!0),u))
        then return '!0;
      subfg!* := x;
      return prepsq !*msq2sq y
   end;

put('symmetrize,'simpfn,'simpsymmetrize);

put('symmetrize,'rtype,'indexed!-quantity);

put('symmetrize,'riccirfn,'riccirsymmetrize);

put('antisymmetrize,'simpfn,'simpsymmetrize);

put('antisymmetrize,'rtype,'indexed!-quantity);

put('antisymmetrize,'riccirfn,'riccirsymmetrize);

flag('(symmetrize antisymmetrize),'full);

symbolic procedure simpsymmetrize u;
   !*msq2sq riccirsymmetrize u;

symbolic procedure riccirsymmetrize u;
   begin scalar x,y,fn;
      if null cddr u or cdddr u 
        then rederr list("Wrong number of arguments to",u);
      fn := car u;
      x := caddr u;
      u := cadr u;
      if not eqcar(x,'list) 
        then rederr list("Wrong second argument:",list(fn,u,x))
       else if not sameposl(x := revalindexlist cdr x) or null cdr x
        then rederr list("Wrong second argument:",list(fn,u,'list . x));
      for each j in permutations x
         do y := (if fn eq 'antisymmetrize and not tpermp(j,x)
                    then list('minus,sub!-perm(indexpair(x,j),u)) 
                   else sub!-perm(indexpair(x,j),u)) . y;
      return riccirsimp!* list('quotient,'plus . y,factorial length x)
   end;

symbolic procedure factorial n;
   if n=1 or n=0 then 1 else n*factorial(n-1);

symbolic procedure sub!-perm(u,v);
   if null v or atom v then v
    else if diffopp v
     then if ldiffopp v then car v . sub!-perm(u,cadr v) . cddr v
           else car v . sub!-perm(u,cadr v) . indexsubla(u,cddr v)
    else if (tensorp v or symbolp v)
     then car v . indexsubla(u,cdr v)
    else if car v memq '(symmetrize antisymmetrize)
     then list(car v,sub!-perm(u,cadr v),indexsubla(u,caddr v))
    else car v . for each j in cdr v collect sub!-perm(u,j);

symbolic procedure riccirsimpatom u;
   if eqcar(riemann!*,u) and subfg!* and !*curralg!* = 4 
     then riccirsimp!* invcurv!-def()
    else mkmsq!* u;

symbolic procedure riccirsimpiden u;
   !*msq2sq ricciriden u;

symbolic procedure ricciriden u;
   % the simplification function for tensors and symbols;
   begin 
      if atom u then return riccirsimpatom u
       else if numberp cdr u then return exptmsq(riccirsimp!* car u,cdr u);
      u := car u . revalindexlist cdr u;
      tcheck(indices u,u);
      if subfg!* then <<
   % Metric tensor & Kronecker delta:  
   %   1. trace  => dimension of space-time;
      if (kroneckerp u or metricp u) 
         and dummyp(cadr u,caddr u) and stdim!*
        then return riccirsimp!* stdim!*
   % Kronecker delta;
   % 1. concrete indices;
       else if kroneckerp u and not indexp cadr u and not indexp caddr u
              then return if indexabs cadr u eq indexabs caddr u
                            then !*msfone!* ./ !*msfone!*
                           else !*msqnull!* 
   % Christoffel symbols:
   %   1. christoffel => definition;
       else if christoffelp u and !*curralg!* = 8 and metric!*
               and defindexpos cdr u
              then return riccirsimp!* christoffel!-def(cadr u,
                                                caddr u,cadddr u)
   % Ricci tensor:
   %   1. ricci => invcurv;
   %   2. ricci => riemann;
       else if riccip u 
              then (if !*curralg!* = 3 
                       and dummyp(cadr u,caddr u) 
                      then return riccirsimp!* car riemann!*
                     else if !*curralg!* = 2
                      then return riccirsimp!* 
                                    ricci!-def(cadr u,caddr u))
   % Riemann tensor: 
   %   1. traceless => 0
   %   2. riemann => ricci
   %   3. riemann => weyl
   %   4. riemann => definition
   %   5. bianchy1
       else if riemannp u 
              then (if symmpropp cdr u then return !*msqnull!*
                     else if !*curralg!* = 1 
                             and sdummy symbolic!-indices cdr u 
                      then return simpriemann cdr u
                     else if !*curralg!* = 6 
                             and weyl!* and weyl!-cond()
                      then return riccirsimp!*
                                   weyl!-in!-def(cadr u,caddr u,
                                          cadddr u,car cddddr u)
                     else if !*curralg!* = 7 
                             and christoffel!* 
                             and (defindexpos cdr u or metric!*)
                      then return riccirsimp!* riemann!-def cdr u
                     else if eqcar(!*curralg!*,'bianchy1)
                      then (if x then return riccirsimp!* bianchy(u,x))
                              where x=bianchy1p(cdr u,cdr !*curralg!*))
   % Weyl tensor:
   %   1. traceless => 0
   %   2. weyl => riemann
   %   3. bianchy1
       else if weylp u and stdim!* 
              then if stdim!* = 3 or sdummy symbolic!-indices cdr u
                     then return !*msqnull!*
                    else if !*curralg!* = 5 
                            and riemann!* and weyl!-cond()
                     then return riccirsimp!*
                                   weyl!-out!-def(cadr u,caddr u,
                                         cadddr u,car cddddr u)
                    else if eqcar(!*curralg!*,'bianchy1)
                      then (if x then return riccirsimp!* bianchy(u,x))
                              where x=bianchy1p(cdr u,cdr !*curralg!*)>>;
      return mkmsq u
   end; 

symbolic procedure defindexpos u;
   atom car u and not samepos(car u,cadr u) and sameposl cdr u;

symbolic procedure weyl!-cond;
   metric!* and stdim!* and (not numberp stdim!* or stdim!* geq 4);

symbolic procedure tcheck(u,v);
   if null u then T
    else if member(car u,cdr u) then rerror('riccir!-simp,1,
      list("An index occurs twice in cov. or contrav. position in",v))
    else tcheck(cdr u,v);   

symbolic procedure dummyp(u,v);
   indexp u and u = indexinv v;

symbolic procedure symmpropp u;
   dummyp(car u,cadr u) or dummyp(caddr u,cadddr u);
 
symbolic procedure simpriemann u;
   (if dummyp(x,z) then simpriemann0(y,v,1)
     else if dummyp(y,v) then simpriemann0(x,z,1)
     else if dummyp(x,v) then simpriemann0(y,z,-1)
     else simpriemann0(x,v,-1))
       where x=car u,y=cadr u,z=caddr u,v=cadddr u;

symbolic procedure simpriemann0(u,v,n);
   multmsqsq(riccirsimp!* list(car riemann!*,u,v),
             n*(cddr riemann!*) ./ 1);              % MTW definition;

symbolic procedure mkmsq u;
   begin scalar x,z,!*symmsign;
      if !*metric then << x := movable!-indices u;
                          u := apply!-metric0!*(u,setdiff(x,sfree x))>>;
      x := getsymm(u := mk!-dummy!* u);
      return if symm!-antisymm x then !*msqnull!*
              else if (z := symm!+opmtch(u,x,movable!-indices u)) 
               then multmsqsq(riccirsimp!* z,n) where n=mksymmsign()
              else << !*symmsign := nil;
                      u := apply!-symm(u,x,truesdummy indices u);
                      multmsqsq(mkmsq!* u,n) where n=mksymmsign()>>
   end;

symbolic procedure apply!-metric0!*(u,v);
   if null v then u
    else if indexinv car v member cdr v
     then if atom car v then apply!-metric0!*(u,cdr v)
           else apply!-metric0!*(where!-list(list(cadar v,car v),
                                             list(car v,cadar v),u),cdr v)
    else apply!-metric0!*(u,cdr v);

 symbolic procedure mksymmsign;
   (if !*symmsign then -1 else 1) ./ 1;

symbolic procedure mkmsq!* u;
   begin scalar x;
      if subfg!* then go to b;
   a: if null x then x := fkern u;
      return !*msk2msq list if atom car x then car x .* 1 else car x;
   b: if null !*nosubs and atom u and (x := get(u,'avalue))
        then go to c
       else if not('used!* memq cddr (x := fkern u)) then aconc(x,'used!*);
      go to a;
   c: if null !*resubs then !*nosubs := t;
      x := riccirsimp!* cadr x;
      !*nosubs := nil;
      return x
   end;

symbolic procedure kvalue!-opmtch u;
   begin scalar x,!*symmsign;
      if null subfg!* then go to a
       else if car u eq 'd and null cdddr u 
              then (if !*curralg!* = 9 and indexp indexabs caddr u
                       and (riemannp cadr u or riccip cadr u)
                       and (x := indexinv caddr u) member cdadr u
                      then return list riccirsimp!*
                                 if riemannp cadr u
                                   then riem!-biancc(cdadr u,x)
                                  else ricci!-biancc(cdadr u,x)
                     else if eqcar(!*curralg!*,'bianchy2) 
                             and riemannp cadr u
                             and (x := bianchy2p(append(cdadr u,cddr u),
                                                 cdr !*curralg!*))
                      then return list riccirsimp!* bianchy(u,x))
       else if x := symm!+opmtch(u,getsymm u,movable!-indices u) 
        then go to c;
   a: return nil;
   c: if null !*resubs then !*nosubs := t;
      x := multmsqsq(riccirsimp!* x,n) where n=mksymmsign();
      !*nosubs := nil;
      return list x
   end;

symbolic procedure symm!+opmtch(u,v,w);
   % U : tensor
   % V : symmetries U
   % W : movable indices of U
   if null subfg!* then nil
    else begin scalar x;
            if null (w := sdummy w) 
              then v := (del!-simple2 caar v . del!-simple2 cdar v) .
                        (del!-simple cadr v . del!-simple cddr v);
            x := indices u;
            return if caar v or cdar v 
                     then symm!+opmtch2(u,caar v,cdar v,x,w,nil)
                    else symm!+opmtch1(u,cadr v,cddr v,x,w,nil) 
         end;

symbolic procedure symm!+opmtch2(u,x,y,v,w,bool);
   % U : tensor
   % X : symmetric 2nd order symmetries
   % Y : antisymmetric 2nd order symmetries
   % V : indices of U
   % W : movable dummy indices of U
   % bool : do we need to change sign
   begin scalar a,b;
      if null(x or y) 
        then << x := cdr getsymm u;
                if null (w := sdummy movable!-indices u)
                   then x := del!-simple car x . del!-simple cdr x;
                return symm!+opmtch1(u,car x,cdr x,indices u,w,bool)>>
       else if null x 
              then << a := for each j in permutations car y collect
                    (nxtsymmsign(not tpermp(j,car y),bool) . appends j);
                      y := cdr y>>
              else << a := for each j in permutations car x
                              collect (bool . appends j);
                      x := cdr x>>;
      while null b and a do <<
         b := symm!+opmtch2(where!-list(cdar a,intersection(v,cdar a),u),
                            x,y,v,w,nxtsymmsign(caar a,bool));
         a := cdr a>>;
      return b
   end;

symbolic procedure symm!+opmtch1(u,x,y,v,w,bool);
   % U : tensor
   % X : symmetric 1nd order symmetries
   % Y : antisymmetric 1nd order symmetries
   % V : indices of U
   % W : movable dummy indices of U
   % bool : do we need to change sign
   if null(x or y) then dummy!+opmtch(u,possible!-dummy w,bool)
    else begin scalar a,b;
            if null x 
              then << a := for each j in permutations car y collect
                              (nxtsymmsign(not tpermp(j,car y),bool) . j);
                      y := cdr y>>
              else << a := for each j in permutations car x
                              collect (bool . j);
                      x := cdr x>>;
            while null b and a do <<
               b := symm!+opmtch1(where!-list(cdar a,intersection(v,cdar a),u),
                                  x,y,v,w,nxtsymmsign(caar a,bool));
               a := cdr a>>;
            return b
         end;

symbolic procedure dummy!+opmtch(u,v,bool);
   begin scalar x,y;
      if v then <<
        y := car v;
        while null x and v do <<
          x := opmtch where!-list(car v,y,u);
          v := cdr v>>>>
       else x := opmtch u;
      !*symmsign := bool;
      return x
   end;
         
symbolic procedure possible!-dummy u;
   possible!-dummy0 for each j in u collect list(j,indexinv j);

symbolic procedure possible!-dummy0 u;
   % the possible combination of the dummy indices;
   if null u then nil
    else (if null cdr u then list(x,y)
           else append(for each j in z collect append(x,j),
                       for each j in z collect append(y,j))
                           where z=possible!-dummy0 cdr u)
                     where x=car u,
                           y=reverse car u;

symbolic procedure del!-simple2 u;
   if null u then nil
    else (if sameposl x
             and null setdiff(x,symbolic!-indices x)
            then del!-simple2 cdr u
           else car u . del!-simple2 cdr u) where x=appends car u;

symbolic procedure del!-simple u;
   if null u then nil
    else if sameposl car u 
            and null setdiff(car u,symbolic!-indices car u)
           then del!-simple cdr u
    else car u . del!-simple cdr u;

symbolic procedure setsymmsign(u,v);
   if null u then v
    else setsymmsign(cdr u,nxtsymmsign(car u,v));

symbolic procedure nxtsymmsign(u,v);
   (u or v) and not(u and v);      
  
endmodule;

end;
