%********************************************************************;
%*****               Special symbol declarations                 ****;
%********************************************************************;

module riccir!-specs;

% Author: Jozsef Kadlecsik
 
global '(christoffelsign!* christoffel!* kronecker!* 
         levi!-civitasign!* levi!-civitas!* levi!-civitat!* metric!* 
         signature!* stdim!* riemann!* weyl!* 
         cursym!*);

fluid '(!*levi!-civita !*metric
        !*nat);

% metric!*        = the name of the metric tensor;
% kronecker!*     = the name of the Kronecker delta;
% christoffel!*   = the name of the Christoffel symbol;
% levi!-civitas!* = the name of the Levi-Civita symbol .  
%                               number of its indices + 1; 
% levi!-civitat!* = the name of the Levi-Civita tensor .  
%                               number of its indices + 1; 
% riemann!*       = the name of the Riemann tensor . 
%       sign according to the MTW def. of the Riemann tensor .
%       sign according to the MTW def. of the Ricci tensor; 
% weyl!*          = the name of the Weyl tensor; 

deflist('((metric specsymbolstat)
          (kronecker specsymbolstat)
          (christoffel specsymbolstat)
          (levi!-civita specsymbolstat)
          (riemann specsymbolstat)
          (weyl specsymbolstat)
         ),'stat);

deflist('((metric formgenclear)
          (kronecker formgenclear)
          (christoffel formgenclear)
          (levi!-civita formgenclear)
          (riemann formgenclear)
          (weyl formgenclear)
         ),'formfn);

deflist('((metric clear!-metric)
          (kronecker clear!-kronecker)
          (christoffel clear!-christoffel)
          (levi!-civita clear!-levi!-civita)
          (riemann clear!-riemann)
          (weyl clear!-weyl)
         ),'clearfn);

deflist('((clear!-metric formrlis)
          (clear!-kronecker formrlis)
          (clear!-christoffel formrlis)
          (clear!-levi!-civita formrlis)
          (clear!-riemann formrlis)
          (clear!-weyl formrlis)
         ),'formfn);

symbolic procedure specsymbolstat;
   begin scalar x,y;
      x := cursym!*;       % the name of the special symbol;
      if (y := scan()) memq '(tensor symbol delta) then scan()
       else specsymerr x;
      if flagp!*!*(cursym!*,'delim) then specsymerr x
       else return x . y . remcomma xread1 'lambda
   end;

symbolic procedure specsymerr u;
   symerr(if u eq 'metric then "METRIC"
           else if u eq 'kronecker then "KRONECKER"
           else if u eq 'christoffel then "CHRISTOFFEL"
           else if u eq 'levi!-civita then "LEVI!-CIVITA"
           else if u eq 'riemann then "RIEMANN"
           else "WEYL",T);

symbolic procedure specsymerr2(u,v);
   msgpri(u,"invalid in",list(v,"statement."),nil,T);

symbolic procedure specmsgpri(u,v);
   begin scalar nat1;
      nat1 := !*nat;
      !*nat := nil;
      terpri();
      lpri list"***";
      maprin u;
      indvarprt v;
      terpri!*(t);
      !*nat := nat1
   end;

symbolic procedure specmsgpri2(u,v);
   begin scalar nat1;
      nat1 := !*nat;
      !*nat := nil;
      terpri();
      lpri list "*****";
      if v then indvarprt u 
       else if cdr u then inprint('!*comma!*,0,u)
       else maprin u;
      lpri list " invalid in RIEMANN statement"; 
      terpri!*(t);
      !*nat := nat1;
      error1()
   end;

symbolic procedure metric u;
   % 
   %   METRIC TENSOR a(i,j);   
   %
   begin
      rmsubs();
      if car u eq 'tensor then u := check!-tensor(cdr u,2)
       else specsymerr2(car u,"METRIC");
      if not tensorp u then mktensor car u;
      u := car u . upping cdr u;
      mksymmetry list(u,'s . cdr u);
      if metric!* then 
        specmsgpri("The metric tensor redefined: ",u);
      metric!* := car u;
      !*metric := t;
   end;

symbolic procedure clear!-metric u;
   %
   %   CLEAR METRIC TENSOR;
   %
   << metric!* := signature!* := !*metric := nil>>;

symbolic procedure kronecker u;
   % 
   %   KRONECKER DELTA a(i,j);   
   %
   begin
      rmsubs();
      if car u eq 'delta then u := check!-tensor(cdr u,2)
       else specsymerr2(car u,"KRONECKER");
      if not tensorp u then mktensor car u;
      u := car u . downing cdr u;
      if kronecker!* then 
        specmsgpri("The Kronecker delta redefined: ",u);
      kronecker!* := car u;
   end;

symbolic procedure clear!-kronecker u;
   %
   %   CLEAR KRONECKER DELTA;
   %
   kronecker!* := nil;

symbolic procedure christoffel u;
   % 
   %   CHRISTOFFEL SYMBOL a(i,j,k);   
   %
   begin
      rmsubs();
      if car u eq 'symbol then u := check!-tensor(cdr u,3)
       else specsymerr2(car u,"CHRISTOFFEL");
      if not symbolp u then mksymbol car u;
      u := car u . indexabs cadr u . downing cddr u;
      mksymmetry list(u,'s . cddr u);
      newtokcall(christoffelsign!*,car u);
      if christoffel!* then
        specmsgpri("The Christoffel symbol redefined: ",u);
      christoffel!* := car u;
    end;

symbolic procedure clear!-christoffel u;
   %
   %   CLEAR CHRISTOFFEL SYMBOL;
   %
   << remnewtok(christoffelsign!*,christoffel!*);
      christoffel!* := nil>>;

symbolic procedure levi!-civita u;
   % 
   %   LEVI!-CIVITA SYMBOL | TENSOR a(i,j) [or a(i,j,k) 
   %                                        or a(i,j,k,l) 
   %                                        or .......... ];   
   begin scalar x;
      rmsubs();
      if not(car u memq '(tensor symbol))
        then specsymerr2(car u,"LEVI!-CIVITA");
      x := car u eq 'symbol;
      u := check!-tensor(cdr u,5);
      if atom u or length u = 2 then errpri2(u,T)
       else if (x and levi!-civitat!* 
                and length u neq cdr levi!-civitat!*) 
               or (not x and levi!-civitas!* 
                   and length u neq cdr levi!-civitas!*)
              then rerror('riccir!-specs,1,
          "The Levi-Civita objects have different number of indices.");
      if x and not symbolp u then mksymbol car u
       else if not(x or tensorp u) then mktensor car u;
      u := car u . downing cdr u;
      mksymmetry list(u,'a . cdr u);
      if x then <<newtokcall(levi!-civitasign!*,car u);
                  if levi!-civitas!* then
                    specmsgpri("The Levi-Civita symbol redefined: ",u);
                  levi!-civitas!* := car u . length u>>
           else <<if levi!-civitat!* then
                    specmsgpri("The Levi-Civita tensor redefined: ",u);
                  levi!-civitat!* := car u . length u>>;
      x := stdim!*;
      stdim!-signature(length cdr u,signature!*);
      if x then
        specmsgpri("The dimension of space-time redefined: ",stdim!*);
  end;

symbolic procedure clear!-levi!-civita u;
   %
   %   CLEAR LEVI-CIVITA SYMBOL | TENSOR:
   << if car u eq 'tensor and levi!-civitat!*
        then levi!-civitat!*:= nil;
      if car u eq 'symbol and levi!-civitas!*
        then << remnewtok(levi!-civitasign!*,levi!-civitas!*);
                levi!-civitas!* := nil>>;
      if not(levi!-civitat!* or levi!-civitas!*)
        then !*levi!-civita := nil>>;

symbolic procedure riemann u;
   % 
   %   RIEMANN TENSOR a(i,j,k,l) , PLUS | MINUS , PLUS | MINUS;   
   %
   begin scalar x;
      rmsubs();
      if car u eq 'tensor then u := cdr u
       else specsymerr2(car u,"RIEMANN");
      if length car u neq 5 then specmsgpri2(car u,t)
       else if cdr u
              then (if not(length u = 3  
                           and cadr u memq '(plus minus)
                           and caddr u memq '(plus minus))
                      then specmsgpri2(cdr u,nil))
             else rerror('riccir!-specs,2,
                         "Syntax error: missing PLUS,PLUS");
      x := cdr u;
      u := car u;
      if not tensorp u then mktensor car u;
      u := car u . downing cdr u;
      % Symmetry properties of the Riemann tensor:
      mksymmetry list(u,'a . list(cadr u,caddr u),
                        'a . cdddr u,
                        's . list(list('list,cadr u,caddr u),'list . cdddr u));
      % Symmetry property of the Ricci tensor:
      mksymmetry list(list(car u,cadr u,caddr u),'s . list(cadr u,caddr u));
      if riemann!* then
        specmsgpri("The Riemann tensor redefined:",u);
      riemann!* := car u . defsign car x . defsign cadr x;
   end;

symbolic procedure clear!-riemann u;
   %
   %   CLEAR RIEMANN TENSOR;
   %
   riemann!* := nil;

symbolic procedure defsign u;
   if u eq 'plus then 1 else -1;

symbolic procedure weyl u;
   % 
   %   WEYL TENSOR a(i,j,k,l);   
   %
   begin
      rmsubs();
      if car u eq 'tensor then u := check!-tensor(cdr u,4)
       else specsymerr2(car u,"WEYL");
      stdim!-signature(stdim!*,signature!*);
      if null stdim!* then
        rerror('riccir!-specs,3,
           "The dimension of space-time has not been declared.")
       else if numberp stdim!*
              then if stdim!* = 1 or stdim!* = 2
        then rerror('riccir!-specs,4,
             list("The Weyl tensor has no definition in",
                  stdim!*,"dimensional space-time."))
             else if stdim!* = 3
        then rerror('riccir!-specs,5,
             list("The Weyl tensor is zero tensor",
                  "in 3-dimensional space-time."));
      if not tensorp u then mktensor car u;
      u := car u . downing cdr u;
      mksymmetry list(u,'a . list(cadr u,caddr u),
                        'a . cdddr u,
                        's . list(list('list,cadr u,caddr u),'list . cdddr u));
      if weyl!* then
        specmsgpri("The Weyl tensor redefined:",u);
      weyl!* := car u;
   end;

symbolic procedure clear!-weyl u;
   %
   %   CLEAR WEYL TENSOR;
   %
   weyl!* := nil;

symbolic procedure check!-tensor(u,n);
   if length u neq 1 then errpri2(u,T)
    else if not(n = 5 or length car u = n + 1)
     then errpri2(car u,T)
    else car u;

endmodule;

end;
