%*********************************************************************;
%*****              Matching products, powers in msq              ****;
%*********************************************************************;

module riccir!-subsmsq;

% Author: Jozsef Kadlecsik

global '(!*msfone!* !*msqnull!* imatch!* ipowlis!* metric!*
         !*match !*resubs powlis1!*);

fluid '(!*levi!-civita
        !*asymp!* !*combinelogs !*exp !*mcd !*rationalize 
        !*structure !*sub2 dmode!*);

version!-dependent();

symbolic procedure mp3msq u;
   begin scalar x;
      x := !*sub2;    %save current enviroment;
      if subfg!* then <<
        if !*sub2 or ipowlis!* then u := subs2msq u;
        if imatch!* then u := subs3msq u>>;
      !*sub2 := x;
      return multmsq(mp3msf numr u,revpr mp3msf denr u)
   end;

symbolic procedure mp3msf u;
   if null u then !*msqnull!*
    else addmsq(mp3mst lt u,mp3msf red u);
   
symbolic procedure mp3mst u;
   (if tpow u = 1 
      then if null numr x then !*msqnull!*
            else !*dsq2msq x
     else multmsqsq(if y then mp3msq riccirsimp!* y
                     else !*msk2msq tpow u,x) 
                       where y=call!-levi!-civita!-rel tpow u) 
          where x=sq!-subs2 tc u;

symbolic procedure sq!-subs2 u;
   begin scalar !*asymp!*,xexp,v,w,x;
      u := u ./ 1;
      x := !*sub2;    %save current enviroment;
      if subfg!* then <<
        if !*sub2 or powlis1!* then u := subs2q u;
        if (!*match and numr u) and null !*exp 
          then <<xexp := t; !*exp := t; v := u; w := u := resimp u>>;
        u := subs3q u;
        if  xexp then <<!*exp := nil; if u=w then u := v>>>>;
      if !*combinelogs then u := clogsq!* u;
      % must be here, since clogsq!* can upset girationalizesq!:.
      if dmode!* eq '!:gi!: then u := girationalize!: u
       else if !*rationalize then u := rationalizesq u;
      !*sub2 := x;
      % If any leading terms have cancelled, a gcd check is required;
      if !*asymp!* and !*rationalize then u := gcdchk u;
      return u
   end;

symbolic procedure call!-levi!-civita!-rel u;
   begin scalar x,y;
      if null subfg!* then return nil;
      if !*levi!-civita and metric!* then x := comb(setdiff(u,scalpart u),2);
   a: if null x then return nil
       else if lcc!-cond(caar x,cadar x)
              then << y := levi!-civita!-rel(caar x,cadar x);
                      return if null(u := setdiff(u,car x)) then y
                              else 'times . (y . mk!*expt u)>>;
      x := cdr x;
      go to a;
   end;

% Routines for substituting powers;

symbolic procedure subs2msq u;
   % Perform power substitutions on u. Check whether substitions
   % on numerator and denominator change these before doing quotient.
   ((if denr x=1 and denr y=1 and numr x=v and numr y=w then u
      else multmsq(x,revpr y)) where x=subs2msf v, y=subs2msf w)
        where v=numr u, w=denr u;

symbolic procedure subs2msf u;
   begin scalar x;
      !*sub2 := nil;
      x := subs2msf1 u;
      if (!*sub2 or ipowlis!*) and !*resubs
        then if numr x=u and msfonep denr x then !*sub2 := nil
              else x := subs2msq x;
      return x
   end;

symbolic procedure subs2msf1 u;
   if null u or lpow u = 1 then u ./ !*msfone!*
    else begin scalar k,v,w,x,y,z;
            k := lpow u;
            z := !*msfone!* ./ !*msfone!*;
         a: if null k 
              then return addmsq(multmsq(if null w then !*dsf2msq lc u
                                          else reverse w .* lc u .+ nil 
                                                  ./ !*msfone!*,z),
                                 subs2msf1 red u);
            if numberp cdr(v := car k) then go to b;
         e: w := v . w;
            k := cdr k;
            go to a;
         b: x := ipowlis!*;
         c: if null x then go to d
             else if (y := mtchmsk(v,caar x,caddar x,caadar x,cdadar x))
                    then go to r;
            x := cdr x;
            go to c;
         r: z := multmsq(y,z);
            k := cdr k;
            go to a;
         d: if not eqcar(car v,'expt) or !*structure or cdr v = 1
              then go to e
             else y := riccirsimpexpt list(caar v,cadar v,
                                           list('times,caddar v,cdr v));
            go to r
         end;

symbolic procedure mtchmsk(u,v,w,flg,bool);
   % U is a msk-power, V a power to be matched against.
   % W is the replacement expression.
   % FLG is a flag which is T if an exact power match required.
   % BOOL is a boolean expression to be satisfied for substitution.
   % Value is the substitution msq if a match found, NIL otherwise;
   begin scalar x;
      x := mtchmsk1(u,v,flg,bool);
   a: if null x then return nil
       else if lispeval subla(car x,bool) then go to b;
      x := cdr x;
      go to a;
   b: v := divide(cdr u,subla(car x,cdr v));
      w := exptmsq(riccirsimp!* subla(car x,w),car v);
      if cdr v neq 0 then w := multmsq(w,list(car u .** cdr v) ./ !*msfone!*);
      return w
   end;

% Routines for matching products.

symbolic procedure subs3msq u;
   multmsq(subs3msf numr u,revpr subs3msf denr u);

symbolic procedure subs3msf u;
   % U is a msf.
   % Value is a msq with all product substitutions made;
   begin scalar x,z;
      z := nil ./ !*msfone!*;
   a: if null u then return z
       else if lpow u = 1 then return addmsq(z,u ./ !*msfone!*);
      x := subs3mst lt u;
      if not msfonep denr x or numr x neq (lt u .+ nil)
        then << if !*sub2 or ipowlis!* then x := subs2msq x;
                x := subs3msq x>>;
      z := addmsq(z,x);
      u := cdr u;
      go to a
   end;

symbolic procedure subs3mst u;
   % U is a mst.
   % Value is a msq for the substituted term.
   begin scalar x,y,z,v,w;
      w := car u;
      v := imatch!*;
   c: if null w 
        then return if (x := subs3dsf1(car u,y,cdr u,v)) then x
                     else u .+ nil ./ !*msfone!*;
      x := mtchmsk!*(car w,v);
      if null x then go to a                      %lpow doesn't match;
       else if null caar x then go to b;          %complete match found;
      y := car w . y;
      v := x;
   a: w := cdr w;
      go to c;                                    %check tc for match;
   b: y := car w . y;
      x := cddar x;                               %list(<subst value>,<denoms>);
      if (v := setdiff(cadr x,y)) then errach list('subs3mst,u,car v);
      y := multmsqsq(riccirsimp!* car x,cdr u ./ 1);
      z := setdiff(car u,cadr x);
      return if null z then y else multmsq(y,!*msk2msq z)
   end;

symbolic procedure subs3dsf1(u,v,w,l);
   % U is the no matching kernels;
   % V is the maching kernels;
   % W is a standard form (coefficient).
   % L is a list of possible matches.
   % Value is a msq with all product substitutions made;
   begin scalar x,y,z;
   a: if null l then return nil;
      y := car l;
      % The part list(<subst value>,<denoms>) of Y is stored in Z
      % and replaced by list(1,nil);
      z := cddr y;
      x := subs3f1(w,list list(car y,cadr y,1,nil),nil);
      if denr x neq 1 or numr x neq w then go to b;
      l := cdr l;
      go to a;
   b: if (y := setdiff(cadr z,u)) then errach list('subs3mst,u,car y);
      x := multmsqsq(riccirsimp!* car z,x);
      u := setdiff(u,cadr z);
      return if null u then x else multmsq(x,!*msk2msq u)    
   end;
 
symbolic procedure mtchmsk!*(u,v);
   % U is a msk, V a list of matching templates.
   % If a match is made, value is of the form:
   % list list(NIL,<boolean form>,<subst value>,<denoms>),
   % otherwise value is an updated list of templates;
   begin scalar flg,v1,w,x,y,z;
   e: if null v then return z;
      v1 := car v;
      w := car v1;
   a: if null w then go to d;
      x := mtchmsk1(u,car w,caadr v1,cdadr v1);
   b: if null x then go to c
       else if car (y := subla(car x,delete(car w,car v1))
                              . list(subla(car x,cadr v1),
                                     subla(car x,caddr v1),
                                     subla(car x,car w)
                                         . cadddr v1))
        then z := y . z
       else if lispeval subla(car x,cdadr v1) then return list y;
      x := cdr x;
      go to b;
   c: w := cdr w; 
      go to a;
   d: z := aconc(z,v1);   % Could also be append(z,list v1).
      v := cdr v;
      go to e
   end;

symbolic procedure mtchmsk1(u,v,flg,bool);
   % U is a msk, V a msk to be matched against.
   % FLG is a flag which is T if an exact power match required.
   % BOOL is a boolean expression to be satisfied for substitution.
   % Value is a list of possible free variable pairings which
   % match conditions;
   begin scalar x,y;
      if u = v then return list nil
       else if not numberp cdr u
        then return if not numberp cdr v then mchk!*(u,v) else nil
       else if not numberp cdr v then return nil
       else if not (x := mchk!*(car u,car v)) then return nil
       else if cdr v memq frlis!*
      % do not match a free power to 1 or a conflicting match.
        then if cdr u = 1 or not(x := powmtch(cdr v,x,cdr u))
               then return nil
              else return mapcons(x,cdr v . cdr u)
       else if (flg and not(cdr u = cdr v)) or (!*mcd and cdr u<cdr v)
        then return nil
       else return x
   end;
     
endmodule;

end;
