%********************************************************************;
%*****                    Symmetry declaration                   ****;
%********************************************************************;

module riccir!-symm;

global '(cursym!*);

% Author: Jozsef Kadlecsik

% 
%       symmetry a(i,j,-k,-l,m,n) => symmetric in {i,j},{-k,-l}
%                                    antisymmetric in {m,n},
%                a(i,j),b(i,j)    => symmetric
%                b(-i,-j,-k,-l)   => antisymmetric in {-i,-j},{-k,-l}
%                                    symmetric in {{-i,-j},{-k,-l}};
%
%       and stored as 
%
%        number of indices . 
%            list(list(list of 2nd symmetric,list of 2nd antisymmetric)
%                 list(list of 1st symmetric,list of 1st antisymmetric))
%
%     A: ((2 ((NIL NIL) ((1 2)) NIL)
%         (6 ((NIL NIL) ((1 2 D D U U) (U U -3 -4 U U)) ((U U D D 5 6)))))
%     B: ((2 ((NIL NIL) ((1 2)) NIL)
%         (4 ((((-1 -2) (-3 -4))) NIL) (NIL ((-1 -2 U U) (U U -3 -4)))))
%

put('symmetry,'stat,'symmetrystat);

symbolic procedure symmetrystat;
   begin scalar x,pr;
      flag('(replaceby symmetric antisymmetric),'delim);
      pr := get('replaceby,'infix);
      remprop('replaceby,'infix);
      x := errorset!*(symmetrystat0(),nil);
      % We have to use errorset!*, in order that even in the case
      % of syntax error, the original property and flag could 
      % be restored.
      remflag('(replaceby symmetric antisymmetric),'delim);
      put('replaceby,'infix,pr);
      return if errorp x then symerr('symmetry,t)
              else car x
   end;

symbolic procedure symmetrystat0;
   begin scalar x,y,z,u;
   a: scan();
   b: x := for each j in remcomma (xread1 nil) collect j . nil;
      if cursym!* neq 'replaceby then go to r
       else if scan() eq 'symmetric then z := 's
       else if cursym!* eq 'antisymmetric then z := 'a
       else go to r;
   d: if scan() eq 'in then go to c;
      x := for each j in x collect car j . ((z . cdar j) . cdr j);
      if cursym!* eq '!*comma!* then << u := append(x,u); go to a>>
       else if cursym!* eq '!*semicol!* then go to e
       else go to r;
   c: if scan() neq '!*lcbkt!* 
        then if cdar x then << u := append(x,u); go to b>>
              else go to r
       else y := z . cdr xreadlist();
      x := for each j in x collect car j . (y . cdr j);
      if cursym!* eq '!*comma!* then go to c
       else if cursym!* eq 'antisymmetric then << z := 'a; go to d>>
       else if cursym!* eq 'symmetric then << z := 's; go to d>>
       else if cursym!* eq '!*semicol!* then go to e;
   r: if z then symerr('symmetry,t)
       else rederr "Sorry, not implemented yet";
   e: return list('symmetry,mkquote append(x,u)) 
   end;

symbolic procedure symmetry u;
   begin rmsubs();
      for each j in u do mksymmetry j
   end;

symbolic procedure mksymmetry u;
   begin scalar a,s,x;
      x := car u;
      if not(tensorp x or symbolp x) then <<
        mktensor car x;
        msgpri(nil,car x,"defined as tensor id.",nil,nil)>>;
      for each j in cdr u do
         (if car j eq 's then s := y . s
           else a := y . a) where y=pick!-symm(x,cdr j);
      u := length cdr x . completing(s,a,car x);
      put(car x,'symmetry,
          if (s := get(car x,'symmetry)) and (a := assoc(car u,s))
            then (car u . completing!*(cdr a,cdr u,car x)) . delete(a,s)
           else u . s)
   end;

symbolic procedure pick!-symm(u,v);
   if listp v 
     then if eqcar(car v,'list)    % 2.nd order symmetry property;
            then if ordsubl(cdr u,v) 
                   then 2 . pick!-symm2(cdr u,v,cdr u)
                  else errpri2(v,t)
           else if xn(cdr u,v) = v % 1.st order symmetry property;
            then 1 . pick!-symm1(cdr u,v,1)
           else errpri2(v,t)
    else errpri2(v,t);

symbolic procedure ordsubl(u,v);
   null v or (eqcar(car v,'list) 
              and xn(u,cdar v) = cdar v 
              and (null cdr v or length car v = length cadr v)
              and ordsubl(u,cdr v));

symbolic procedure pick!-symm1(u,v,n);
   if null u then nil
    else if v and car u = car v 
     then (if atom car u then n else - n) . pick!-symm1(cdr u,cdr v,n + 1)
    else (if atom car u then 'u else 'd) . pick!-symm1(cdr u,v,n + 1);

symbolic procedure pick!-symm2(u,v,w);
   if null u then nil
    else if v and car u member cdar v
     then (for each j in cdar v collect sgnnth!?(j,w,1)) .
             pick!-symm2(setdiff(u,cdar v),cdr v,w)
    else sgnnth!?(car u,w,1) . pick!-symm2(cdr u,v,w);

%********************************************************************;
%***** Symmetry collecting :                                    *****;
%*****                        (ij)k   i(jk)      (ijk)          *****;
%*****                       a     , a      =>  a               *****;
%***** Symmetry checking:                                       *****;
%*****                        (ij)k   i[jk]                     *****;
%*****                       a     , a      =>  zero tensor     *****;
%*****                                                          *****;
%********************************************************************;

symbolic procedure completing(s,a,fn);
   check!-symmetry(completing0 symm2nd1st(s,nil,nil),
                   completing0 symm2nd1st(a,nil,nil),
                   fn);

symbolic procedure completing0 u;
   list(completing2 car u,completing1 cadr u);

symbolic procedure symm2nd1st(u,v,w);
   if null u then list(reverse v,reverse w)
    else if caar u = 1 then symm2nd1st(cdr u,v,cdar u . w)
    else symm2nd1st(cdr u,cdar u . v,w);

symbolic procedure completing1 u;
   if null u then nil
    else (if x then completing1(car x . delete(cdr x,cdr u))
           else car u . completing1 cdr u) 
               where x=newsymm1(car u,cdr u);

symbolic procedure newsymm1(u,v);
   if null v then nil
    else if commonsymm1p xn(u,car v)
            then (if x then x . car v else newsymm1(u,cdr v))
                    where x=newsymm10(u,car v,nil)
    else newsymm1(u,cdr v);

symbolic procedure commonsymm1p u;
   u and (numberp car u or commonsymm1p cdr u);

symbolic procedure newsymm10(u,v,w);
   if null u then reverse w
    else if car u = car v then newsymm10(cdr u,cdr v,car u . w)
    else if find!-symm1(car u,car v) then newsymm10(cdr u,cdr v,car u . w)
    else if find!-symm1(car v,car u) then newsymm10(cdr u,cdr v,car v . w)
    else nil;

symbolic procedure find!-symm1(u,v);
   numberp u and (minusp u and v eq 'd or (not minusp u and v eq 'u));
   
symbolic procedure completing2 u;
   if null u then nil
    else (if x then completing2(car x . delete(cdr x,cdr u))
           else car u . completing2 cdr u) 
               where x=newsymm2(car u,cdr u);

symbolic procedure newsymm2(u,v);
   if null v then nil
    else if commonsymm2p xn(u,car v)
            then (if x then x . car v else newsymm2(u,cdr v))
                    where x=newsymm20(u,car v,nil)
    else newsymm2(u,cdr v);

symbolic procedure commonsymm2p u;
   u and (pairp car u or commonsymm2p cdr u);

symbolic procedure newsymm20(u,v,w);
   if null u then reverse w
    else if car u = car v then newsymm20(cdr u,cdr v,car u . w)
    else (if x then newsymm20(cdr u,car x,car u . w)
           else if (x := find!-symm2(car v,u)) 
            then newsymm20(car x,cdr v,car v . w)
           else nil) where x=find!-symm2(car u,v);

symbolic procedure find!-symm2(u,v);
   begin scalar x,y;
      if atom u then return nil;
   a: if null u then return list v;
      if atom car v 
        then if find!-symm1(car u,car v) then v := cdr v
              else return nil
       else if car u member car v then rederr "Bad symmetry definition"
       else if find!-symm1(car u,x := firstnot2ndsymm v) then v := delete(x,v)
       else return nil;
      u := cdr u;
      go to a
   end;

symbolic procedure firstnot2ndsymm u;
   if null u then nil
    else if pairp car u then firstnot2ndsymm cdr u
    else car u;
 
symbolic procedure check!-symmetry(u,v,fn);
   if check!-symmetry2(car u,car v) or check!-symmetry1(cadr u,cadr v)
     then rerror('riccir!-symm,1,list(
       "Only the zero tensor satisfies the symmetry property of",
       fn))
    else list(list(ordn car u,ordn car v),
              list(reverse ordn cadr u,reverse ordn cadr v));

symbolic procedure check!-symmetry1(u,v);
   u and (newsymm1(car u,v) or check!-symmetry1(cdr u,v));

symbolic procedure check!-symmetry2(u,v);
   u and (newsymm2(car u,v) or check!-symmetry2(cdr u,v));

symbolic procedure completing!*(u,v,fn); 
   check!-symmetry(completing0 list(append(caar u,caar v),        
                                    append(caadr u,caadr v)),      
                   completing0 list(append(cadar u,cadar v),      
                                    append(cadadr u,cadadr v)),   
                   fn);

put('symmetry,'clearfn,'clear!-symmetry);
put('clear!-symmetry,'formfn,'formgenclear);

symbolic procedure clear!-symmetry u;
   %
   %   CLEAR SYMMETRY t1,t2,...tn;
   %
   begin scalar x,y;
      rmsubs();
      for each j in u do clear!-symm!-prop(j,'symmetry)
   end;

symbolic procedure clear!-symm!-prop(u,type);
   if pairp u then clear!-symm!-prop0(car u,length cdr u,type)
    else errpri2(u,t);

symbolic procedure clear!-symm!-prop0(u,n,type);
   begin scalar x,y;
      if (x := get(u,type)) and (y := assoc(n,x))
        then if (x := delete(y,x)) then put(u,type,x)
              else remprop(u,type)
       else nil
   end;

endmodule;

end;
