showtime;

% Declaration commands, notation of indices;

tensor a,b,c,f;
index i,j,k,l,m,n,o,p,q,t,x,y,z,w,a,b;

a(i,j);
a(-i,-j);

a(0,0);
a(-0,-0);

i*a(i,-i);

i**2*a(i,-i);
showtime;

% Examples with symmetry properties;

symmetry a(i,j,-k,-l) => antisymmetric in {i,j}
                         symmetric in {-k,-l},
         b(-i,-j,k,l) => symmetric in {-i,-j};

a(i,j,-i,-j);
a(i,j,-k,-l)*b(-i,-j,x,y);
a(i,j,-k,-l)*b(-i)*b(-j);

showtime;

% Clearing symmetry property...

clear symmetry a(i,j,-k,-l);

a(i,j,-i,-j);
a(i,j,-k,-l)*b(-i,-j,x,y);
a(i,j,-k,-l)*b(-i)*b(-j);
showtime;

clear symmetry b(-i,-j,k,l);

% Examples with symmetry property of second order;

symmetry a(i,j,k,l,m,n)=>antisymmetric in {{i,j},{k,l},{m,n}};

a(i,j,k,l,m,n)*b(-i,-j)*b(-m,-n);

a(i,j,k,l,m,n)*b(-i,-j,o)*c(-o)*b(-m,-n,p)*c(-p);

a(i,j,k,l,m,n)*c(-i,-j)*a(-k,-l)*b(-m,-n);

clear symmetry a(i,j,k,l,m,n);

symmetry a(i,j,k,l,m,n)=>symmetric in {{i,j},{k,l},{m,n}};

a(i,j,k,l,m,n)*b(-i,-j)*b(-m,-n);

a(i,j,k,l,m,n)*b(-i,-j,o)*c(-o)*b(-m,-n,p)*c(-p);

a(i,j,k,l,m,n)*c(-i,-j)*a(-k,-l)*b(-m,-n);

a(i,j,k,l,m,n)*b(-i,-j,o)*c(-o)*a(-k,-l)*b(-m,-n,p)*c(-p)*
a(x,y,z,w,a,b)*b(-x,-y,q)*c(-q)*a(-z,-w)*b(-a,-b,t)*c(-t);

a(i,j,k,l,m,n)*b(-i,-j,o)*c(-o)*a(-k,-l)*b(-m,-n,p)*c(-p)*
a(x,y,z,w,a,b)*b(-x,-y,q)*a(-q)*a(-z,-w)*b(-a,-b,t)*c(-t);

symmetry a(i,j,k,l,m,n)=>symmetric in {i,j},{k,l},{m,n};

a(i,j,k,l,m,n)*b(-i)*c(-j)*a(-l)*c(-k)*b(-n)*a(-m);

clear symmetry a(i,j,k,l,m,n);

% Declaration of special objects. 

metric tensor g(-i,-j);
levi!-civita tensor eps(-i,-j,-k);
riemann tensor r(-i,-j,-k,-l),plus,plus;
showtime;

% Some examples with the Riemann and Ricci tensors
% and symmetric, antisymmetric tensors;

tensor s;
index a,b,c,d,e,f,g,h;

symmetry a(-i,-j) => antisymmetric,
         s(-i,-j) => symmetric;

r(a,b)*r(-a,-i,-b,-j) + r(a,b)*r(-a,-j,-b,-i);

r(-a,-b,-c,-d)*r(-e,-f,c,d)*r(a,b,e,f) + r(-c,-d,g,h)*r(-e,-f,-g,-h)*r(c,d,e,f);

r(-a,-c)*r(-b,c)*r(a,b) + r(-a,-b)*r(-c,a)*r(b,c);

r(-a,-b,-c,-d)*r(a,b,c,d)+r(-a,-c,-b,-d)*r(a,c,b,d);

a(-a,-b)*s(a,b);

a(-b,-d)*s(-c,b)-a(-d,b)*s(-b,-c);

n*a(-b,-d)*s(-c,b)-m*a(-d,b)*s(-b,-c);

clear s;
clear symmetry a(-i,-j);
clear index a,b,c,d,e,f,g,h;

% Partial differentiation;

@(s,i);

@(a,i);

@(a(i),-i);

@(a(i)*b(-i),-j);

@(g(i,j),-k);

@(eps(i,j,k),-l);
showtime;

% Covariant differentiation;

d(s,i);

d(a,i);

d(a(i),-i);

d(a(i)*b(-i),-j);

d(g(i,j),-k);

d(eps(i,j,k),-l);

showtime;

% Lie differentiation;

ld(s,a(i));

ld(a(i)*a(-i),a(i));

ld(b(i),a(i));

ld(a(i),b(i)+c(i));

ld(a(i),a*b(i));

a(i):=c(i)*f(k)*f(-k);

ld(b(i),a(j,k,-k));

ld(b(i,j,-j),a(k,-l,l));

ld(b(i),a(j));

ld(a(i),b(j));

clear a(i);

a(i);

showtime;

symbol h,s,q;

% Assignment statements: I.;

a(i,j):=b(i,j,k,l)*c(-k,-l);

% We have not to remember what names nave been used
% for dummy indices;

a(k,l);
showtime;

% Assignment statements: II.;
% The assignment valids for the symbolic dummy indices;

a(i,-k,k,l):=c(i,k)*c(-k,l);

a(k,-l,l,j);

a(k,-0,0,j);
showtime;

% Assignment statements: III.;
% In the case of symbols different cases may exsist;

h(i,j):=s(i,j);

h(-i,-j):=q(-i,-j);

h(k,l);

h(-k,-l);

h(k,-l);

% Rule list;

r1 := {h(i,j) => a(i,-k,k,j)};

h(k,l) where r1;

h(k,l);

clear h,s,q,r1;

showtime;

% Mixed expressions;

a := sin(f*a(i)*b(-i));

d(a,-i);

operator bb;

bb(a(i)*a(-i)+1,b(i)*b(-i)+1);

a := bb(f*a(i)*b(-i));

d(a**2,-j);

for all z,x let d(bb(z),x)=d(z,x);   

d(a**2,-j);

showtime;

% Application of the metric tensor in expressions;

g(i,j)*a(-j);

g(i,j)*g(-j,k)*a(-k,l);

% Application of the metric tensor is suppressed...

off metric;

g(i,j)*a(-j);

g(i,j)*g(-j,k)*a(-k,l);
showtime;

clear a(i,j);

a(i,j);

% Product of Levi-Civita tensors can be transformed
% to expressions consisting of metric tensors only;

signature 1;

on metric,levi!-civita;

eps(i,j,k)*eps(l,x,y);

eps(i,j,k)*eps(l,x,-k);

eps(i,j,k)*eps(l,-j,-k);

eps(i,j,k)*eps(-i,-j,-k);

showtime;

% Symmetrizing, antisymmetrizing;

symmetry a(i,j),a(i,j,k)=>symmetric,
         b(i,j),b(i,j,k)=>antisymmetric;

symmetrize(a(i,j),{i,j});
symmetrize(a(i,j,k),{i,j,k});
symmetrize(c(i,j),{i,j});
symmetrize(c(i,j,k),{i,j,k});

antisymmetrize(b(i,j),{i,j});
antisymmetrize(b(i,j,k),{i,j,k});
antisymmetrize(c(i,j),{i,j});
antisymmetrize(c(i,j,k),{i,j,k});

symmetrize(c(-l,-j,-k),{-l,-j})*b(l,j);

symmetrize(antisymmetrize(c(i,j),{i,j}),{i,j});

showtime;

end;
