% Examples with RELATION

showtime;

tensor a,b,c,f,h;
index i,j,k,l,m,n;
metric tensor g(i,j);
christoffel symbol gam(i,j,k);
riemann tensor r(i,j,k,l),plus,plus;
showtime;

% Bianchy identities and Jacobi identity;

relation bianchy1=>{i,{-j,-k,-l}},
         bianchy2=>{{-k,-l,-m},i,j},
         jacobi=>ld(a(i),b(i),c(i));

r(i,-j,-k,-l) + r(i,-k,-l,-j);

d(r(i,j,-k,-l),-m) + d(r(i,j,-l,-m),-k);

ld(a(i),b(i),c(i)) + ld(b(i),c(i),a(i));

showtime;
status;

% Unfortunately STATUS doesn't work yet quite nicely in the case
% of RELATIONs. Please, patience :-);

% Applying of relations locally;

r(i,-j,-k,-l) + r(i,-k,-l,-j)
   where relation={bianchy1=>{i,{-j,-k,-l}}};

r(i,-j,-k,-l) + r(i,-k,-l,-j)
   where relation={bianchy1=>{~i,{-~j,-~k,-~l}}};

r(i,-j,-k,-l) + r(-j,-k,i,-l)
   where relation={bianchy1=>{{i,-j,-k},-l}};

d(r(i,j,-k,-l),-m) + d(r(i,j,-l,-m),-k)
    where relation={bianchy2=>{{-k,-l,-m},i,j}};

d(r(i,j,-k,-l),-m) + d(r(i,j,-l,-m),-k)
    where relation={bianchy2=>{{-~k,-~l,-~m},~i,~j}};

d(r(i,j,-k,-l),-m) + d(r(i,j,-l,-m),-k)
    where relation={bianchy2=>{{-l,-m,-k},i,j}};

d(r(i,j,-k,-l),-m) + d(r(j,-m,-k,-l),i)
    where relation={bianchy2=>{{i,j,-m},-k,-l}};

ld(a(i),b(i),c(i)) + ld(b(i),c(i),a(i))
    where relation={jacobi=>ld(a(i),b(i),c(i))};

ld(a(i),b(i),c(i)) + ld(b(i),c(i),a(i))
    where relation={jacobi=>ld(b(i),c(i),a(i))};

d(a(i),-j,-k)+d(a(i),-k,-j) where relation={ordercov};

d(a(i),-j,-k)+d(a(i),-k,-j) 
   where relation={ordercov=>d(a(i),-j,-k)};

d(a(i),-j,-k)+d(a(i),-k,-j) 
   where relation={ordercov=>d(a(i),-k,-j)};

%d(a(i),-j,-k)+d(a(i),-k,-j) 
%   where relation={ordercov=>{d(a(i),-j,-k),d(a(i),-k,-j)}};

showtime;

% Transformation of partial derivatives to covariant derivatives;

relation partial => covariant;

@(a,-i);

@(a(i),-j);
showtime;

% Transformation of covariant derivatives to partial derivatives;

relation nil,covariant => partial;

d(a,-i);

d(a(i),-j);

showtime;

b(i,-j,-k) := d(a(i),-j,-k);
showtime;

% Does it work backwards?

relation nil,partial => covariant;

b(i,-j,-k) := b(i,-j,-k);

showtime;

% Yes, it's all right;

relation ordercov;

b(i,-k,-j);

status;

off metric;
showtime;
relation nil;

% Transformation of Lie-derivatives to covariant ones and back;

relation lie=>covariant;

h(i) := ld(a(i),b(i));

h(j) where relation={covariant=>lie};

f(i) := ld(a(i,j)*b(-j),b(i,j)*c(-j));

f(i) where relation={covariant=>lie};

showtime;
end;
