%********************************************************************;
%*****                  Utility functions                        ****;
%********************************************************************;

module riccir!-util;

% Author: Jozsef Kadlecsik

global '(christoffel!* indexlist!* indxlist!* kronecker!* 
         levi!-civitas!* levi!-civitat!* metric!* riemann!* 
         symbollist!* tensorlist!* weyl!*);

fluid '(dummy!* free!*);

indxlist!* := '(!&000 !&001 !&002 !&003 !&004 
                !&005 !&006 !&007 !&008 !&009);

% Symbolic indices;

symbolic procedure msfindices u;
   % symbolic indices of a modified standard form;
   if null u or lpow u = 1 then nil
    else union(argindices lpow u,msfindices red u);

symbolic procedure argindices u;
   % symbolic indices from a multiplication of tensors;
   if null u or numberp cdar u then nil 
    else append(indices car u,argindices cdr u);

symbolic procedure indices u;
   % symbolic indices of a main variable;
   if atom u then nil
    else if flagp(car u,'diffop) 
     then if car u eq 'ld then indices cadr u
           else append(indices cadr u,symbolic!-indices cddr u)
    else if get(car u,'riccirfn) eq 'ricciriden 
     then symbolic!-indices cdr u
    else nil;

symbolic procedure argallindices u;
   % all (symbolic and non symbolic) indices from a multiplication;
   if null u or numberp cdar u then nil
    else append(allindices car u,argallindices cdr u);

symbolic procedure allindices u;
   % all (symbolic and non symbolic) indices of a main variable;
   if atom u then nil
    else if flagp(car u,'diffop) 
     then if car u eq 'ld then allindices cadr u
           else append(allindices cadr u,cddr u)
    else if get(car u,'riccirfn) eq 'ricciriden then cdr u
    else nil;

% Movable symbolic indices;

symbolic procedure movable!-msfindices u;
   % symbolic movable indices of a modified standard form;
   if null u or lpow u = 1 then nil
    else if msfindices red u
     then xn(movable!-argindices lpow u,movable!-msfindices red u)
    else movable!-argindices lpow u;

symbolic procedure movable!-argindices u;
   % symbolic movable indices from a multiplication;
   metric!* and movable!-argindices0 u;

symbolic procedure movable!-argindices0 u;
   if null u or numberp cdar u then nil 
    else append(movable!-indices0 car u,
                movable!-argindices0 cdr u);

symbolic procedure movable!-indices u;
   % symbolic movable indices of a main variable;
   metric!* and movable!-indices0 u;

symbolic procedure movable!-indices0 u;
   if atom u then nil
    else if flagp(car u,'diffop) 
           then if commutep(car u,'metric)
                  then if car u eq 'ld then movable!-indices0 cadr u
                        else append(movable!-indices0 cadr u,
                                    symbolic!-indices cddr u)
                 else nil
    else if tensorp u then symbolic!-indices cdr u
    else nil;

% Indices with symmetry property;

symbolic procedure argsymm u;
   if null u or numberp cdar u then (nil . nil) . (nil . nil)
     else symm!-add(getsymm car u,argsymm cdr u);

symbolic procedure symm!-add(u,v);
   symm!-add0(car u,car v) . symm!-add0(cdr u,cdr v);

symbolic procedure symm!-add0(u,v);
   append(car u,car v) . append(cdr u,cdr v);

symbolic procedure getsymm u;
   if atom u then (nil . nil) . (nil . nil)
    else if flagp(car u,'diffop)
           then if cdddr u
                  then if car u eq 'partdiff 
                         then symm!-add(getsymm cadr u,
                                    (nil . nil) . ((list cddr u) . nil))
                        else if car u eq 'd and atom cadr u
                         then symm!-add(getsymm cadr u,
                                       (nil . nil) .
                                       ((list first!-two cddr u) . nil))
                        else getsymm cadr u
                 else getsymm cadr u
    else if get(car u,'riccirfn) eq 'ricciriden
     then getsymm!-prop(car u,cdr u,symbolp u or null metric!*)
    else (nil . nil) . (nil . nil);

symbolic procedure getsymm!-prop(u,v,bool);
   % U : name of tensor
   % V : args of tensor
   (if x then (pick!-indices2(v,caar x,bool) . 
               pick!-indices2(v,cadar x,bool))
            . (pick!-indices1(v,caadr x,bool) . 
               pick!-indices1(v,cadadr x,bool))
     else (nil . nil) . (nil . nil)) 
      where x=if (u := assoc(length v,get(u,'symmetry)))
                               then cdr u else nil;

symbolic procedure pick!-indices1(u,v,bool);
   if bool then pick!-indices10(u,v)
    else for each j in v collect pick!-indices1!*(u,j);

symbolic procedure pick!-indices10(u,v);
   if null v then nil
    else (if x then x . pick!-indices10(u,cdr v) 
           else pick!-indices10(u,cdr v))
              where x=pick!-indices11(u,car v,nil);
 
symbolic procedure pick!-indices11(u,v,w);
   if null u then reverse w
    else if numberp car v and same!-ind!-nr(car u,car v)
           then pick!-indices11(cdr u,cdr v,car u . w)
    else if idp car v and same!-ind!-idp(car u,car v)
            then pick!-indices11(cdr u,cdr v,w)
    else nil;

symbolic procedure same!-ind!-nr(u,v);
   if atom u then not minusp v else minusp v;

symbolic procedure same!-ind!-idp(u,v);
   v eq if atom u then 'u else 'd;

symbolic procedure pick!-indices1!*(u,v);
   if null v then nil
    else if idp car v then pick!-indices1!*(cdr u,cdr v)
    else car u . pick!-indices1!*(cdr u,cdr v);

symbolic procedure pick!-indices2(u,v,bool);
   if bool then pick!-indices20(u,v)
    else for each j in v collect pick!-indices2!*(u,j);

symbolic procedure pick!-indices20(u,v);
   if null v then nil
    else (if x then x . pick!-indices20(u,cdr v) 
           else pick!-indices20(u,cdr v))
              where x=pick!-indices21(u,car v,nil);
 
symbolic procedure pick!-indices21(u,v,w);
   if null v then reverse w
    else if numberp car v and same!-ind!-nr(nth(u,abs car v),car v)
           then pick!-indices21(u,cdr v,w)
    else (if pairp car v and x 
            then pick!-indices21(u,cdr v,x . w)
           else nil) where x=pick!-indices22(u,car v,nil);

symbolic procedure pick!-indices22(u,v,w);
   if null v then reverse w
    else (if same!-ind!-nr(x,car v)
            then pick!-indices22(u,cdr v,x . w)
           else nil) where x=nth(u,abs car v);

symbolic procedure pick!-indices2!*(u,v);
   if null v then nil
    else if numberp car v then pick!-indices2!*(u,cdr v)
    else (for each j in car v collect nth(u,abs j)) . pick!-indices2!*(u,cdr v);

% Dummy and free indices, symbolic indices;

symbolic procedure sdummy u;
   if null u then nil
    else if indexinv car u member cdr u 
     then indexabs car u . sdummy delete(indexinv car u,cdr u)
    else sdummy cdr u;

symbolic procedure sfree u;
   if null u then nil
    else if indexinv car u member cdr u 
     then sfree delete(indexinv car u,cdr u)
    else car u . sfree cdr u;

symbolic procedure free!-and!-dummy2 u;
   % return free . dummy of indices u;
   begin scalar dummy!*,free!*;
      free!-and!-dummy u;
      return free!* . dummy!*
   end;

symbolic procedure free!-and!-dummy u; 
   << free!-and!-dummy0 u;
      free!*  := reverse free!*;
      dummy!* := reverse dummy!*>>;    %simpler solution?

symbolic procedure free!-and!-dummy0 u;
   if null u then nil
    else if indexinv car u member cdr u 
           then << dummy!* := indexabs car u . dummy!*;
                   free!-and!-dummy0 delete(indexinv car u,cdr u)>>
    else << free!* := car u . free!*;
            free!-and!-dummy0 cdr u>>;

symbolic procedure symbolic!-indices u;
   % collect the symbolic indices from an indexlist;
   if null u then nil
    else if indexp car u then car u . symbolic!-indices cdr u
    else symbolic!-indices cdr u;

% Boolean functions;

symbolic procedure tensor!-exprp u;
   getrtype u eq 'indexed!-quantity;
 
symbolic procedure diffopp u;
   flagp(car u,'diffop);

symbolic procedure diffopp!* u;
   flagpcar(u,'diffop);

symbolic procedure ldiffopp u;
   flagp(car u,'lie!-type);

symbolic procedure tensorp u;
   car u memq tensorlist!*;

symbolic procedure symbolp u;
   car u memq symbollist!*;

symbolic procedure indexedp u;
   flagp(car u,'diffop) or get(car u,'riccirfn) eq 'ricciriden;

symbolic procedure indexedp!* u;
   (atom u and indexedp list u) or eqcar(u,'!*times);

symbolic procedure tvecp u;
   pairp u and tensorp u and null cddr u and atom cadr u;

symbolic procedure metricp u;
   eqcar(u,metric!*) and length u = 3;

symbolic procedure kroneckerp u;
   eqcar(u,kronecker!*) and length u = 3;

symbolic procedure christoffelp u;
   eqcar(u,christoffel!*) and length u = 4;

symbolic procedure riemannp u;
   riemann!* and eqcar(u,car riemann!*) and length u = 5;

symbolic procedure riccip u;
   riemann!* and eqcar(u,car riemann!*) and length u = 3;

symbolic procedure weylp u;
   weyl!* and eqcar(u,weyl!*) and length u = 5;

symbolic procedure levi!-civitasp u;
   levi!-civitap0(u,levi!-civitas!*);

symbolic procedure levi!-civitatp u;
   levi!-civitap0(u,levi!-civitat!*);

symbolic procedure levi!-civitap0(u,v);
   v and eqcar(u,car v) and length u = cdr v;

symbolic procedure downp u; eqcar(u,'minus);

symbolic procedure indexp u;
   (x memq indexlist!* or x memq indxlist!*) where x=indexabs u;

symbolic procedure samepos(u,v);
   (atom u and atom v) or (pairp u and pairp v);

symbolic procedure xnp(u,v);
   u and v and (car u member v or xnp(cdr u,delete(car u,v)));

% Algebraic boolean functions;

symbolic procedure contrav u; 
   atom car revalindexlist list u;

symbolic procedure dummy u; 
   indexp car revalindexlist list u;

flag('(contrav dummy),'boolean);

% Index functions;

symbolic procedure where!-dummy(u,v,w);
   where!-list(appinvs u,appinvs v,w);

symbolic procedure appinvs u;
   append(u,indicesinv u);

symbolic procedure indexsubst(indx1,indx2,w);
   where!-list(list indx1,list indx2,w);

symbolic procedure indexabs u;
   if atom u then u else cadr u;

symbolic procedure upping u;
   if null u then nil
    else indexabs car u . upping cdr u;

symbolic procedure downing u;
   if null u then nil
    else (if atom car u then list('minus,car u)
           else car u) . downing cdr u;

symbolic procedure indexinv u;
   if atom u then list('minus,u) else cadr u;

symbolic procedure indicesinv u;
   if null u then nil
    else indexinv car u . indicesinv cdr u;

symbolic procedure downs u;
   if null u then nil
    else if atom car u then downs cdr u
    else car u . downs cdr u;

symbolic procedure ups u;
   if null u then nil
    else if atom car u then car u . ups cdr u
    else ups cdr u;

symbolic procedure sameposl u;
   null u or null cdr u or (samepos(car u,cadr u) and sameposl cdr u);

symbolic procedure first!-two u;
   list(car u,cadr u);

symbolic procedure first!-three u;
   list(car u,cadr u,caddr u);

symbolic procedure appends u;
   if null u then nil else append(car u,appends cdr u);

symbolic procedure even!-odd!-perms u;
   begin scalar x,y;
      for each j in permutations u do
         if tpermp(j,u) then x := j . x
          else y := j . y;
      return x . y
   end;

symbolic procedure tpermp(u,v);
   if null u then t
    else if car u = car v then tpermp(cdr u,cdr v)
    else not tpermp(cdr u,subst(car v,car u,cdr v));

symbolic procedure nth!?(u,v,n);
   if null v then 0
    else if u = car v then n 
    else nth!?(u,cdr v,n+1);

symbolic procedure sgnnth!?(u,v,n);
   if null v then 0
    else if u = car v 
           then if atom u then n else -n 
    else sgnnth!?(u,cdr v,n+1);

endmodule;

end;
